/* bookeep.c   6-3-96  bookeeping functions for the Tierra Simulator */
/* Tierra Simulator V4.3: Copyright (c) 1991, 1992, 1993, 1994, 1995, 1996 */
/* Tom Ray & Virtual Life */
/* MemProf labels: #01XX, in use: none */
/* FEError labels: #01XX, in use: 0100-0107 */

#ifndef lint
static char sccsid[] = "@(#)bookeep.c    1.5 7/21/92";
#endif /* lint */

#include "license.h"
#include "tierra.h"
#include "globals.h"

void DivideBookeep(mc, nc)
    Pcells   mc; /* mc = mother cell */
    Pcells   nc; /* nc = daughter cell */
{   I8s      same = 0;
    I32s     flag;

    LastDiv = InstExe;
    if (!ce->d.fecundity && !mc->d.mut && !mc->d.flaw)
    {   mc->d.d1.flags = mc->d.flags;    /* record metabolic data 1st repl */
        mc->d.d1.inst = mc->d.inst + 1;
        mc->d.d1.mov_daught = mc->d.mov_daught;
    }
    mc->d.fecundity++;
    nc->d.gen.size = nc->mm.s;
    if (GeneBnker)
    {   nc->d.genome = soup + nc->mm.p;
        if (mc->mm.s == nc->mm.s &&    /* if cell breeds true */
            IsSameGen(nc->mm.s, soup + nc->mm.p, soup + mc->mm.p))
        {   if (mc->d.fecundity == 1)
                nc->d.d1.BreedTrue = mc->d.d1.BreedTrue = 1;
            nc->d.parent = mc->d.parent;
            nc->d.gen = mc->d.gen;
            nc->d.gi = mc->d.gi;
            nc->d.hash = mc->d.hash;
            same = 1;
        }
        else     /* if daughter is a new genotype (same = 0) */
        {   nc->d.parent = mc->d.gen; /* this will assign a gen.label */
        nc->d.hash = Hash(nc->d.gen.size,nc->d.genome) ;
            if (DiskBank)
                flag = 25;
            else
                flag = 0;
            CheckGenotype(&(nc->d), flag, &siz_sl, &sl); /* check .gen files */
            strcpy(nc->d.gen.label, Int2Lbl(nc->d.gi));
        }
        DivGenBook(nc, InstExe, reaped, 0, same, 0);
        DivGenBook(mc, InstExe, reaped, 1, same, 0);
    }
    mc->d.mov_daught = mc->d.mut = 0;
    OutDisk((I32s)'b', nc);
#if FRONTEND != STDIO
    FEStats(); 
#endif /* FRONTEND != STDIO */
}

void ReapBookeep(cr)
Pcells  cr; /* cell to be reaped */
{   OutDisk((I32s)'d', cr);
    if (GeneBnker)
    {   
      ReapGenBook(cr);
    }

#ifdef MICRO
    if (cr == BrkTrapCondBuf.curbrkdat.MicroSlice)
    {
      GoBrkClr(1);
      sprintf(Fbuf, "cell being monitored has died");
      FEError(-505, NOEXIT, NOWRITE, Fbuf);
      FEBanner();
      FEStats();
    }
#endif /*MICRO */

    InitCell(cr->q.this.a,cr->q.this.i,cr);
    NumCells--;
#ifdef MICRO
    ChkBrkBd();
#endif
    reaped = 1;
}

void MutBookeep(i)
    I32s i;
{
    I8s    md;
    Pcells cm; /* cell to be mutated */
    TCell  cmt;
    I32s   flag ;

    if (!GeneBnker || IsFree(i)) return;
    WhichCell(i, &cm, &md);
    cmt = *cm;
    if (md == 'd')
        return;
    OutDisk((I32s)'d', cm);
    cm->d.parent = cm->d.gen;    /* assign new genotype */
    cm->d.gi = -1;
    strcpy(cm->d.gen.label, "---");

    cm->d.hash = Hash(cm->d.gen.size,cm->d.genome);/* new hash <- mutation */
    if (DiskBank)
        flag = 25;
    else
        flag = 0;
    CheckGenotype(&(cm->d), flag, &siz_sl, &sl); /* by checking .gen files */
    strcpy(cm->d.gen.label, Int2Lbl(cm->d.gi));
    DivGenBook(cm, InstExe, reaped, 0, 0, 0) ;
    OutDisk((I32s)'b', cm);
    ReapGenBook(&cmt) ;
    cm->d.d1.flags = cm->d.d1.mov_daught = 0L;
    cm->d.fecundity = cm->d.flags = 0L;
    cm->d.d1.inst = cm->d.inst = 0L;
    cm->d.mut++;
}

void OutDisk(bd, nc)
    I32s bd;
    Pcells nc;
{   I32s ttime;
    I8s label[4];

    if (DiskOut)
    {   if (FirstOutDisk)
        {   FirstOutDisk = 0;
            BrkupCum = 0;
            BrkupCou = 1;
#ifdef IBM3090
            if (BrkupSiz)
                sprintf(Buff, "break.1.d");
            else sprintf(Buff, "tierra.run.d");
            oufr = fopen(Buff, "w");
#else /* IBM3090 */
            if (BrkupSiz)
                sprintf(Buff, "%sbreak.1", OutPath);
            else sprintf(Buff, "%stierra.run", OutPath);
            oufr = fopen(Buff, "w");
#endif /* IBM3090 */
            if (oufr == NULL)
            {   sprintf(Fbuf,
                   "Tierra OutDisk() 1: file %s not opened, exiting\n", Buff);
                FEError(-100, EXIT, NOWRITE, Fbuf);
            }
            sprintf(label, nc->d.gen.label);
#ifdef IBM3090
            Ascii2Ebcdic(label);
#endif /* IBM3090 */
            BrkupCum += fprintf(oufr, "%lx %c %ld", InstExe.i, (I8s) bd,
                nc->d.gen.size);
            if (GeneBnker)
                BrkupCum += 1 + fprintf(oufr, " %s\n", label);
            else BrkupCum += 1 + fprintf(oufr, "\n");
        }
        else
        {   ttime = InstExe.i - lo.time;
            if (ttime < 0)
                ttime += 1000000L;
            BrkupCum += fprintf(oufr, "%lx", ttime);
            if (lo.bd != bd)
                BrkupCum += fprintf(oufr, " %c", bd);
            if (lo.size != nc->d.gen.size)
                BrkupCum += fprintf(oufr, " %ld", nc->d.gen.size);
            if (GeneBnker && strcmp(lo.label, nc->d.gen.label))
            {   sprintf(label, nc->d.gen.label);
#ifdef IBM3090
                Ascii2Ebcdic(label);
#endif /* IBM3090 */
                BrkupCum += fprintf(oufr, " %s", label);
            }
            BrkupCum += 1 + fprintf(oufr, "\n");
            if (BrkupSiz && BrkupCum > BrkupSiz * 1024L)
            {   fclose(oufr);
                BrkupCum = 0;
                BrkupCou++;
#ifdef IBM3090
                sprintf(Buff, "break.%ld.d", BrkupCou);
                oufr = fopen(Buff, "w");
#else /* IBM3090 */
                sprintf(Buff, "%sbreak.%ld", OutPath, BrkupCou);
                oufr = fopen(Buff, "w");
#endif /* IBM3090 */
                if (oufr == NULL)
                {   sprintf(Fbuf,
                   "Tierra OutDisk() 2: file %s not opened, exiting\n", Buff);
                    FEError(-101, EXIT, WRITE, Fbuf);
                }
            }
        }
    }
    else
    {   if (FirstOutDisk) FirstOutDisk = 0;
        else
        {   ttime = InstExe.i - lo.time;
            if (ttime < 0) ttime += 1000000L;
        }
    }
    lo.bd = bd;
    lo.size = nc->d.gen.size;
    lo.time = InstExe.i;
    strcpy(lo.label, nc->d.gen.label);
    TimePop += (double) ttime *(double) NumCells;
    if ((I8s) bd == 'b')
        TimeBirth++;
    else TimeDeath++;
}

#ifdef VRML

void UpdateVrml()
{   FILE  *inf;
    I16u  i;
    struct in_addr addr;

    inf = fopen("vrml.tie", "w");
    fprintf(inf, "%s\n", inet_ntoa(LocalHostAddr));
    for (i = 0; i < NumNodes; i++)
    {   addr.s_addr = IPMap[i].address.node;
        fprintf(inf, "%s:%ld:%ld:%ld:%ld:%ld:%ld:%lu:%lu\n",
            inet_ntoa(addr),
            IPMap[i].data.Speed,
            IPMap[i].data.SoupSize,
            IPMap[i].data.NumCells,
     (I32s) IPMap[i].data.OS,
            IPMap[i].data.InstExec.m,
            IPMap[i].data.InstExecConnect.m,
            IPMap[i].data.TransitTime,
            IPMap[i].data.Time);
    }
    fclose(inf);
    system("run-mail");
}

#endif /* VRML */

void plan()
{   I32s OClock;

    OClock = TierraClock;
    TierraClock = (I32s) time(NULL);
    if (OClock)
        Speed = 1000000L / (TierraClock - OClock);
    if (reaped)
        CellsGarbageCollect(); /* call this function only on Tierra nodes */

#ifdef VRML
    UpdateVrml();
#endif /* VRML */

    /* adjust control variables */
    AdjCtrlVars();

    FEPlan(1);
#ifdef MON
    Mplan();
#endif /* MON */
    TimePop = 0.; /* these are used in CalcSoupStats() and FEPlan() */
    TimeBirth = TimeDeath = 0L;
    TimeStats = InstExe;
}

void stats()
{   if (GeneBnker)
    {   if (reaped)
        {   GBGarbageCollect();
#ifdef ERROR
            VerifyGB();
#endif /* ERROR */
        }
        CalcGBStats(sl, siz_sl); 
    }
    else
        CalcSoupStats();
    CalcTimeSoup(); /* call this function only on Tierra nodes */
}

void CalcTimeSoup()
{   Event  TimeSince;
    double  AvgBD, dt;

    if (InstExe.m)
    {   dt = (double) SubEvent(&InstExe, &TimeStats, &TimeSince);
        AvgBD =  (double) (TimeBirth + TimeDeath) / 2.;
        TimeGenIndiv = (I32s) (dt / AvgBD);
        AvgPop = TimePop / (double) dt; /* average pop over time-interval */
        Generations += AvgBD / AvgPop;
    }
}

void CalcAverages() /* call this function only on Tierra nodes */
{   I32s    tNumCells = 0, ar, ci;
    Pcells  tc; /* temporary cell pointer */

    AverageSize = 0;
    for (ar = 0; ar < NumCelAr; ar++) for (ci = 0; ci < CelArSiz; ci++)
    {   if (ar == 0 && ci < 2)
            continue;
        tc = &cells[ar][ci];
        if (tc->ld)
        {   tNumCells++;
            AverageSize += tc->d.gen.size;
        }
    }
#ifdef ERROR
    if (tNumCells != NumCells)
    {   sprintf(Fbuf,
            "Tierra CalcAverages() NumCells = %ld  count of cells = %ld\n",
            NumCells, tNumCells);
        FEError(-102, EXIT, NOWRITE, Fbuf);
    }
#endif /* ERROR */
    AverageSize /= tNumCells;
}

void CalcSoupMaxes() /* call this function only on Tierra nodes */
/* This function builds a local soup genebank from a scan of the soup */
/* evidently this function is never called */
{   I32s  lsiz_sl = 1, flag;
    SList  Fp Fp  lsl;

    if (DiskBank)
        flag = 25;
    else
        flag = 0;
    lsl = MkGBFromSoup(&lsiz_sl, &lsl, flag);
    CalcGBMaxes(lsl, lsiz_sl);
    FreeGB(&lsl, &lsiz_sl);
}

void CalcSoupStats() /* call this function only on Tierra nodes */
/* This function builds a local soup genebank from a scan of the soup */
{   I32s  lsiz_sl = 1, flag;
    SList  Fp Fp  lsl;

    if (DiskBank)
        flag = 25;
    else
        flag = 0;
    lsl = MkGBFromSoup(&lsiz_sl, &lsl, flag);
    CalcGBStats(lsl, lsiz_sl);
    FreeGB(&lsl, &lsiz_sl);
}

void CalcGBStats(lsl, lsiz_sl)
SList Fp Fp  lsl;
I32s  lsiz_sl;
{   I32s  si, gi, pop, mem, tNumCells;
    Pgl  pgl;

    tNumCells = AverageSize = MaxPop = MaxMemry = 0;
    for (si = lsiz_sl - 1; si >= 0; si--) /* each size class */
    {   if (lsl[si] && lsl[si]->num_c)
        {   if (GeneBnker)
            for (gi = lsl[si]->a_num - 1; gi >= 0; gi--) /* each genotype */
            {   if (((I32u) (pgl = lsl[si]->g[gi]) > 4)
                    && (pop = pgl->pop))
                {   mem = pop * si;
                    if (pop > MaxPop)
                    {   MaxPop = pop;
                        MaxGenPop.size = si;
                        strcpy(MaxGenPop.label, Int2Lbl(gi));
                    }
                    if (mem > MaxMemry)
                    {   MaxMemry = mem;
                        MaxGenMem.size = si;
                        strcpy(MaxGenMem.label, Int2Lbl(gi));
                    }
                    AverageSize += mem;
                    tNumCells += pop;
                }
            }
            else
            {   tNumCells += lsl[si]->num_c;
                AverageSize += lsl[si]->num_c * si;
            }
        }
    }
    if (tNumCells != NumCells)
    {   sprintf(Fbuf,
            "Tierra CalcGBStats() NumCells = %ld  count of cells = %ld\n",
            NumCells, tNumCells);
        FEError(-103, EXIT, NOWRITE, Fbuf);
    }
    AverageSize=(tNumCells)?(AverageSize/tNumCells):0;
}

void CalcGBMaxes(lsl, lsiz_sl)
SList Fp Fp  lsl;
I32s  lsiz_sl;
{   I32s  si, gi, pop, mem;
    Pgl  pgl;

    MaxPop = MaxMemry = 0;
    for (si = lsiz_sl - 1; si >= 0; si--) /* each size class */
    {   if (lsl[si] && lsl[si]->num_c)
        {   for (gi = lsl[si]->a_num - 1; gi >= 0; gi--) /* each genotype */
            {   if (((I32u) (pgl = lsl[si]->g[gi]) > 4)
                    && (pop = pgl->pop))
                {   mem = pop * si;
                    if (pop > MaxPop)
                    {   MaxPop = pop;
                        MaxGenPop.size = si;
                        strcpy(MaxGenPop.label, Int2Lbl(gi));
                    }
                    if (mem > MaxMemry)
                    {   MaxMemry = mem;
                        MaxGenMem.size = si;
                        strcpy(MaxGenMem.label, Int2Lbl(gi));
                    }
                }
            }
        }
    }
}

void FreeGB(lsl, lsiz_sl)
SList Fp Fp Fp lsl;
I32s * lsiz_sl;
{   I32s  si;
    I16s  gi;
    Pgl   tgl;

    if (*lsiz_sl < 1 || (*lsl) == NULL)
        return;

    for (si = 0; si < *lsiz_sl; si++) /* for each size */
    {   if ((*lsl)[si])
        {   if ((*lsl)[si]->num_c)
                NumSizes--;
            if ((*lsl)[si]->g)
            {   for (gi = 0; gi < (*lsl)[si]->a_num; gi++) /* for each gen */
                {   if ((I32u) ((*lsl)[si]->g[gi]) > 4)
                    {   tgl = (*lsl)[si]->g[gi];
                        if (tgl->genome)
                        {   tfree((void *) (tgl->genome),
                                si * sizeof(Instruction), 1403);
                            tgl->genome = NULL;
                        }
                        if (tgl->gbits)
                        {   tfree((void *) (tgl->gbits),
                                si * sizeof(GenBits), 1404);
                            tgl->gbits = NULL;
                        }
                        tfree((void *) ((*lsl)[si]->g[gi]), sizeof(GList),
                            1402);
                        (*lsl)[si]->g[gi] = NULL;
                        NumGenotypes--;
                    }
                }
                tfree((void *) ((*lsl)[si]->g),
                    (*lsl)[si]->a_num * sizeof(Pgl), 1401);
                (*lsl)[si]->g = NULL;
            }
            if ((*lsl)[si]->hash)
            {   tfree((void *) ((*lsl)[si]->hash),
                    (*lsl)[si]->a_num * sizeof(I32s), 1405);
                (*lsl)[si]->hash = NULL;
            }
            tfree((void *) (*lsl)[si], sizeof(SList), 1400);
            (*lsl)[si] = NULL;
        }
    }
    tfree((void *) (*lsl), *lsiz_sl * sizeof(SList Fp), 1903);
    (*lsl) = NULL;
    *lsiz_sl = 0;
    if (NumGenotypes || NumSizes)
    {   sprintf(Fbuf, "Tierra FreeGB(): NumGenotypes = %ld  NumSizes = %ld\n",
            NumGenotypes, NumSizes);
        FEError(-104, EXIT, NOWRITE, Fbuf);
    }
}

SList Fp Fp MkGBFromSoup(lsiz_sl, lsl, flag)
I32s * lsiz_sl, flag;
SList Fp Fp Fp lsl;
{   I32s  si, ar, ci, i, tNumCells = 0;
#if PLOIDY > 1
    I32s  j;
#endif
    I16s  gi;
    Pgl   tgl;
    Pcells tc; /* temporary cell pointer */

    *lsiz_sl = 1;
    NumSizes = NumGenotypes = 0;
    *lsl = (SList Fp Fp) tcalloc(1, sizeof(SList Fp), 1903);
    for (ar = 0; ar < NumCelAr; ar++) for (ci = 0; ci < CelArSiz; ci++)
    {   if (ar == 0 && ci < 2)
            continue;
        tc = &cells[ar][ci];
        if (tc->ld)
        {   si = tc->d.gen.size;
            gi = tc->d.gi;
            tNumCells++;
            if (GeneBnker && gi < 0)
            {   tc->d.hash = Hash(si, tc->d.genome);
                CheckGenotype(&(tc->d), flag, lsiz_sl, lsl);
                strcpy(tc->d.gen.label, Int2Lbl(tc->d.gi));
                gi = tc->d.gi;
            }
            if (si >= *lsiz_sl)
            {   *lsl = (SList Fp Fp) trecalloc((void *) *lsl,
                    (si + 1) * sizeof(SList Fp), *lsiz_sl * sizeof(SList Fp),
                    1903);
                *lsiz_sl = si + 1;
            }
            if (!(*lsl)[si])
            {   (*lsl)[si] = (SList Fp) tcalloc(1, sizeof(SList), 1400);
                if (GeneBnker)
                {   (*lsl)[si]->g =
                        (Pgl Fp) tcalloc(gi + 1, sizeof(Pgl), 1401);
                    (*lsl)[si]->hash =
                        (I32s Fp) tcalloc(gi + 1, sizeof(I32s), 1405);
                    (*lsl)[si]->a_num = gi + 1;
                }
            }
#ifdef ERROR
            if (!(*lsl)[si]->num_c)
            {   if ((*lsl)[si]->num_g)
                    FEError(-105, NOEXIT, NOWRITE,
           "Tierra MkGBFromSoup() !(*lsl)[si]->num_c but (*lsl)[si]->num_g");
            }
#endif /* ERROR */
            if (!(*lsl)[si]->num_c)
                NumSizes++;
            (*lsl)[si]->num_c++;
            if (GeneBnker)
            {   if (gi >= (*lsl)[si]->a_num)
                {   (*lsl)[si]->g = (Pgl Fp)
                        trecalloc((void *) ((*lsl)[si]->g),
                        (gi + 1) * sizeof(Pgl),
                        (*lsl)[si]->a_num * sizeof(Pgl), 1401);
                    (*lsl)[si]->hash = (I32s Fp)
                        trecalloc((void *) ((*lsl)[si]->hash),
                        (gi + 1) * sizeof(I32s),
                        (*lsl)[si]->a_num * sizeof(I32s), 1405);
                    (*lsl)[si]->a_num = gi + 1;
                }
                if ((I32u) (*lsl)[si]->g[gi] < 4)
                {   (*lsl)[si]->g[gi] = (Pgl)
                        tcalloc(1, sizeof(GList), 1402);
                }
                tgl = (*lsl)[si]->g[gi];
                if (!tgl->genome)
                {   tgl->genome = (FpInst)
                        tcalloc(si, sizeof(Instruction), 1403);
                    for (i = 0; i < si; i++)
#if PLOIDY == 1
                        tgl->genome[i] = tc->d.genome[i];
#else
                        for (j = 0; j < PLOIDY; j++)
                            tgl->genome[i][j] = tc->d.genome[i][j];
#endif
                }
                if (!tgl->gbits)
                {   tgl->gbits = (FpGenB) tcalloc(si, sizeof(GenBits), 1404);
                }
                if (!tgl->pop)
                {   (*lsl)[si]->num_g++;
                    NumGenotypes++;
                }
                tgl->pop++;
                tgl->hash = (*lsl)[si]->hash[gi] = tc->d.hash;
            }
        }
    }
#ifdef ERROR
    if (tNumCells != NumCells)
    {   sprintf(Fbuf,
            "Tierra MkGBFromSoup() NumCells = %ld  count of cells = %ld\n",
            NumCells, tNumCells);
        FEError(-106, EXIT, NOWRITE, Fbuf);
    }
#endif /* ERROR */
    return *lsl;
}

void CellsGarbageCollect()
{   I32s  ar, ci;
    I8s   LInUse = 0, PInUse = 0;
    Pcells  tc; /* temporary cell pointer */

    if (NumCelAr <= 2)
        return;
    ar = NumCelAr - 2; /* check for use in penultimate cells array */
    for (ci = 0; ci < CelArSiz; ci++)
    {   if (cells[ar][ci].ld)
        {   PInUse = 1;
            break;
        }
    }
    if (PInUse)
        return;
    ar = NumCelAr - 1; /* check for use in last cells array */
    for (ci = 0; ci < CelArSiz; ci++)
    {   if (cells[ar][ci].ld)
        {   LInUse = 1;
            break;
        }
    }
    if (LInUse)
        return;

    if (cells[ar])    /* free the last cells array */
    {   for (ci = 0; ci < CelArSiz; ci++)
        {   tc = &cells[ar][ci];
            if (tc->c.ar)
            {   tfree((void *) (tc->c.ar), tc->c.n * sizeof(Cpu), 1902);
                tc->c.ar = NULL;
            }
#ifdef NET
            if (tc->c.io.io)
            {   tfree((void *) (tc->c.io.io), tc->c.io.siz * sizeof(IOS),
                    700);
                tc->c.io.siz = 0;
                tc->c.io.io = NULL;
            }
#endif /* NET */
#ifdef IO
            if (tc->c.pg.pb)
            {   tfree((void *) (tc->c.pg.pb), tc->c.pg.psiz * sizeof(Reg),
                    704);
                tc->c.pg.psiz = 0;
                tc->c.pg.pb = NULL;
            }
            if (tc->c.pg.gb)
            {   tfree((void *) (tc->c.pg.gb), tc->c.pg.gsiz * sizeof(Reg),
                    703);
                tc->c.pg.gsiz = 0;
                tc->c.pg.gb = NULL;
            }
#endif /* IO */
        }
        tfree((void *) cells[ar], CelArSiz * sizeof(TCell), 1901);
        cells[ar] = NULL;
    }
    NumCelAr--;
    cells = (Pcells Fp) trecalloc((void *) cells,
        (I32u) NumCelAr * sizeof(Pcells Fp),
        (I32u) (NumCelAr + 1) * sizeof(Pcells Fp), 1900);
    CellsSize = NumCelAr * CelArSiz;
}

I16s Lbl2Int(s)
    I8s *s;
{
#ifdef BIGNAMES
    I32s  index ;

    if (s[0] == '-')
    return -1;
    if (s[0] >= 'a')
      index = s[0] - 'a' ;
    else
      index = 26 + s[0] - 'A' ;
    if (s[1] >= 'a')
      index += (s[1] - 'a') * 52  ;
    else
      index += (26 + s[1] - 'A') * 52  ;
    if (s[2] >= 'a')
      index += (s[2] - 'a') * 52 * 52 ;
    else
      index += (26 + s[2] - 'A') * 52 * 52 ;
    return index ;

#else /* BIGNAMES */
   
    if (s[0] == '-')
    return -1;
    return (I16s) ((s[2] - 'a') + (26 * (s[1] - 'a')) +
           (676 * (s[0] - 'a')));

#endif /* BIGNAMES */
}

I8s *Int2Lbl(i)
    I32s i;
{
    static I8s s[4];

#ifdef BIGNAMES
    I32s index ;

    if (i < 0)
    {   strcpy(s, "---");
        return s;
    }
    index = i / (52 * 52) ;
    if (index >= 26)
      s[0] = 'A' + index - 26;
    else
      s[0] = 'a' + index ;
    i %= (52 * 52) ;
    index = i / 52 ;
    if (index >= 26)
      s[1] = 'A' + index - 26;
    else
      s[1] = 'a' + index ;
    i %= 52 ;
    index = i ;
    if (index >= 26)
      s[2] = 'A' + index - 26;
    else
      s[2] = 'a' + index ;
    s[3] = 0;
    return s;

#else /* BIGNAMES */

    if (i < 0)
    {   strcpy(s, "---");
        return s;
    }
    s[0] = 'a' + (I16s) i / 676;
    i %= 676;
    s[1] = 'a' + (I16s) i / 26;
    i %= 26;
    s[2] = 'a' + (I16s) i;
    s[3] = 0;
    return s;

#endif /* BIGNAMES */
}

I32u WhoIs(cp, a)
    Pcells  Fp cp; /* pointer to pointer to cell */
    I32s a;
{
    I8s md;

    if (a >= (*cp)->mm.p && a < (*cp)->mm.p + (*cp)->mm.s)
        return 0;          /* same cell */
    if (a >= (*cp)->md.p && a < (*cp)->md.p + (*cp)->md.s)
        return 1;          /* daughter cell */
    if (IsFree(a))
        return 3;          /* is free memory */
    WhichCell(a, cp, &md);
    if (md == 'm')
        return 2;          /* is other cell */
    return 4;              /* is the daughter of another cell */
}

I8s IsSameGen(size, g1, g2)/* compare two genomes */
    I32s size;
    FpInst g1, g2;
{
    I32s i;
#if PLOIDY > 1
    I32s j;
#endif

    for (i = 0; i < size; i++) 
#if PLOIDY == 1
        if (*(g1 + i) != *(g2 + i))
#else /* PLOIDY > 1 */
        for (j = 0; j < PLOIDY; j++)
            if (g1[i][j] != g2[i][j])
#endif /* PLOIDY > 1 */
            return 0;
    return 1;
}

void AdjCtrlVars()  /* adjust control variables */
{
    I32s pop_gen_time;
    double prob_of_hit;

    if (GenPerMovMut)
        RateMovMut = (I32s) 2L * GenPerMovMut * AverageSize * PLOIDY;
    if (JmpSouTra)
        RateJmpSou = (I32s) (2 * JmpSouTra * AverageSize);
    if (InstExe.m)
        pop_gen_time = AvgPop * TimeGenIndiv;
    else
    {   TimeGenIndiv = 10L * AverageSize;
        pop_gen_time=(AverageSize)?(TimeGenIndiv *
            (SoupSize / (4L * AverageSize))):0;
    }
    prob_of_hit = (double) AverageSize / (double) SoupSize;
    if (GenPerBkgMut)
        RateMut = (I32s) (pop_gen_time * 2L * GenPerBkgMut * prob_of_hit);
    if (GenPerFlaw)
        RateFlaw = (I32s) TimeGenIndiv * GenPerFlaw * 2L;
    if (DropDead) DropDead = 1L + AverageSize / 80L;  /* DAN */
    Search_limit = (I32s) (SearchLimit * AverageSize);
    Put_limit = (I32s) (PutLimit * AverageSize);
    MalLimit = MalTol * AverageSize;
    if (MalLimit >= SoupSize)
        MalLimit = SoupSize - 1;
}

