/* @(#)frontend.c    1.2 11/25/94*/
/* frontend.c   6-3-96  Tierra Simulator frontend routines */
/* Tierra Simulator V4.3: Copyright (c) 1991, 1992, 1993, 1994, 1995, 1996 */
/* Tom Ray & Virtual Life */
/* MemProf labels: #05XX, in use: none */
/* FEError labels: #05XX, in use: 0500-0505 */

/* frontend rationale:  to have a common set of functions for passing info
   from Tierra to the front end.  Any particular front end will likely
   use only a subset of all frontend functions.  Those functions that a
   particular front end does not use, will either have to be implemented
   here as dummy functions, or they will have to be ifdef'ed where they
   are called from Tierra */

#ifndef lint
static char     sccsid[] = "@(#)frontend.c    1.5 7/21/92";
#endif /* lint */

#include "license.h"
#include "tierra.h"
#ifdef __GUI__
#define WEXTERN_H
#endif /* __GUI__ */
#include "globals.h"
#ifdef ALCOMM
#include "tmonitor.h"
#include "trequest.h"
#endif /* ALCOMM */
#if INTERFACE == BASICD
#include <gppconio.h>
#endif /* INTERFACE == BASICD */

#ifdef __GUI__
#include "wntierra.h"
static char *ENDBLANK[80];
#endif /* __GUI__ */

#ifdef MON
#include "obs.h"
#include "tmon_global.h"
#endif /* MON */

typedef struct {
    I32s  x; /* x position (column) */
    I32s  y; /* y position (row) */
} Spos; /* last x,y position of cursor on screen */

Spos  spos;

#if (INTERFACE == BASICB) || (INTERFACE == BASICD)

I32s  DispText = 14, DispBack = 0, InfoText = 15, InfoBack = 6;
I32s  COLS = 80, LINES = 25;

#endif /* (INTERFACE == BASICB) || (INTERFACE == BASICD) */

void FECeol()
{
#if (INTERFACE == BASICB) || (INTERFACE == BASICD)
    clreol();
#endif /* (INTERFACE == BASICB) || (INTERFACE == BASICD) */

#if INTERFACE == BASICC
    clrtoeol();
    refresh();
#endif /* INTERFACE == BASICC */

#if INTERFACE == BASICM
     GrafPtr            oldPort;
     Rect               destRect;

     GetPort( &oldPort );
     SetPort( (GrafPtr)wp );
     destRect.left = spos.x*((windowWidth-displayOffset+5)/COLS1)
                     + (displayOffset +5);
     destRect.right = windowWidth;
     destRect.top = spos.y*rowHeight;
     destRect.bottom = (spos.y+1)*rowHeight;
     EraseRect( &destRect );
     SetPort( oldPort );
#endif /* INTERFACE == BASICM */

#if INTERFACE == BASICA
    clrtoeol();
    refresh();
#endif /* INTERFACE == BASICA */

#if INTERFACE == BASICV
    clrtoeol();
    refresh();
#endif /* INTERFACE == BASICV */

}

void FEClear()
{
#if FRONTEND == STDIO
    printf("\n");
    spos.x = spos.y = 0;
#endif /* FRONTEND == STDIO */

#if (INTERFACE == BASICB) || (INTERFACE == BASICD)
#ifndef __GUI__
    clrscr();
#else /* __GUI__ */
    int i ;
    for (i = 0; i < 25; i++)
        FEPrintf (0, i, 0, 0, " ");
    EndPaint(hwnd, &ps);
    InvalidateRect(hwnd, NULL, TRUE);
    hdcnew = BeginPaint(hwnd, &ps);
    SelectObject(hdcnew, GetStockObject(SYSTEM_FIXED_FONT));
#endif /* __GUI__ */

#endif /* (INTERFACE == BASICB) || (INTERFACE == BASICD) */

#if INTERFACE == BASICC
    clear();
    refresh();
#endif /* INTERFACE == BASICC */

#if INTERFACE == BASICM
    GrafPtr     oldPort;
    Rect        destRect;

    GetPort( &oldPort );
    SetPort( (GrafPtr)wp );
    destRect = wp->port.portRect;
    destRect.left = displayOffset + 1;
    EraseRect( &destRect );
    SetPort( oldPort );
#endif /* INTERFACE == BASICM */

#if INTERFACE == BASICA
    clear();
    refresh();
#endif /* INTERFACE == BASICA */

#if INTERFACE == BASICV
    clear();
    refresh();
#endif /* INTERFACE == BASICV */

}

void FEExit(ENO)
I32s ENO;
{
#if FRONTEND != STDIO
    sleep(1);
#endif /* FRONTEND != STDIO */

#ifdef MON
    if(MSock.fd != -1) {
      sprintf(Fbuf, "Closing connection to observation tool:fd = %d\n",MSock.fd);
      FEPrintf(HLP_X, HLP_Y, 1, -1, Fbuf);
      exit(END);
    }
#endif /* MON */

#ifdef ALCOMM2
    if (AL2_run_flag == 1)
    {   kill(AL_PID,SIGKILL);
        sprintf(Fbuf, "TIERRA: ALCOMM process is KILLED\n");
        FEPrintf(HLP_X, HLP_Y, 1, -1, Fbuf);
    }
#endif /* ALCOMM2 */

#if (INTERFACE == BASICB) || (INTERFACE == BASICD)
#ifndef __GUI__
    textmode(C80);
    textbackground(0);
    textcolor(15);
    clrscr();
#endif /* __GUI__ */
#endif /* (INTERFACE == BASICB) || (INTERFACE == BASICD) */

#if INTERFACE == BASICC
    echo(); nl(); nocbreak();
    clear(); refresh(); endwin();
#endif /* INTERFACE == BASICC */

#if INTERFACE == BASICA
    echo(); nl(); nocbreak();
    clear(); refresh(); endwin();
#endif /* INTERFACE == BASICA */

#if INTERFACE == BASICV
    echo(); nl();
    clear(); refresh(); endwin();
#endif /* INTERFACE == BASICV */

#ifndef __GUI__
    exit(ENO);
#else /* __GUI__ */
    MessageBox (hwnd, "Tierra quits, and Windows too!", "Tierra Info", MB_OK);
    ExitWindows (0,0); /* This was the only way to exit cleanly... */
#endif /* __GUI__ */    /* with non-preemptive multitasking...  */
}

I16s FEGetch(x, y)
I32s  x, y;
{
#ifndef AMIGA
    I16s  rv;
#else /* AMIGA */
#if FRONTEND == BASIC
    I16s  rv;
#endif /* FRONTEND == BASIC */
#endif /* AMIGA */

#if FRONTEND == STDIO
    fgets(Fbuf,84,stdin);
    return Fbuf[0];
#endif /* FRONTEND == STDIO */

#if FRONTEND == BASIC

#if INTERFACE == BASICC
    move(y,x);
#endif /* INTERFACE == BASICC */

#if INTERFACE == BASICM
        EventRecord event;
        short           character = 0;
        GrafPtr         oldPort;
        short           x_pixel, y_pixel;
        Rect            destRect;

        GetPort( &oldPort );
        SetPort( (GrafPtr)wp );
        x_pixel = x*((windowWidth-displayOffset+5)/COLS1) + displayOffset +5;
        y_pixel = (y+1)*rowHeight + fontInfo.ascent;
        destRect.left = x_pixel;
        destRect.right = destRect.left + CharWidth( character );
        destRect.top = y_pixel-fontInfo.ascent;
        destRect.bottom = y_pixel+fontInfo.descent + fontInfo.leading;

        while (! character)
        {       if (GetNextEvent(everyEvent, &event))
                {       if (event.what == keyDown)
                        {       character = event.message & charCodeMask;
                                EraseRect( &destRect );
                                MoveTo( x_pixel, y_pixel);
                                DrawChar( character );
                        }
                }
                else if (event.what == nullEvent)
                        character = 0;
        }
        SetPort( oldPort );
        return(character);
#endif /* INTERFACE == BASICM */

#if INTERFACE == BASICA
    nodelay(stdscr, FALSE);
    rv = mvgetch(y, x);
    nodelay(stdscr, TRUE);
    return rv;
#endif /* INTERFACE == BASICA */

#if INTERFACE == BASICV
    move(y, x);
    crmode();
    rv=getch();
    nocrmode();
    return rv;
#endif /* INTERFACE == BASICV */

#ifndef __GUI__
#if (INTERFACE != BASICA) && (INTERFACE != BASICV)
#ifndef __MWERKS__
    return getch();
#endif /* __MWERKS__ */
#endif /* (INTERFACE != BASICA) && (INTERFACE != BASICV)*/
#else /* __GUI__ */
    int i = ' '; /*initialize to 'space bar' -the return value if no key
         pressed GUI FEGetch is used only by Micro_Spy() in ttools.c */
    MSG msg;            /* to clear the Windows Message Queue */
    PeekMessage( &msg, (HWND) NULL, 0, 0, PM_REMOVE);
    if ( GetAsyncKeyState (VK_LBUTTON) < 0 )
    {   i = 'z';
        return i;
    }
    if ( GetAsyncKeyState (0x1b) < 0 )
    {   i = 0x1b;
        return i;
    }
    if ( GetAsyncKeyState ( 84 ) < 0 )
        i = 't';
    if ( GetAsyncKeyState ( 85 ) < 0 )
        i = 'u';
    if ( GetAsyncKeyState ( 78 ) < 0 )
        i = 'n';
    if ( GetAsyncKeyState (VK_SHIFT) < 0 )
        i = toupper(i);
    return i ;
#endif /* __GUI__ */
#endif /* FRONTEND == BASIC */
}

void FEBanner()
{
#if INTERFACE == BASICC
    I8s tt, dummy ;

    /* nice stats banner stuff */
    standout();
    /* not sure why the following indef is necessary but now it seems */
    /* to be.  - kurt  (feb. 13, 1994)                                */
    for(tt = 0; tt < (COLS1 - 1) ; tt++)
    {   dummy = tt ;
        mvaddch(0,dummy,' ');
        mvaddch(1,dummy,' ');
        mvaddch(LINES1-1,dummy,' ');
    }
    standend();
#endif /* INTERFACE == BASICC */

#if INTERFACE == BASICA
    I8s tt, dummy ;

    /* nice stats banner stuff */
    standout();
    for(tt = 0; tt < (COLS1 - 1) ; tt++)
    {   dummy = tt ;
        mvaddch(0,dummy,' ');
        mvaddch(1,dummy,' ');
        mvaddch(LINES1-1,dummy,' ');
    }
    standend();
#endif /* INTERFACE == BASICA */

#if INTERFACE == BASICV
    I8s tt, dummy ;

    /* nice stats banner stuff */
    standout();
    for(tt = 0; tt < (COLS1 - 1) ; tt++)
    {   dummy = tt ;
        mvaddch(0,dummy,' ');
        mvaddch(1,dummy,' ');
        mvaddch(LINES1-1,dummy,' ');
    }
    standend();
#endif /* INTERFACE == BASICV */

#if (INTERFACE == BASICB) || (INTERFACE == BASICD)
#ifndef __GUI__
    textbackground(InfoBack);
    textcolor(InfoText);
    gotoxy(1, 1);
    clreol();
    gotoxy(1, 2);
    clreol();
    gotoxy(1, HLP_Y + 1);
    clreol();
    textbackground(DispBack);
    textcolor(DispText);
#else /* __GUI__ */
    FEPrintf (20,0,1,0, ENDBLANK ); FEPrintf (0,0,1,0, ENDBLANK);
    FEPrintf (20,1,1,0, ENDBLANK ); FEPrintf (0,1,1,0, ENDBLANK);
#endif /* __GUI__ */
#endif /* (INTERFACE == BASICB) || (INTERFACE == BASICD) */
}

/* called immediately after the soup_in file is read */
/* an opportunity to interactively set soup_in vars */
void FEStartup()
{
#if (INTERFACE == BASICB) || (INTERFACE == BASICD)
#ifndef __GUI__
    FILE  *inf;
    I8s   first[12], second[12];
    I16s  gmode, color;
#endif /* __GUI__ */
#endif /* (INTERFACE == BASICB) || (INTERFACE == BASICD) */

#ifndef DECVAX
#if FRONTEND == BASIC
    LINES1=LINES;
    COLS1=COLS;
#endif
#endif

#if FRONTEND == STDIO

    fe_width = 60;
    MSG_X = 1;
    MSG_Y = 1;
    ERR_X = 1;
    ERR_Y = 1;
    PLN_X = 1;
    PLN_Y = 1;

#ifdef __TURBOC__
#ifndef __GUI__
    clrscr();
#else /* __GUI__ */
    FEClear();
#endif /* __GUI__ */
#else /* __TURBOC__ */
    system("clear");
#endif /* __TURBOC__ */

#endif /* FRONTEND == STDIO */

#if (INTERFACE == BASICB) || (INTERFACE == BASICD)

#if INTERFACE == BASICD

    gppconio_init();

#endif /* INTERFACE == BASICD */

#ifndef __GUI__

#if (INTERFACE == BASICB) && (__TURBOC__ >= 0x200) && (__TURBOC__ < 0x400)
    /* deal with old and (stupid) new compilers */

    FE_DV_Mode = DETECT;
    detectgraph(&FE_DV_Mode, &gmode);

#endif /* (INTERFACE == BASICB) && (__TURBOC__ >= 0x200) &&
          (__TURBOC__ < 0x400) */

    if(inf = fopen("tcolors.cfg","r"))
    {   fgets(Buff, 84, inf);
        do
        {   if (*Buff != '#' && strlen(Buff) > 3)
            {   sscanf(Buff, "%11s%11s%d", first, second, &color);
                if (!strcmp(first,"information"))
                {   if (!strcmp(second,"text:"))
                    InfoText = color;
                    else if (!strcmp(second,"background:"))
                    InfoBack = color;
                }
                else if (!strcmp(first,"display"))
                {   if (!strcmp(second,"text:"))
                    DispText = color;
                    else if (!strcmp(second,"background:"))
                    DispBack = color;
                }
            }
        } while (fgets(Buff, 84, inf) != NULL);
        fclose(inf);
    }
    textmode(C80);
    textbackground(DispBack);
    textcolor(DispText);
    clrscr();

#else /* __GUI__ */
    strcpy (ENDBLANK,
        "                                                                ");
    FEClear();
#endif /* __GUI__ */
    COLS1 = 80;
    LINES1 = 25;
    fe_width = 57;
    fe_lines = LINES1 - 1;
    ERR_Y = LINES1 - 2;
    HLP_Y = LINES1 - 1;

#endif /* (INTERFACE == BASICB) || (INTERFACE == BASICD) */

#if INTERFACE == BASICC

    initscr();
    LINES1=LINES;
    COLS1=COLS;
    if ((LINES1 < 20) || (COLS1 < 80))
    {   fprintf(stderr,"\nTIERRA: not enough space on screen ... \n");
        FEExit(-666);
    }
    FEClear();
    nonl(); cbreak(); noecho(); /* set up keyboard */
    fe_width = COLS1 - 23;
    fe_lines = LINES1 - 1;
    ERR_Y = LINES1 - 2;
    HLP_Y = LINES1 - 1;

#endif /* INTERFACE == BASICC */

#if INTERFACE == BASICM
    if ((LINES1 < 20) || (COLS1 < 80))
    {   FEPrintf(ERR_X, ERR_Y, 0, 0,
            "TIERRA: not enough space on screen ... \n");
        FEExit(-666);
    }
#endif /* INTERFACE == BASICM */

#if INTERFACE == BASICA

    initscr();
    LINES1=LINES;
    COLS1=COLS;
    if ((LINES1 < 20) || (COLS1 < 80))
    {   fprintf(stderr,"\nTIERRA: not enough space on screen ... \n");
        FEExit(-666);
    }
    FEClear();
    nonl(); cbreak(); noecho(); /* set up keyboard */
    fe_width = COLS1 - 23;
    fe_lines = LINES1 - 1;
    ERR_Y = LINES1 - 2;
    HLP_Y = LINES1 - 1;

#endif /* INTERFACE == BASICA */

#if INTERFACE == BASICV
    initscr();
    LINES1=LINES;
    COLS1=COLS;
    if ((LINES1 < 20) || (COLS1 < 80))
    {   fprintf(stderr,"\nTIERRA: not enough space on screen ... \n");
        FEExit(-666);
    }
    FEClear();
    fe_width = COLS1 - 23;
    fe_lines = LINES1 - 1;
    ERR_Y = LINES1 - 2;
    HLP_Y = LINES1 - 1;

#endif /* INTERFACE == BASICV */

    IMode = PLN_STATS;
    PLN_X = 0;
    PLN_Y = 2;
    ERR_X = 0;
    HLP_X = 0;
    MSG_X = 0;
#ifdef MEM_PROF
    MSG_Y = PLN_Y + 9;
#else  /* MEM_PROF */
    MSG_Y = PLN_Y + 8;
#endif /* MEM_PROF */

#ifndef __GUI__    /* Windows title screen in wntierra.c */
#if INTERFACE != BASICM
    sprintf(Fbuf, "\n\n");              /* Title screen */
    FEPrintf(0, 3, 0, 0, Fbuf);
#endif /* INTERFACE != BASICM */
    sprintf(Fbuf, "Tierra\n\n");
    FEPrintf(37, 5, 0, 0, Fbuf);
    sprintf(Fbuf, "March 96  VER %2.2f\n\n", VER);
    FEPrintf(31,  7, 0, 0, Fbuf);
    sprintf(Fbuf, "Information Wants To Be Alive!\n\n");
    FEPrintf(25,  9, 0, 0, Fbuf);
    sprintf(Fbuf, "By\n\n");
    FEPrintf(39, 11, 0, 0, Fbuf);
    sprintf(Fbuf, "Thomas S. Ray\n\n");
    FEPrintf(33, 13, 0, 0, Fbuf);
    sprintf(Fbuf, "Matt Jones, Roger Gouin, Jeremy Bruestle, Joseph Hart\n\n");
    FEPrintf(14, 15, 0, 0, Fbuf);
    sprintf(Fbuf, "Kurt Thearling, Tsukasa Kimezawa, Agnes Charrel\n\n");
    FEPrintf(16, 17, 0, 0, Fbuf);
    sprintf(Fbuf, "Daniel Pirone, Tom Uffner, Marc Cygnus\n\n");
    FEPrintf(20, 19, 0, -4, Fbuf);
    FEClear();
#endif /* __GUI__ */
    FEBanner();

#if FRONTEND != STDIO
    sprintf(Fbuf, "Tierra initializing, please wait ...\n");
    FEPrintf(HLP_X + 15, HLP_Y, 1, 0, Fbuf);
#endif /* FRONTEND != STDIO */

#ifdef MON
    M_MSG_X = MSG_X;
    M_MSG_Y = MSG_Y;
#endif /* MON */
}

void FEPrintf(scr_x, scr_y, scr_a, slee, str)
I32s scr_x, scr_y, scr_a, slee;
I8s * str;
{
#ifndef AMIGA
    I8s  CF = 0;
#else /* AMIGA */
#if FRONTEND == BASIC
    I8s  CF = 0;
#endif /* FRONTEND == BASIC */
#endif /* AMIGA */
    I32s  len = strlen(str);
#if INTERFACE == BASICM
    char                temp_str[500],meas_str[500], *token;
    char                *cr = "\n";
    GrafPtr             oldPort;
    short               x_pixel, y_pixel;
    Rect                destRect;
#endif /* INTERFACE == BASICM */

    if (str[0] == 0)
    return;

#if FRONTEND == STDIO

    if (scr_x < spos.x)
    {   fprintf(stdout, "\n");
        spos.x = 0;
    }
    if (scr_x > spos.x)
    {   while (spos.x < scr_x)
        {   fprintf(stdout, " ");
            spos.x++;
        }
    }
    fprintf(stdout, str);
#ifndef AMIGA
#ifndef DECVAX
    fflush(stdout);
#endif /* DECVAX */
#endif /* AMIGA */
    if (str[len - 1] == '\n')
    {   spos.x = 0;
    }
    else
        spos.x += len;

#endif /* FRONTEND == STDIO */

#if (INTERFACE == BASICB) || (INTERFACE == BASICD)
#ifndef __GUI__
/* the upper left character is 1,1; the lower right is 80,25 */

    scr_x++; scr_y++;
    if((scr_x < 1) || (scr_x + len > COLS1) || (scr_y < 1) || (scr_y > LINES1))
    {   sound(500);
        delay(500);
        nosound();
        return;
    }
#endif /* __GUI__ */
    if (str[len - 1] == '\n')
    {   str[len - 1] = '\0';
        CF = 1;
    }
#ifndef __GUI__
    if (scr_a)
    {   textbackground(InfoBack);
        textcolor(InfoText);
    }
    gotoxy(scr_x,scr_y);
    cprintf(str);
    if (CF)
        clreol();
    if (scr_a)
    {   textbackground(DispBack);
        textcolor(DispText);
    }
#else /* __GUI__ */
    if (scr_a)
        SetBkColor (hdcnew, RGB (255,255,0)) ;     /* yellow */
    else
        SetBkColor (hdcnew, RGB (255,255,255)) ;   /* white  */

    if (CF)
    {   (FEw + scr_y )->szMarg = 0L;
        strcpy(( FEw + scr_y )->szDesc, ENDBLANK ) ;
        FEwTextOut(FEw, scr_y ) ;
        (FEw + scr_y )->szMarg = 20L;
        FEwTextOut(FEw, scr_y ) ;
    }
    (FEw + scr_y )->szMarg = scr_x ;
    strcpy(( FEw + scr_y )->szDesc, str );
    FEwTextOut(FEw, scr_y ) ;

#endif /* __GUI__ */

#endif /* (INTERFACE == BASICB) || (INTERFACE == BASICD) */

#if INTERFACE == BASICC

/* the upper left character is 0,0; the lower right is COLS1,LINES1 */

    if((scr_x < 0) || (scr_x +len >= COLS1) || (scr_y < 0) || (scr_y >= LINES1))
    {   fprintf(stderr,"\0x07");
        fprintf(stderr,"\b");
        return;
    }
    if (str[len - 1] == '\n')
    {   str[len - 1] = '\0';
        CF = 1;
    }
    if (scr_a)
        standout();
    mvprintw(scr_y, scr_x, str);
    if (scr_a)
        standend();
    refresh();

#endif /* INTERFACE == BASICC */

#if INTERFACE == BASICM
    GetPort( &oldPort );
    SetPort( (GrafPtr)wp );
    /* strip out the carriage returns */
    if (str[len - 1] == '\n')
    {   spos.x = 0;
        CF = 1;
    }
    token = strtok( str, cr );
    strcpy( temp_str, token );
    while ( token != NULL )
    {   token = strtok( NULL, cr );
        if (token != NULL)
            strcat( temp_str, token );
    }
    strcpy( meas_str, temp_str );
    /* Erase the current block of text */
    x_pixel = scr_x*((windowWidth-displayOffset+5)/COLS1) + (displayOffset +5);
    y_pixel = (scr_y)*rowHeight + fontInfo.ascent;
    destRect.left = x_pixel;
    destRect.right = destRect.left + StringWidth( CtoPstr( meas_str ) );
    destRect.top = y_pixel-fontInfo.ascent;
    destRect.bottom = y_pixel+fontInfo.descent + fontInfo.leading;
    EraseRect( &destRect );
    /* draw the string to screen */
    MoveTo( x_pixel, y_pixel );
    DrawString( CtoPstr( temp_str ) );
    if (CF)
    {   destRect.left = destRect.right;
        destRect.right = windowWidth;
        EraseRect( &destRect );
    }
    SetPort( oldPort );
#endif /* INTERFACE == BASICM */

#if INTERFACE == BASICA

/* the upper left character is 0,0; the lower right is COLS1,LINES1 */

    if((scr_x < 0) || (scr_x +len >= COLS1) || (scr_y < 0) || (scr_y >= LINES1))
    {   fprintf(stderr,"\0x07");
        fprintf(stderr,"\b");
        return;
    }
    if (str[len - 1] == '\n')
    {   str[len - 1] = '\0';
        CF = 1;
    }
    if (scr_a)
        standout();
    mvprintw(scr_y, scr_x, str);
    if (scr_a)
        standend();
    refresh();

#endif /* INTERFACE == BASICA */

#if INTERFACE == BASICV

/* the upper left character is 0,0; the lower right is COLS1,LINES1 */

    if((scr_x < 0) || (scr_x +len >= COLS1) || (scr_y < 0) || (scr_y >= LINES1))
    {   fprintf(stderr,"\0x07");
        fprintf(stderr,"\b");
        return;
    }
    if (str[len - 1] == '\n')
    {   str[len - 1] = '\0';
        CF = 1;
    }
    if (scr_a)
        standout();
    move(scr_y, scr_x);
    printw(str);
    if (scr_a)
        standend();
    refresh();

#endif /* INTERFACE == BASICV */

#if FRONTEND != STDIO
#ifndef MON
    if (slee < 0)
        sleep(-slee);
    else if (slee &&
        ((IMode == SIZ_HIST) || (IMode == SIZM_HIST) || (IMode == GEN_HIST)))
    {   sleep(slee);
        query_species(fe_lines);
    }
#endif /* MON */
#endif
}

void FEGetStr(x, y)
I32s  x, y;
{
#if INTERFACE == BASICC
    nl();
    echo();
    move(y, x);
    refresh();
    getstr(FGSbuf);
    noecho();
    nonl();
#endif /* INTERFACE == BASICC */

#if INTERFACE == BASICM
    EventRecord event;
    char                character[2];
    GrafPtr             oldPort;
    short               x_pixel, y_pixel;
    Rect                destRect;

    GetPort( &oldPort );
    SetPort( (GrafPtr)wp );
    strcpy( FGSbuf, "" );
    character[1] = '\0';
    FEClrmsg( HLP_Y-3,"");
    x_pixel = HLP_X*((windowWidth-displayOffset+5)/COLS1) + displayOffset +5;
    y_pixel = (HLP_Y-3)*rowHeight + fontInfo.ascent;
    destRect.top = y_pixel-fontInfo.ascent;
    destRect.bottom = y_pixel+fontInfo.descent + fontInfo.leading;

    while (character[0] != '\n' && character[0] != '\r')
    {   if (GetNextEvent(everyEvent, &event))
        {       if (event.what == keyDown)
                {       character[0] = event.message & charCodeMask;
                        strcat( FGSbuf, (char *)character );
                        destRect.left = x_pixel;
                        destRect.right = destRect.left + CharWidth(
                        (short)character[0] );
                        EraseRect( &destRect );
                        MoveTo( x_pixel, y_pixel);
                        DrawChar( (short)character[0] );
                        x_pixel = destRect.right;
                }
        }
    }
    FEClrmsg( HLP_Y-3,"");
    return;
#endif /* INTERFACE == BASICM */

#if INTERFACE == BASICA
    nodelay(stdscr, FALSE);
    nl();
    echo();
    move(y, x);
    refresh();
    getstr(FGSbuf);
    noecho();
    nonl();
    nodelay(stdscr, TRUE);
#endif /* INTERFACE == BASICA */

#if INTERFACE == BASICV
    move(y, x);
    refresh();
    getstr(FGSbuf);
#endif /* INTERFACE == BASICV */

#if (INTERFACE == BASICB) || (INTERFACE == BASICD)
#ifndef __GUI__
    gotoxy(HLP_X + 1, HLP_Y - 2);
    gets(FGSbuf);
#else /* __GUI__ */
    int return_code;
    return_code = DialogBox (hInst,
        MAKEINTRESOURCE(GETSTRBOX), hwnd, lpfnGetStrDlgProc);
        /* gets text into FGSbuf thru GetStrDlgProc */
    if (return_code == -1)
    {   MessageBox(hwnd, "Unable to display dialog", "System Error",
            MB_ICONHAND);
    }
#endif /* __GUI__ */
#endif /* (INTERFACE == BASICB) || (INTERFACE == BASICD) */

#if FRONTEND == STDIO
    fgets(FGSbuf,84,stdin);
    if (FGSbuf[strlen(FGSbuf) - 1] == '\n')
        FGSbuf[strlen(FGSbuf) - 1]  = '\0';
#endif /* FRONTEND == STDIO */
}

void FEStats()
{
    if (GeneBnker)
    {   if(InstExe.m){
           sprintf(Fbuf,
           "InstExec  = %ld,%06ld  Cells = %4ld  Genotypes = %4ld  Sizes =%4ld\n",
           InstExe.m,InstExe.i,NumCells, NumGenotypes, NumSizes);
        }
        else{
           sprintf(Fbuf,
           "InstExec  = %6ld  Cells = %4ld  Genotypes = %4ld  Sizes =%4ld\n",
           InstExe.i,NumCells, NumGenotypes, NumSizes);
        }
        FEPrintf(0, 0, 1, 0, Fbuf);
        sprintf(Fbuf, "Extracted = %-20.20s  %s\n",
          ExtrG, (MC_step==(-1))?"normal":
          (MC_step== 1)?"step":"delay");
        FEPrintf(0, 1, 1, 0, Fbuf);
    }
    else /* (GeneBnker) */
    {
        if(InstExe.m){
           sprintf(Fbuf, "InstExec  = %ld,%06ld  Cells = %4ld\n",
           InstExe.m, InstExe.i, NumCells);
           FEPrintf(0, 0, 1, 0, Fbuf);
        }
        else{
           sprintf(Fbuf, "InstExec  = %6ld  Cells = %4ld\n",
           InstExe.i, NumCells);
           FEPrintf(0, 0, 1, 0, Fbuf);
        }
       sprintf(Fbuf, "%s\n",(MC_step==(-1))?"off":
         (MC_step== 1)?"keypress":"delay");
       FEPrintf(0, 1, 1, 0, Fbuf);
    }
#if FRONTEND == BASIC
#if (INTERFACE != BASICD) && (INTERFACE != BASICB)
    move(2, 0);
    refresh();
#endif /* (INTERFACE != BASICD) && (INTERFACE != BASICB) */
#endif /* FRONTEND == BASICC */
}

#ifdef __TURBOC__
int T_sig_int(void)
#elif __MWERKS__ || __SOLARIS2__
void T_sig_int(sig)
I32s  sig;
#else /* __TURBOC__ */
#ifdef DECVAX
void T_sig_int(int sig)
#else
void T_sig_int(sig, code, scp, addr)
I32s  sig, code;
I32s *scp; /* DO NOT USE !!!!! */ /* struct sigcontext *scp; */
I8s  *addr;
#endif /* DECVAX */
#endif /* __TURBOC__ */

{
#ifdef DECVAX
    signal(SIGINT, SIG_DFL);
    signal(SIGINT,  T_sig_int);
#endif /* DECVAX */
    TC_Menu = 1;
#ifdef __TURBOC__
    return TC_Menu;
#endif /* __TURBOC__ */
#ifdef MON
    if(MonConnectedFlg != 0) {
      sprintf(Fbuf, "Monitor is logging in. Frontend Menu isn't accepted.\n");
      FEPrintf(0, 1, 1, 0, Fbuf);
      TC_Menu = 0;
      return;
    }
#endif /* MON */
}

void FEPrVQ(data)
I8s  *data;
{   I8s  found = 1;

    if (vqu.type == 's')
        sprintf(Fbuf, "%s = %s", data, vqu.s);
    else if (vqu.type == 'h')
        sprintf(Fbuf, "%s = %hu", data, vqu.h);
    else if (vqu.type == 'i')
        sprintf(Fbuf, "%s = %ld", data, vqu.i);
    else if (vqu.type == 'f')
        sprintf(Fbuf, "%s = %f", data, vqu.f);
    else if (vqu.type == 'd')
        sprintf(Fbuf, "%s = %lf", data, vqu.d);
    else
        found = 0;
#if INTERFACE == BASICM
    FEClrmsg(MSG_Y+5, "");
#endif /* INTERFACE == BASICM */
    if (found)
        FEPrintf(MSG_X, MSG_Y + 5, 0, 0, Fbuf);
}

#ifndef __GUI__
void FEMenu ()
{   I8s  answer;
#ifndef AMIGA
    I32s tsz, psi, pgi;
#endif

#ifdef ALCOMM
    AL_run_flag = 1;
#endif

#if INTERFACE == BASICC
    signal(2,T_sig_int);
    touchwin(stdscr);
    refresh();
#endif /* INTERFACE == BASICC */

#if INTERFACE == BASICA
    touchwin(stdscr);
    refresh();
#endif /* INTERFACE == BASICA */

#if INTERFACE == BASICV
    /* signal(2,T_sig_int); */
    signal(SIGINT,  T_sig_int);
    touchwin(stdscr);
    refresh();
#endif /* INTERFACE == BASICV */

#if FRONTEND == STDIO
    sprintf(Fbuf,
    "\n==========================================================\n");
    FEPrintf(0, 1, 1, 0, Fbuf);
    FEPrintf(1, 1, 1, 0, "TIERRA: Main Menu");
    sprintf(Fbuf,
    "VER=%1.2f PLOIDY=%d  %s %s %s %s \n", VER, PLOIDY,

#ifdef ERROR
    "ERROR",
#else
    "",
#endif

#ifdef MEM_PROF
    "MEM_PROF",
#else
    "",
#endif

#ifdef MICRO
    "MICRO",
#else
    "",
#endif

#ifdef ALCOMM
    "ALCOMM"
#else
    ""
#endif
        );
    FEPrintf(0, 1, 1, 0, Fbuf);
    sprintf(Fbuf,
        "---------------------------------------------------------\n");
    FEPrintf(0, 1, 1, 0, Fbuf);
    sprintf(Fbuf, "InstExe.m    = %6ld  InstExec.i = %6ld  NumCells = %4ld ",
    InstExe.m,InstExe.i,NumCells);
    FEPrintf(1, 1, 1, 0, Fbuf);
    if(GeneBnker)
    {   sprintf(Fbuf,
        "NumGenotypes =   %4ld  NumSizes   =   %4ld", NumGenotypes, NumSizes);
        FEPrintf(1, 1, 1, 0, Fbuf);
    }
#ifdef __TURBOC__
    sprintf(Fbuf, "CoreLeft     =  %ld",(I32s) coreleft());
    FEPrintf(1, 1, 1, 0, Fbuf);
#endif
#ifdef ALCOMM
    sprintf(Fbuf, "VPORT        =    %hd", VPORT);
    FEPrintf(1, 1, 1, 0, Fbuf);
#endif

    sprintf(Fbuf,
    "---------------------------------------------------------\n");
    FEPrintf(0, 1, 1, 0, Fbuf);
    sprintf(Fbuf, "        Key      Function\n");
    FEPrintf(1, 1, 1, 0, Fbuf);
    sprintf(Fbuf, "        i        Information on simulation\n");
    FEPrintf(1, 1, 1, 0, Fbuf);
    sprintf(Fbuf, "        v        Change or examine a soup_in variable\n");
    FEPrintf(1, 1, 1, 0, Fbuf);
    sprintf(Fbuf, "        m        Misc. commands\n");
    FEPrintf(1, 1, 1, 0, Fbuf);
    sprintf(Fbuf, "        S        Execute a system Shell\n");
    FEPrintf(1, 1, 1, 0, Fbuf);
    sprintf(Fbuf, "        s        Save the soup\n");
    FEPrintf(1, 1, 1, 0, Fbuf);
    sprintf(Fbuf, "        q        Save the soup & quit\n");
    FEPrintf(1, 1, 1, 0, Fbuf);
    sprintf(Fbuf, "        Q        Quit/Abort simulation\n");
    FEPrintf(1, 1, 1, 0, Fbuf);
    sprintf(Fbuf, "        c        To Continue simulation\n");
    FEPrintf(1, 1, 1, 0, Fbuf);
    sprintf(Fbuf,
    "---------------------------------------------------------\n");
    FEPrintf(1, 1, 1, 0, Fbuf);
#endif /* FRONTEND == STDIO */

    while(1)
    {   FEVidInit(0);
        if (IMode==PLN_STATS) {
           FEPlan(0);
        }
        sprintf(Fbuf,
"i-info  v-var  s-save  S-shell  q-save&quit  Q-quit  m-misc  c-continue |->"
            );
        FEPrintf(HLP_X, HLP_Y, 1, 0, Fbuf);
#ifdef AMIGA
#if FRONTEND == BASIC
      nodelay(stdscr, FALSE);
#endif /* FRONTEND == BASIC */
#endif /* AMIGA */
        answer = FEGetch(HLP_X + strlen(&(Fbuf[0])), HLP_Y);
#ifdef AMIGA
#if FRONTEND == BASIC
      nodelay(stdscr, TRUE);
#endif /* FRONTEND == BASIC */
#endif /* AMIGA */
        if (answer == 'c')
             break;
        else if (answer == 's')
            GoSave();
        else if (answer == 'S')
            GoShell();
        else if (answer == 'q')
            GoSaveQuit();
        else if (answer == 'Q')
            GoQuit();
        else if (answer == 'v')
            GoVariable();
        else if (answer == 'i')
            GoInfo();
        else if (answer == 'm')
            GoMisc();
        FEBanner();
        FEStats();
        answer = ' ';
    }   /* end while loop */
    if ((FRONTEND == STDIO) ||
        (IMode != SIZ_HIST) && (IMode != SIZM_HIST) && (IMode != GEN_HIST))
        FEClrmsg(MSG_Y,
            ((OPSYS == BCDOS) || (OPSYS == DJGPPDOS) || (OPSYS == AMIGADOS)
            || (OPSYS == MAC) ?
"                       Press Any Key for Menu ...                           "
            :
"                    Press Interrupt Key for Menu ...                        "
            ));
    TC_Menu = 0;
}    /* end of FEMenu */
#endif /* __GUI__ */

void GoSave()
{   I8s  answer;

    FEBanner();
    FEStats();
    sprintf(Fbuf, "Are you sure ? (y/n)->");
    FEPrintf(HLP_X, HLP_Y, 1, 0, Fbuf);
#ifdef AMIGA
#if FRONTEND == BASIC
    nodelay(stdscr, FALSE);
#endif /* FRONTEND == BASIC */
#endif /* AMIGA */
    answer = FEGetch(HLP_X + strlen(&(Fbuf[0])), HLP_Y);
#ifdef AMIGA
#if FRONTEND == BASIC
    nodelay(stdscr, TRUE);
#endif /* FRONTEND == BASIC */
#endif /* AMIGA */
    if ((answer=='y')||(answer=='Y')){
        FEPrintf(MSG_X, MSG_Y, 0, 0, "Writing soup ...");
        WriteSoup(0);
        FEPrintf(MSG_X, MSG_Y + 2, 0, 0, "Done saving soup...");
    }
}

#ifndef __GUI__
void GoShell()
{   sprintf(Fbuf,
        "TIERRA: shell (%s), type exit to return ... ", SHELL);
    FEPrintf(HLP_X, HLP_Y, 1, 0, Fbuf);
#if INTERFACE == BASICC
    echo(); nl(); nocbreak();
    clear(); refresh(); endwin();
#endif /* INTERFACE == BASICC */

#if INTERFACE == BASICA
    echo(); nl(); nocbreak();
    clear(); refresh(); endwin();
#endif /* INTERFACE == BASICA */

#if INTERFACE == BASICV
    echo(); nl();
    clear(); refresh(); endwin();
#endif /* INTERFACE == BASICV */

    system(SHELL);
    FEVidInit(0);
    FEPlan(0);
}
#endif /* __GUI__ */

void GoSaveQuit()
{   I8s  answer;

    FEBanner();
    FEStats();
    sprintf(Fbuf, "Are you sure ? (y/n)->");
    FEPrintf(HLP_X, HLP_Y, 1, 0, Fbuf);
#ifdef AMIGA
#if FRONTEND == BASIC
    nodelay(stdscr, FALSE);
#endif /* FRONTEND == BASIC */
#endif /* AMIGA */
    answer = FEGetch(HLP_X + strlen(&(Fbuf[0])), HLP_Y);
#ifdef AMIGA
#if FRONTEND == BASIC
    nodelay(stdscr, TRUE);
#endif /* FRONTEND == BASIC */
#endif /* AMIGA */
    if ((answer=='y')||(answer=='Y'))
        FEError(0,EXIT,WRITE,"");
}

#ifndef __GUI__
void GoQuit()
{   I8s  answer;

    FEBanner();
    FEStats();
    sprintf(Fbuf, "Are you sure ? (y/n)->");
    FEPrintf(HLP_X, HLP_Y, 1, 0, Fbuf);
#ifdef AMIGA
#if FRONTEND == BASIC
    nodelay(stdscr, FALSE);
#endif /* FRONTEND == BASIC */
#endif /* AMIGA */
    answer = FEGetch(HLP_X + strlen(&(Fbuf[0])), HLP_Y);
#ifdef AMIGA
#if FRONTEND == BASIC
    nodelay(stdscr, TRUE);
#endif /* FRONTEND == BASIC */
#endif /* AMIGA */
    if ((answer=='y')||(answer=='Y'))
        FEExit(0);
}

void GoVariable()
{   I8s  answer;

    FEClrmsg(MSG_Y, "");
    FEBanner();
    FEStats();
    sprintf(&(Fbuf[0]),"VAR |  a-alter variable  e-examine variable->");
    FEPrintf(HLP_X, HLP_Y, 1, 0, Fbuf);
#ifdef AMIGA
#if FRONTEND == BASIC
      nodelay(stdscr, FALSE);
#endif /* FRONTEND == BASIC */
#endif /* AMIGA */
    answer = FEGetch(HLP_X + strlen(&(Fbuf[0])), HLP_Y);
#ifdef AMIGA
#if FRONTEND == BASIC
      nodelay(stdscr, TRUE);
#endif /* FRONTEND == BASIC */
#endif /* AMIGA */
    if (answer == 'a')
        GoAlter();
    else if (answer == 'e')
        GoExamine();
    answer = ' ';
}
#endif /* __GUI__ */

void GoAlter()
{   I8s  answer;
    sprintf(Fbuf, "To alter any global variable from soup_in, type\n");
    FEPrintf(MSG_X, MSG_Y, 0, 0, Fbuf);
    sprintf(Fbuf, "the variable name (using proper case), a space,\n");
    FEPrintf(MSG_X, MSG_Y + 1, 0, 0, Fbuf);
    sprintf(Fbuf, "an equal sign, a space, and the new value.\n");
    FEPrintf(MSG_X, MSG_Y + 2, 0, 0, Fbuf);
    sprintf(Fbuf, "Use no space at start of line.  Some examples:\n");
    FEPrintf(MSG_X, MSG_Y + 3, 0, 0, Fbuf);
    sprintf(Fbuf, "alive = 0\n");
    FEPrintf(MSG_X, MSG_Y + 4, 0, 0, Fbuf);
    sprintf(Fbuf, "DistProp = .6\n");
    FEPrintf(MSG_X, MSG_Y + 5, 0, 0, Fbuf);
    sprintf(Fbuf, "GenebankPath = newpath/\n");
    FEPrintf(MSG_X, MSG_Y + 6, 0, 0, Fbuf);
    FEGetStr(HLP_X, HLP_Y - 3);
    TGeneBnker = GeneBnker;
    if (!GetAVar(FGSbuf))
    {   sprintf(Fbuf,
            "Tierra T_sig_int() Not a valid soup_in variable: %s",
            FGSbuf);
        FEError(-500, NOEXIT, NOWRITE, Fbuf);
    }
    else{
        if(TGeneBnker != GeneBnker){
            ChangeGeneBnker(GeneBnker);
            if(!GeneBnker){
               GoBrkClr(1);
            }
        }
    }
    TGeneBnker = GeneBnker;
    ToggleLog();
    sprintf(Fbuf,"FE: Var change @ %ld,%5.5ld - %.30s",
        InstExe.m, InstExe.i, FGSbuf);
#if INTERFACE == BASICM
    FEClrmsg(MSG_Y+8, "");
#endif /* INTERFACE == BASICM */
    FEPrintf(MSG_X, MSG_Y + 8, 0, 0, Fbuf);
    FEBanner();
    FEStats();
    sprintf(&(Fbuf[0]),"hit any key to continue");
    FEPrintf(HLP_X, HLP_Y, 1, 0, Fbuf);
#ifdef AMIGA
#if FRONTEND == BASIC
    nodelay(stdscr, FALSE);
#endif /* FRONTEND == BASIC */
#endif /* AMIGA */
    answer = FEGetch(HLP_X + strlen(&(Fbuf[0])), HLP_Y);
#ifdef AMIGA
#if FRONTEND == BASIC
    nodelay(stdscr, TRUE);
#endif /* FRONTEND == BASIC */
#endif /* AMIGA */
}

void GoExamine()
{   I8s  answer;
    sprintf(Fbuf, "To examine any global variable from soup_in,\n");
    FEPrintf(MSG_X, MSG_Y, 0, 0, Fbuf);
    sprintf(Fbuf, "the name of the variable:\n");
    FEPrintf(MSG_X, MSG_Y + 1, 0, 0, Fbuf);
    FEGetStr(HLP_X, HLP_Y - 3);
    if (!GetAVar(FGSbuf))
    {   sprintf(Fbuf,
            "Tierra T_sig_int() Not a valid soup_in variable: %s",
            FGSbuf);
        FEError(-501, NOEXIT, NOWRITE, Fbuf);
    }
    FEPrVQ(FGSbuf);
    FEBanner();
    FEStats();
    sprintf(&(Fbuf[0]),"hit any key to continue");
    FEPrintf(HLP_X, HLP_Y, 1, 0, Fbuf);
#ifdef AMIGA
#if FRONTEND == BASIC
    nodelay(stdscr, FALSE);
#endif /* FRONTEND == BASIC */
#endif /* AMIGA */
    answer = FEGetch(HLP_X + strlen(&(Fbuf[0])), HLP_Y);
#ifdef AMIGA
#if FRONTEND == BASIC
    nodelay(stdscr, TRUE);
#endif /* FRONTEND == BASIC */
#endif /* AMIGA */
}

void GoInfo()
{   I8s  answer;
#if FRONTEND == BASIC
    I8s tt, dummy ;
#endif /* FRONTEND == BASIC */

    FEBanner();
    FEStats();
#if FRONTEND == BASIC
#if (INTERFACE != BASICD) && (INTERFACE != BASICB)
    standout();
    for(tt = 0; tt < (COLS1 - 1) ; tt++)
    {   dummy = tt ;
        mvaddch(1,dummy,' ');
    }
    standend();
#endif /* (INTERFACE != BASICD) && (INTERFACE != BASICB) */
#endif /* FRONTEND == BASIC */

#ifdef MEM_PROF
    sprintf(Buf85, "TotNumAloc: %8ld DynMemUse: %8ld seed: %ld",
        MemStat.TotNumAloc, MemStat.DynMemUse, seed);
#else  /* MEM_PROF */
    sprintf(Buf85, "seed: %ld", seed);
#endif /* MEM_PROF */

    sprintf(Fbuf,
#ifdef ALCOMM
        "Port = %5ld %s   ", VPORT, Buf85);
#else /* ALCOMM */
        "%s", Buf85);
#endif /* ALCOMM */
#ifndef __GUI__
    FEPrintf(0, 1, 1, 0, Fbuf);
    sprintf(Fbuf,
      "INFO |  p-plan  s-size_histo g-gen_histo m-mem_histo z-size_query  |->");
    FEPrintf(HLP_X, HLP_Y, 1, 0,&(Fbuf[0]));
#ifdef AMIGA
#if FRONTEND == BASIC
      nodelay(stdscr, FALSE);
#endif /* FRONTEND == BASIC */
#endif /* AMIGA */
    answer = FEGetch(HLP_X + strlen(&(Fbuf[0])), HLP_Y);
#ifdef AMIGA
#if FRONTEND == BASIC
      nodelay(stdscr, TRUE);
#endif /* FRONTEND == BASIC */
#endif /* AMIGA */
    switch(answer){
       case 'p':
          histostrtidx=0;
          GoPlan();
          break;
       case 's':
          histostrtidx=0;
          GoSizeHist();
          break;
       case 'm':
          histostrtidx=0;
          GoMem();
          break;
       case 'g':
          histostrtidx=0;
          GoGene();
          break;
       case 'z':
          histostrtidx=0;
          GoSizeQuery();
          break;
       default:
          break;
    }
    answer = ' ';
#else /* __GUI__ */
    MessageBox(hwnd, Fbuf,"Tierra Info", MB_OK);
#endif /* __GUI__ */
}

void GoPlan()
{   if (Hist)
    {   thfree((void*) Hist, HistAlocSiz, 2101);
        Hist = (HistType *) NULL;
        HistAlocSiz = 0;
    }
    if (TGeneBnker != GeneBnker)
    {   GeneBnker = TGeneBnker;
        ChangeGeneBnker(GeneBnker);
    }

    IMode = PLN_STATS;
}
void GoSizeQuery()
{
   int abortflag, num_spec;
   I32s size_class;
   size_buf *buf;

   /* enter size class */
   size_class=EnterLongVal(MSG_Y,
      "Enter a size class ( eg: 80 ) to examine -> ",
      &abortflag);
   if(!abortflag){
      if(!GeneBnker){
         TGeneBnker=GeneBnker;
         GeneBnker=1;
         ChangeGeneBnker(GeneBnker);
      }
      if((size_class>0)&&(size_class<siz_sl)){
         if(sl[size_class]){
            num_spec = sl[size_class]->num_g;
            /* initialize size query buffer */
            if(buf=InitSizeQuery(size_class, num_spec)){

               /* display size query buffer */
               DispSizeQuery(buf, num_spec, size_class);

               thfree((void*)buf, num_spec*sizeof(size_buf), 2100);
            }
         }
      }
   }
   /* query termination screen handling */
#if FRONTEND == BASIC
   FEClrmsg(MSG_Y, "");
   FEBanner();
   FEStats();
#endif /* FRONTEND == BASIC */
}

void GoSizeHist()
{   IMode = SIZ_HIST;
    if (!GeneBnker)
    {   TGeneBnker = GeneBnker;
        GeneBnker = 1;
        ChangeGeneBnker(GeneBnker);
    }
}

void GoMem()
{   IMode = SIZM_HIST;
    if (!GeneBnker)
    {   TGeneBnker = GeneBnker;
        GeneBnker = 1;
        ChangeGeneBnker(GeneBnker);
    }
}

void GoGene()
{   IMode = GEN_HIST;
    if (!GeneBnker)
    {   TGeneBnker = GeneBnker;
        GeneBnker = 1;
        ChangeGeneBnker(GeneBnker);
    }
}

void GoMisc()
{   I8s  answer;

    FEClrmsg(MSG_Y, "");
    FEBanner();
    FEStats();
    sprintf(Fbuf, "VER=%1.2f PLOIDY=%d  %s %s %s %s\n", VER,PLOIDY,
#ifdef ERROR
        "ERROR",
#else /* ERROR */
        "",
#endif /* ERROR */
#ifdef MEM_PROF
        "MEM_PROF",
#else /* MEM_PROF */
        "",
#endif /* MEM_PROF */
#ifdef MICRO
        "MICRO",
#else /* MICRO */
        "",
#endif /* MICRO */
#ifdef ALCOMM
        "ALCOMM"
#else /* ALCOMM */
        ""
#endif /* ALCOMM */
        );

#ifdef __GUI__

    MessageBox(hwnd, Fbuf, "Tierra Simulation Status", MB_OK);

#else /*__GUI__ */

    FEPrintf(0, 1, 1, 0, Fbuf);
    sprintf(Fbuf,
        "H-Histo Log  I-Inj Gene  %s  %s  %s  %s->",

#ifdef MICRO
        "M-Micro Tog",
#else /* MICRO */
        "",
#endif /* MICRO */

#ifdef ALCOMM
        "P-Alm Pause",
#else /* ALCOMM */
        "",
#endif /* ALCOMM */

#ifdef MICRO
        "B-Brk Menu",
#else /* MICRO */
        "",
#endif /* MICRO */
#ifndef MON
        "O-Histo Ctrl"
#else
        ""
#endif /* MON */
        );
    FEPrintf(HLP_X, HLP_Y, 1, 0, Fbuf);

#ifdef AMIGA
#if FRONTEND == BASIC
      nodelay(stdscr, FALSE);
#endif /* FRONTEND == BASIC */
#endif /* AMIGA */

    answer = FEGetch(HLP_X + strlen(&(Fbuf[0])), HLP_Y);

#ifdef AMIGA
#if FRONTEND == BASIC
      nodelay(stdscr, TRUE);
#endif /* FRONTEND == BASIC */
#endif /* AMIGA */

    switch(answer){
       case 'H':                  /* Histo Logging */
           GoHistoLog();
           break;
       case 'I':                  /* Inject Gene */
           GoInject();
           break;
       case 'M':                  /* Micro Toggle */
           GoMicroToggle();
           break;
#ifdef ALCOMM
       case 'P':                  /* Almond Pause */
           GoALComm();
           break;
#endif /* ALCOMM */
#ifdef MICRO
       case 'B':                  /* Break Menu */
           GoBrkMenu();
           break;
#endif /* MICRO */
#ifndef MON
       case 'O':                  /* Display Control */
           GoHistoCtrl();
           break;
#endif /* MON */
       default:
           break;
    }
    FEBanner();
    FEStats();
    answer = ' ';

#endif /*__GUI__ */
}

void GoHistoLog()
{   HistPrint = (Log && !HistPrint) ? 1 : 0;
    sprintf(Fbuf,
        "%s \n", (Log) ? (HistPrint) ? "Logging Histograms \n" : 
        "NOT Logging Histograms \n" : "Log NOT on ! \n");
#ifdef __GUI__
    MessageBox(hwnd, Fbuf, "Tierra Simulation Status", MB_OK);
#else /* __GUI */
    FEPrintf(MSG_X, MSG_Y, 0, 0, Fbuf);
#endif /*__GUI__ */
}

#ifdef ALCOMM
void GoALComm()
{   AL_run_flag = (AL_run_flag) ? 0 : 1; 
    sprintf(Fbuf," ALmond Pause = %s\n",
        (AL_run_flag) ? "off" : "on");
    FEPrintf(MSG_X, MSG_Y, 0, 0, Fbuf);
}
#endif /* ALCOMM */

void GoInject()
{
   I32s injectnum, prefinjaddr, actnuminject;
   int abortflag, loopflag;
   GlIndex injgen;
   char injgennam[8];
   GList *g;

#if FRONTEND == BASIC
      FEClrmsg(MSG_Y, "");
      FEBanner();
      FEStats();
#endif /* FRONTEND == BASIC */

   abortflag=0;

   /* enter genome to be injected */
   loopflag=1;
   while(loopflag){
      injgen=EnterGenome(MSG_Y,
         "Enter genome to be injected (eg 0080aaa) (enter to abort)->",
         &abortflag, 0);
      if(!abortflag){
         sprintf(&(injgennam[0]), "%04ld%-3s",
            injgen.si, Int2Lbl(injgen.gi));
         if(g=GetAGen(&(injgennam[0]))){
            FreeGen(g,0);
            loopflag=0;
         }
      }
      else{
         loopflag=0;
      }
   }
   if(!abortflag){
      loopflag=1;
      while(loopflag){

         /* enter number of creatures to be injected */
         injectnum=EnterLongVal(MSG_Y+1,
            "Enter number of creatures to inject (enter to abort)->",
            &abortflag);
         if(!abortflag){
            if(injectnum<1){
               sprintf(Fbuf, "invalid entry");
               FEError(-502, NOEXIT, NOWRITE, Fbuf);
            }
            else{
               loopflag=0;
            }
         }
         else{
            loopflag=0;
         }
      }
      if(!abortflag){

         /* display soup limits */
            sprintf(Fbuf,"soup address limits = 0 - %ld", SoupSize-1);
            FEPrintf(MSG_X, MSG_Y+2, 0, 0, Fbuf);

         loopflag=1;
         while(loopflag){
            /* enter preferred injection address */
            prefinjaddr=EnterLongVal(MSG_Y+3,
               "Enter preferred address, (Enter to abort, or -1 for no preference)->",
               &abortflag);
            if(!abortflag){
               if((prefinjaddr>=0)&&(prefinjaddr<SoupSize)){
                  loopflag=0;
               }
               else{
                  if(prefinjaddr==(-1)){
                     loopflag=0;
                  }
                  else{
                     sprintf(Fbuf, "invalid soup address");
                     FEError(-503, NOEXIT, NOWRITE, Fbuf);
                  }
               }
            }
            else{
               loopflag=0;
            }
         }
         if(!abortflag){
            actnuminject=InjectFromBank(
               &(injgennam[0]), prefinjaddr, 0, injectnum);
            if(actnuminject!=injectnum){
               sprintf(Fbuf, "only %ld actually injected", actnuminject);
               FEError(-505, NOEXIT, NOWRITE, Fbuf);
            }
            if(NumCells){
               if (!(NumCells-actnuminject)) {
                  CalcAverages();
                  AdjCtrlVars();
               }
            }
         }
      }
   }
#if FRONTEND == BASIC
      FEClrmsg(MSG_Y, "");
      FEBanner();
      FEStats();
#endif /* FRONTEND == BASIC */
}

void FEClrmsg(n, buf)
I32s n;
I8s  *buf;
{
#ifndef AMIGA
    I8s t;
#endif
#if INTERFACE == BASICM
    GrafPtr             oldPort;
    Rect                destRect;
#endif /* INTERFACE == BASICM */

    if(n < 0)
    return;

#if (INTERFACE == BASICB) || (INTERFACE == BASICD)
    for(t = n + 1;  t <= fe_lines; t++)
    {
#ifndef __GUI__
        gotoxy(1, t);
        clreol();
#else /* _GUI__ */
        FEPrintf (20, t-1, 0, 0, ENDBLANK); FEPrintf (0, t-1, 0, 0, ENDBLANK);
#endif /* __GUI__ */
    }
    FEPrintf(HLP_X, HLP_Y, 1, 0, buf);
#endif /* (INTERFACE == BASICB) || (INTERFACE == BASICD) */

#if INTERFACE == BASICC
    move(n, 0);
    clrtobot();
    FEPrintf(HLP_X, HLP_Y, 1, 0, buf);
    refresh();
#endif /* INTERFACE == BASICC */

#if INTERFACE == BASICM
    GetPort( &oldPort );
    SetPort( (GrafPtr)wp );
    destRect.left = displayOffset+1;
    destRect.right = windowWidth;
    destRect.top = (n)*rowHeight;
    destRect.bottom = (n+1)*rowHeight;
    EraseRect( &destRect );
    SetPort( oldPort );
    FEPrintf(HLP_X, HLP_Y, 1, 0, buf);
#endif /* INTERFACE == BASICM */

#if INTERFACE == BASICA
    move(n, 0);
    clrtobot();
    FEPrintf(HLP_X, HLP_Y, 1, 0, buf);
    refresh();
#endif /* INTERFACE == BASICA */

#if INTERFACE == BASICV
    move(n, 0);
    clrtobot();
    FEPrintf(HLP_X, HLP_Y, 1, 0, buf);
    refresh();
#endif /* INTERFACE == BASICV */

#if FRONTEND == STDIO
    FEPrintf(HLP_X, HLP_Y, 1, 0, buf);
#endif /* FRONTEND == STDIO */
}

#if (OPSYS == DECVMS) || (OPSYS == MAC)
void FEVidInit(I16s mode)
#else
void FEVidInit(mode)
I16s mode;
#endif /* OPSYS */
{
#ifndef AMIGA
    int tt;
#endif

#if INTERFACE == BASICC
    echo(); nl(); nocbreak();
    initscr();
    LINES1=LINES;
    COLS1=COLS;
    cbreak(); nonl(); noecho();
    FEClear();
#endif /* INTERFACE == BASICC */

#if INTERFACE == BASICA
    echo(); nl(); nocbreak();
    initscr();
    LINES1=LINES;
    COLS1=COLS;
    cbreak(); nonl(); noecho();
    FEClear();
#endif /* INTERFACE == BASICA */

#if INTERFACE == BASICV
    initscr();
    LINES1=LINES;
    COLS1=COLS;
    FEClear();
#endif /* INTERFACE == BASICV */

#if INTERFACE == BASICB
#if (__TURBOC__ >= 0x200) && (__TURBOC__ < 0x400)
    /* deal with old and (stupid) new compilers */
    if (mode == EGA)
    {   textmode(C4350); /* use C4350 for ega/vga */
        LINES1 = 43;
    }
    else if (mode == VGA)
    {   textmode(C4350);
        LINES1 = 50;
    }
    else
#endif /* __TURBOC__ >= 0x200 */
    {
#ifndef __GUI__
        textmode(C80);
#endif /* __GUI__ */
        LINES1 = 25;
    }
    ERR_Y = LINES1 - 2;
    HLP_Y = fe_lines = LINES1 - 1;
#endif /* INTERFACE == BASICB */

#if INTERFACE == BASICD
    textmode(C80);
    LINES1 = 25;
    ERR_Y = LINES1 - 2;
    HLP_Y = fe_lines = LINES1 - 1;
#endif /* INTERFACE == BASICD */

#if FRONTEND != STDIO
    if (IMode == PLN_STATS)
#ifdef MEM_PROF
        MSG_Y = PLN_Y + 9;
#else  /* MEM_PROF */
        MSG_Y = PLN_Y + 8;
#endif /* MEM_PROF */
    else
        MSG_Y =  PLN_Y ;
    FEClrmsg(MSG_Y, "");
    FEBanner();
    FEStats();
#endif /* FRONTEND != STDIO */

#ifndef MON
    if ((IMode == SIZ_HIST) || (IMode == SIZM_HIST) || (IMode == GEN_HIST))
        query_species(fe_lines);
#endif /* MON */
}

#if (OPSYS == BCDOS) || (OPSYS == DJGPPDOS)
I8s KeyHit()
{
#ifndef __GUI__
    if (kbhit())
    {   getch();
        return 1;
    }
#else /* __GUI__ */
    if ( GetKeyState(VK_SPACE) < 0 ) /* input from Windows message buffer */
        return 1;
#endif /* __GUI__ */
    return 0;
}
#endif /* (OPSYS == BCDOS) || (OPSYS == DJGPPDOS) */

void FEError(errno1, exit, write, buff)
I32s errno1, exit, write;
I8s * buff;
{   I8s  ch;

    if (buff[0] != 0)
    {
        FEPrintf(ERR_X, ERR_Y, 1, 0, buff);

#if FRONTEND == STDIO
        sprintf(buff, "\n");
        FEPrintf(ERR_X, ERR_Y, 0, 0, buff);
#endif /* FRONTEND == STDIO */
        if(Log)
            fprintf(tfp_log, "\n");
    }

#if FRONTEND != STDIO
#ifndef __GUI__
#ifndef MON
    if(strlen(buff) > 1 )
    {   FEPrintf(HLP_X, HLP_Y, 1, 0,
"                     Tierra ERROR Press any key to continue               ");

#ifdef AMIGA
#if FRONTEND == BASIC
      nodelay(stdscr, FALSE);
#endif /* FRONTEND == BASIC */
#endif /* AMIGA */

#ifdef DECVAX
        ch = FEGetch(HLP_X + 74, HLP_Y);
#else
        ch = FEGetch(-1,-1);
#endif

#ifdef AMIGA
#if FRONTEND == BASIC
      nodelay(stdscr, TRUE);
#endif /* FRONTEND == BASIC */
#endif /* AMIGA */

#if FRONTEND == BASIC
        FEClrmsg(ERR_Y, "");
        FEBanner();
        FEStats();
#endif /* FRONTEND == BASIC */
    }
#endif /* MON */
#endif /* __GUI__ */
#endif /* FRONTEND != STDIO */

    if (write == WRITE)
    {   sprintf(Fbuf, "Writing soup ...");
        FEPrintf(MSG_X, MSG_Y, 0, 0, Fbuf);
        WriteSoup(exit);
    }
    if (exit == EXIT)
    {
#ifdef NET
        Bye_send();
#endif /* NET */
        if (hangup)
            FEMenu();
        sprintf(Fbuf, "Tierra going down ... ");
        FEPrintf(MSG_X, MSG_Y, 0, 0, Fbuf);
        if(Log)
        {   fclose(tfp_log);
            Log = 0;
        }
        sprintf(Fbuf,
#ifdef __GUI__
            "Tierra exiting");
        MessageBox(hwnd, Fbuf,"Tierra Error", MB_ICONEXCLAMATION | MB_OK);
#else /* __GUI__ */
"                      Tierra exiting  (PRESS ANY KEY)                   \n");
        FEPrintf(HLP_X, HLP_Y, 1, 0, Fbuf);

#ifdef AMIGA
#if FRONTEND == BASIC
      nodelay(stdscr, FALSE);
#endif /* FRONTEND == BASIC */
#endif /* AMIGA */

#ifdef DECVAX
        ch = FEGetch(HLP_X + 72, HLP_Y);
#else
        ch = FEGetch(-1,-1);
#endif

#ifdef AMIGA
#if FRONTEND == BASIC
      nodelay(stdscr, TRUE);
#endif /* FRONTEND == BASIC */
#endif /* AMIGA */

        if (ch == '~')
            FEMenu(); /* so that we can examin params */
#endif /* __GUI__ */
        FEExit(errno1);
    }
    if(Log){
#ifndef AMIGA
#ifndef DECVAX
         fflush(tfp_log);
#endif /* DECVAX */
#endif /* AMIGA */
    }
}

void  FEPlan(log)
I32s  log;
{   time_t  tp;
    tp = time(NULL);

#if FRONTEND != STDIO
    if ((IMode == SIZ_HIST) || (IMode == SIZM_HIST)||(IMode == GEN_HIST)) {
#ifndef MON
        query_species (fe_lines);
#endif /* MON */
    }
    else
#endif /* FRONTEND != STDIO */
    {   sprintf(Fbuf, "InstExeC      = %8ld  Generations  = %8.0f  %s",
            InstExe.m, Generations, ctime(&tp));
        FEPrintf(PLN_X, PLN_Y, 0, 0, Fbuf);
        if (GeneBnker)
        {   sprintf(Fbuf, 
              "    NumCells  = %8ld  NumGenotypes = %8ld  NumSizes  = %8ld\n",
                NumCells, NumGenotypes, NumSizes);
            FEPrintf(PLN_X, PLN_Y + 1, 0, 0, Fbuf);
            sprintf(Fbuf, 
               "    AvgSize   = %8ld  NumGenDG     = %8ld\n", 
                AverageSize, NumGenDG);
            FEPrintf(PLN_X, PLN_Y + 2, 0, 0, Fbuf);
        }
        else /* (GeneBnker) */
        {   sprintf(Fbuf, "    NumCells  = %8ld\n", NumCells);
            FEPrintf(PLN_X, PLN_Y + 1, 0, 0, Fbuf);
            sprintf(Fbuf, "    AvgSize   = %8ld\n", AverageSize);
            FEPrintf(PLN_X, PLN_Y + 2, 0, 0, Fbuf);
        }
        sprintf(Fbuf,
            "    AvgPop    = %8.0lf  Births       = %8ld  Deaths    = %8ld\n",
            AvgPop,TimeBirth, TimeDeath);
        if (InstExe.m)
        {   FEPrintf(PLN_X, PLN_Y + 3, 0, 0, Fbuf);
#ifdef NET
            sprintf(Fbuf, "    Speed     = %8ld  NumNodes     = %8hu\n",
                Speed, NumNodes);
#else /* NET */
            sprintf(Fbuf, "    Speed     = %8ld\n", Speed);
#endif /* NET */
            FEPrintf(PLN_X, PLN_Y + 6, 0, 0, Fbuf);
            if (GeneBnker)
            {
#if INTERFACE != BASICM
                sprintf(Fbuf,
                    "    MaxGenPop = %8ld  (%4.4ld%3.3s)",
                    MaxPop,MaxGenPop.size, MaxGenPop.label);
                FEPrintf(PLN_X, PLN_Y + 5, 0, 0, Fbuf);
                sprintf(Fbuf,
                    "    MaxGenMem = %8ld (%4.4ld%3.3s)\n",
                    MaxMemry / MaxGenMem.size, MaxGenMem.size,
                    MaxGenMem.label);
                FEPrintf(PLN_X + 37, PLN_Y + 5, 0, 0, Fbuf);
#else /* INTERFACE != BASICM */
                sprintf(Fbuf,
  "    MaxGenPop = %8ld  (%4.4ld%3.3s)    MaxGenMem = %8ld  (%4.4ld%3.3s)\n",
                    MaxPop,MaxGenPop.size, MaxGenPop.label,
                    MaxMemry / MaxGenMem.size, MaxGenMem.size,
                    MaxGenMem.label);
                FEPrintf(PLN_X, PLN_Y + 5, 0, 0, Fbuf);
#endif /* INTERFACE != BASICM */
            } /* end: if (GeneBnker) */
        } /* end: if (InstExe.m) */
        sprintf(Fbuf,
            "    RateMut   = %8ld  RateMovMut   = %8ld  RateFlaw  = %8ld\n",
            RateMut, RateMovMut, RateFlaw);
        FEPrintf(PLN_X, PLN_Y + ((InstExe.m) ? 4 : 3), 0, 0, Fbuf);
#ifdef MEM_PROF
#if (defined __TURBOC__) && (!defined __GUI__)
        sprintf(Fbuf,
            "    DynMemUse = %8ld  TotNumAloc   = %8ld  Coreleft  = %8lu\n",
            MemStat.DynMemUse, MemStat.TotNumAloc, (I32u) coreleft());
        FEPrintf(PLN_X, PLN_Y + 7, 0, 0, Fbuf);
#endif /* (defined __TURBOC__) && (!defined __GUI__) */
#if (!defined __TURBOC__) || (defined __GUI__)
        sprintf(Fbuf, "    DynMemUse = %8ld  TotNumAloc   = %8ld\n",
            MemStat.DynMemUse, MemStat.TotNumAloc);
        FEPrintf(PLN_X, PLN_Y + 7, 0, 0, Fbuf);
#endif /* (!defined __TURBOC__) || (defined __GUI__) */
#endif  /* MEM_PROF */
    } /* end: if ((IMode == SIZ_HIST) || (IMode == SIZM_HIST) ||
          (IMode == GEN_HIST)) */
    if (Log && log)
    {   fprintf(tfp_log,"\nie%ld gn%.0f %s",
        InstExe.m, Generations, ctime(&tp));
        fprintf(tfp_log, "nc%ld ng%ld ns%ld\n",
        NumCells, NumGenotypes, NumSizes);
        fprintf(tfp_log,"as%ld dg%ld\n",
        AverageSize, NumGenDG);
        if (InstExe.m)
        {   fprintf(tfp_log,"bi%ld de%ld ap%.0f\n",
            TimeBirth, TimeDeath, AvgPop);
            if (GeneBnker)
            {   fprintf(tfp_log,"mp%ld @ %4.4ld%3.3s",
                    MaxPop,MaxGenPop.size, MaxGenPop.label);
                fprintf(tfp_log," mg%ld @ %4.4ld%3.3s\n",
                    MaxMemry / MaxGenMem.size, MaxGenMem.size,
                    MaxGenMem.label);
            }
#ifdef NET
            fprintf(tfp_log, "sp%ld nn%hu\n", Speed, NumNodes);
#else /* NET */
            fprintf(tfp_log, "sp%ld\n", Speed);
#endif /* NET */
        }
        fprintf(tfp_log, "rm%ld mm%ld rf%ld\n",
            RateMut, RateMovMut, RateFlaw);
#ifdef MEM_PROF
        if (Log && InstExe.m)
        {   fprintf(tfp_log, "du%ld ta%ld",
                MemStat.DynMemUse, MemStat.TotNumAloc);
#if (defined __TURBOC__) && (!defined __GUI__)
            fprintf(tfp_log," co%lu",(I32u) coreleft());
#endif /* (defined __TURBOC__) && (!defined __GUI__) */
            fprintf(tfp_log,"\n");
        }
#endif  /* MEM_PROF */
#ifndef AMIGA
#ifndef DECVAX
        fflush(tfp_log);
#endif /* DECVAX */
#endif /* AMIGA */
    }
#ifdef ALCOMM
    if (log && MIsDFEnabled( TrtPlanEvent ) )
    {   TPlan();
    }
#endif /* ALCOMM */
}

void ChangeGeneBnker(gb)
I32s  gb;
{   I32s  ar, ci, flag;
#ifndef AMIGA
    I16s  gi;
    I32s  si;
#endif
    Pcells cp;

    if (gb)
    {   if (DiskBank)
        {   flag = 25;
        }
        else
            flag = 0;
        sprintf(Fbuf, "Please wait while I create a Genebank ...");
        FEPrintf(MSG_X, MSG_Y, 0, 0, Fbuf);
        sl = MkGBFromSoup(&siz_sl, &sl, flag);
#ifdef ERROR
        VerifyGB();
#endif /* ERROR */
    }
    else
    {   FreeGB(&sl, &siz_sl);
        for (ar = 0; ar < NumCelAr; ar++) for (ci = 0; ci < CelArSiz; ci++)
        {   if (ar == 0 && ci < 2)
            continue;
            cp = &cells[ar][ci];
            if (cp->ld)
            {   sprintf(cp->d.gen.label, "---");
                sprintf(cp->d.parent.label, "---");
                cp->d.gi = -1;
            }
        }
    }
}

#ifdef AMIGA
#if FRONTEND == BASIC
int kbhit()
{   return((getch() == ERR) ? FALSE : TRUE);
}
#endif
#endif /* AMIGA */

void MenuLine (linemsg, linecnt, menucol, menunum, menuopt)
char *linemsg;
int  *linecnt, menucol, *menunum, menuopt;
{
   if(menuopt){
      sprintf(Fbuf, "%2d - %s", (*menunum), linemsg);
      (*menunum)++;
   }
   else{
      sprintf(Fbuf, "%s", linemsg);
   }
   FEPrintf(menucol, MSG_Y + (*linecnt), 0, 0, Fbuf);
   (*linecnt)++;
}

void GoHistoCtrl()
{
   switch(IMode){
      case SIZ_HIST:
      case SIZM_HIST:
      case GEN_HIST:
         PageHist();
         break;
       default:
           sprintf(Fbuf, "no histogram active");
           FEError(-504, NOEXIT, NOWRITE, Fbuf);
           break;
   }
}

void PageHist()
{
   I8s  answer;
   int loopflag;
   I32s maxhistostrt;

   loopflag=1;
   while(loopflag){
      FEVidInit(0);
      sprintf(Fbuf,
         "U-pg up  D-pg dwn  u-scroll up  d-scroll dwn  q-quit->");
      FEPrintf(HLP_X, HLP_Y, 1, 0, Fbuf);

#ifdef AMIGA
#if FRONTEND == BASIC
      nodelay(stdscr, FALSE);
#endif /* FRONTEND == BASIC */
#endif /* AMIGA */

      answer = FEGetch(HLP_X + strlen(&(Fbuf[0])), HLP_Y);

#ifdef AMIGA
#if FRONTEND == BASIC
      nodelay(stdscr, TRUE);
#endif /* FRONTEND == BASIC */
#endif /* AMIGA */

      switch(answer){
         case 'U':
            histostrtidx+=histodsplnct;
            break;
         case 'D':
            histostrtidx-=histodsplnct;
            break;
         case 'u':
            histostrtidx++;
            break;
         case 'd':
            histostrtidx--;
            break;
         case 'q':
            loopflag=0;
            break;
         default:
            break;
      }
      maxhistostrt=histoentcnt-histodsplnct;
      histostrtidx=(histostrtidx<0)?
         0:((histostrtidx>maxhistostrt)?
         maxhistostrt:histostrtidx);
   }
}
