/* @(#)portable.c	1.13 1/13/95*/
/* portable.c   7-3-96  conditionally compiled functions for portability */
/* Tierra Simulator V4.3: Copyright (c) 1991, 1992, 1993, 1994, 1995, 1996 */
/* Tom Ray & Virtual Life */
/* MemProf labels: #12XX, in use: 1200 */
/* FEError labels: #12XX, in use: 1201-1203 */

#ifndef lint
static char     portable_sccsid[] = "@(#)portable.c	1.5        7/21/92";
#endif

/* #define ARG   */

#include "license.h"

#include <stdio.h>
#include "tierra.h"
#include "globals.h"

#ifdef AMIGA
#include "clib/dos_protos.h"
#endif /* AMIGA */

#ifdef LUNA
     extern int errno;
#endif

#ifdef SIGBLOCK

I32s TSigBlock(value)
I32s  value;
{
#ifdef TSysV
    sighold(SIGIO);
    return sighold(SIGPIPE);
#endif /* TSysV */
#ifdef BSD
    return sigblock(value);
#endif /* BSD */
#ifdef DECVAX
    return sigblock(value);
#endif /* DECVAX */
}

I32s TSigRelease(value)
I32s value;
{
#ifdef TSysV
    sigrelse(SIGPIPE);
    return sigrelse(SIGIO);
#endif /* TSysV */
#ifdef BSD
    return sigsetmask(value);
#endif /* BSD */
#ifdef DECVAX
    return sigsetmask(value);
#endif /* DECVAX */
}

#endif /* SIGBLOCK */

/* free for arrays not greater than 64K */
void tfree(ptr, siz, label)
void *  ptr;
I32s    siz;
I32s    label;
{
#ifdef SIGBLOCK
    I32s SigSaveSet;

    SigSaveSet=TSigBlock(SigSet);
#endif /* SIGBLOCK */

#ifndef OS2_MC
    free(ptr);
#else /* OS2_MC */
    DosFreeSeg((I16u) ptr >> 16);
#endif /* OS2_MC */

#ifdef SIGBLOCK
    (void) TSigRelease(SigSaveSet);
#endif /* SIGBLOCK */

#ifdef MEM_PROF
    MemStats(label, 0, siz);
#endif /* MEM_PROF */
}

/* free for arrays of greater than 64K */
void thfree(ptr, siz, label)
void *  ptr;
I32s    siz;
I32s    label;
{
#ifdef SIGBLOCK
    I32s SigSaveSet;

    SigSaveSet=TSigBlock(SigSet);
#endif /* SIGBLOCK */

#ifdef __TURBOC__
    farfree(ptr);
#endif /* __TURBOC__ */

#ifdef OS2_MC
    DosFreeSeg((I16u) ptr >> 16);
#endif /* OS2_MC */

#ifdef unix
    free(ptr);
#endif /* unix */

#ifdef __MWERKS__
    free(ptr);
#endif /* __MWERKS__ */

#ifdef DJGPP
    free(ptr);
#endif /* DJGPP */

#ifdef IBM3090
    free(ptr);
#endif /* IBM3090 */

#ifdef AMIGA
    free(ptr);
#endif /* AMIGA */

#ifdef DECVAX
    free(ptr);
#endif /* DECVAX */

#ifdef SIGBLOCK
    (void)TSigRelease(SigSaveSet);
#endif /* SIGBLOCK */

#ifdef MEM_PROF
    MemStats(label, 0, siz);
#endif /* MEM_PROF */
}

/* recalloc for Unix, or for DOS arrays not greater than 64K */
void * trecalloc(ptr, nsiz, osiz, label)
void *  ptr;
I32u    nsiz; /* new array size */
I32u    osiz; /* old array size */
I32s    label;
{   I8s   Fp  tp;
    I32u  swapsiz = (osiz < nsiz) ? osiz : nsiz;

    tp = tcalloc(nsiz, 1, label);
    if (ptr)
    {   memcpy(tp, ptr, swapsiz);
        tfree(ptr, osiz, label);
        ptr = NULL;
    }
    return (void *) tp;
}

/* huge recalloc for arrays of greater than 64K */
void * threcalloc(ptr, nsiz, osiz, label)
void *  ptr;
I32u    nsiz; /* new array size */
I32u    osiz; /* old array size */
I32s    label;
{
    I8s   Hp hp, Hp dp, Hp sp;
    I32s  segs, rem, i;
    I32u  swapsiz = (osiz < nsiz) ? osiz : nsiz;

    hp = (void *) thcalloc(nsiz, 1, label);
    if (ptr)
    {   segs = (I32s) (swapsiz / (I32u) MaxSizeT);
        rem  = (I32s) (swapsiz % (I32u) MaxSizeT);
        dp = hp; sp = ptr;
        if(segs) for(i = 0; i < segs; i++)
        {   memcpy(dp, sp, MaxSizeT);
            dp += MaxSizeT;
            sp += MaxSizeT;
        }
        if(rem)
            memcpy(dp, sp, rem);
        thfree(ptr, osiz, label);
        ptr = NULL;
    }
    return (void *) hp;
}

/* calloc for arrays not greater than 64K */
I8s Fp tcalloc(num, siz, label)
I32u   num;
I32u   siz;
I32s   label;
{   register I8s  Fp tp;
#ifdef OS2_MC
    I32u     HugeSize, i, ASize;
    I16u     NumSegs, Remain, rc;
    I8s Hp  tp;
    SEL      Selector;
#endif /* OS2_MC */
#ifdef SIGBLOCK
    I32s SigSaveSet;

    SigSaveSet=TSigBlock(SigSet);
#endif /* SIGBLOCK */

#ifndef OS2_MC
    tp = (I8s  Fp) calloc(num, siz);
#else /* OS2_MC */
    HugeSize = num * siz;
    i = HugeSize / 65536ul;
    NumSegs = (I16u) i;
    Remain  = (I16u) HugeSize % 65536ul;
    rc = DosAllocHuge(NumSegs,Remain,&Selector,3 * NumSegs,0);
    if(rc)
    {   sprintf(Fbuf,
        "Tierra tcalloc() error: calloc failed, num: %ld siz: %ld label: %ld ",
            num, siz, label);
#ifdef ARG
        fprintf(stderr, Fbuf);
        exit(errno);
#else  /* ARG */
#ifdef MEM_PROF
        MemStatReport();
#endif /* MEM_PROF */
        FEError(-label,EXIT,NOWRITE, Fbuf);
#endif /* ARG */
    }
    tp = (I8s Fp) ((I32u ) Selector << (I32u) 16);
    for(i = 0; i < HugeSize; i++)
    {   *(tp + i) = (I8s) 0;
    }
#endif /* OS2_MC */

#ifdef SIGBLOCK
    (void)TSigRelease(SigSaveSet);
#endif /* SIGBLOCK */

    if (!tp)
    {   sprintf(Fbuf,
        "Tierra tcalloc() error: calloc failed, num: %ld siz: %ld label: %ld ",
            num, siz, label);
#ifdef ARG
        fprintf(stderr, Fbuf);
        exit(errno);
#else /* ARG */
#ifdef MEM_PROF
        MemStatReport();
#endif /* MEM_PROF */
        FEError(-label, EXIT, NOWRITE, Fbuf);
#endif /* ARG */
    }

#ifdef MEM_PROF
    MemStats(label, 1, num * siz);
#endif /* MEM_PROF */
    return  tp;
}

/* calloc for arrays of greater than 64K */
I8s Hp thcalloc(num, siz, label)
I32u   num;
I32u   siz;
I32s   label;
{   register I8s  Hp hp;
#ifdef OS2_MC
    I32u     HugeSize, i, ASize;
    I16u     NumSegs, Remain, rc;
    SEL      Selector;
#endif /* OS2_MC */
#ifdef SIGBLOCK
    I32s SigSaveSet;

    SigSaveSet=TSigBlock(SigSet);
#endif /* SIGBLOCK */

#ifdef __TURBOC__
    hp = (I8s  Hp) farcalloc(num, siz);
#endif /* __TURBOC__ */

#ifdef unix
    hp = (I8s *) calloc(num, siz);
#endif /* unix  */

#ifdef __MWERKS__
    hp = (I8s *) calloc(num, siz);
#endif /* __MWERKS__  */

#ifdef DJGPP
    hp = (I8s *) calloc(num, siz);
#endif /* DJGPP */

#ifdef IBM3090
    hp = (I8s *) calloc(num, siz);
#endif /* IBM3090 */

#ifdef AMIGA
    hp = (I8s *) calloc(num, siz);
#endif /* AMIGA */

#ifdef DECVAX
    hp = (I8s *) calloc(num, siz);
#endif /* DECVAX */

#ifdef OS2_MC
    HugeSize = num * siz;
    i = HugeSize / 65536ul;
    NumSegs = (I16u) i;
    Remain  = (I16u) HugeSize % 65536ul;
    rc = DosAllocHuge(NumSegs,Remain,&Selector,3 * NumSegs,0);
    if(rc)
    {   sprintf(Fbuf,
       "Tierra thcalloc() error: calloc failed, num: %ld siz: %ld label: %ld ",
            num, siz, label);
#ifdef ARG 
        fprintf(stderr, Fbuf);
        exit(errno);
#else  /* ARG */
#ifdef MEM_PROF
        MemStatReport();
#endif /* MEM_PROF */
        FEError(-label, EXIT, NOWRITE, Fbuf);
#endif /* ARG */
    }
    hp = (I8s Hp) ((I32u ) Selector << (I32u) 16);
    for(i = 0; i < HugeSize; i++)
    {   *(hp + i) = (I8s) 0;
    }
#endif /* OS2_MC */

#ifdef SIGBLOCK
    (void)TSigRelease(SigSaveSet);
#endif /* SIGBLOCK */

    if (!hp)
    {   sprintf(Fbuf,
       "Tierra thcalloc() error: calloc failed, num: %ld siz: %ld label: %ld ",
        num, siz, label);
#ifdef ARG
        fprintf(stderr, Fbuf);
        exit(errno);
#else  /* ARG */
#ifdef MEM_PROF
        MemStatReport();
#endif /* MEM_PROF */
        FEError(-label, EXIT, NOWRITE, Fbuf);
#endif /* ARG */
    }

#ifdef MEM_PROF
    MemStats(label, 1, num * siz);
#endif /* MEM_PROF */
    return  hp;
}

I32u tfread(ptr, size, n, stream)
I8s Hp  ptr;
I32s  size, n;
FILE  *stream;
{   I32u  r_size = 0;
#ifdef __TURBOC__
    I32s  segs, rem, i = 0;

    segs = (I32s) ((unsigned int) n / MaxSizeT);
    rem  = (I32s) ((unsigned int) n % MaxSizeT);
    if(segs) for(i = 0; i < segs; i++)
        r_size += fread((I8s *) ((I8s Hp) ptr + ((unsigned int) i *
        (unsigned int) size * MaxSizeT)), (size_t) size, (size_t) MaxSizeT,
        stream);
    if(rem) r_size += fread((I8s *) ((I8s Hp) ptr + ((unsigned int) segs *
        (unsigned int) size * MaxSizeT)), (size_t) size, (size_t) rem, stream);
#endif /* __TURBOC__ */

#ifdef OS2_MC
    I32s  segs, rem, i = 0;

    segs = (I32s) ((unsigned int) n / MaxSizeT);
    rem  = (I32s) ((unsigned int) n % MaxSizeT);
    if(segs) for(i = 0; i < segs; i++)
        r_size += fread((I8s *) ((I8s Hp) ptr + ((unsigned int) i *
        (unsigned int) size * MaxSizeT)), (size_t) size, (size_t) MaxSizeT,
        stream);
    if(rem) r_size += fread((I8s *) ((I8s Hp) ptr + ((unsigned int) segs *
        (unsigned int) size * MaxSizeT)), (size_t) size, (size_t) rem, stream);
#endif /* OS2_MC */

#ifdef unix
    r_size = fread(ptr, size, n, stream);
#endif /* unix */

#ifdef __MWERKS__
    r_size = fread(ptr, size, n, stream);
#endif /* __MWERKS__ */

#ifdef DJGPP
    r_size = fread(ptr, size, n, stream);
#endif /* DJGPP */

#ifdef IBM3090
    r_size = fread(ptr, size, n, stream);
#endif /* IBM3090 */

#ifdef AMIGA
    r_size = fread(ptr, size, n, stream);
#endif /* AMIGA */

#ifdef DECVAX
    r_size = fread(ptr, size, n, stream);
#endif /* DECVAX */

    if(r_size != n)
    {   sprintf(Fbuf,
            "Tierra tfread() inconsistency: n = %ld  r_size = %ld", n, r_size);
#ifdef ARG 
        fprintf(stderr, Fbuf); 
        exit(errno);
#else
        FEError(-1201, NOEXIT, NOWRITE, Fbuf);
#endif
    }
    return r_size;
}

I32u tfwrite(ptr, size, n, stream)
I8s Hp  ptr;
I32s  size, n;
FILE *  stream;
{   I32u  r_size = 0;
#ifdef __TURBOC__
    I32s  segs, rem, i = 0;

    segs = (I32s) ((unsigned int) n / MaxSizeT);
    rem  = (I32s) ((unsigned int) n % MaxSizeT);
    if(segs) for(i = 0; i < segs; i++)
        r_size += fwrite((const I8s *) ((I8s Hp) ptr + ((unsigned int) i *
        (unsigned int) size * MaxSizeT)), (size_t) size, (size_t) MaxSizeT,
        stream);
    if(rem)
        r_size += fwrite((const I8s *) ((I8s Hp) ptr + ((unsigned int) segs *
        (unsigned int) size * MaxSizeT)), (size_t) size, (size_t) rem, stream);
#endif

#ifdef OS2_MC
    I32s  segs, rem, i = 0;

    segs = (I32s) ((unsigned int) n / MaxSizeT);
    rem  = (I32s) ((unsigned int) n % MaxSizeT);
    if(segs) for(i = 0; i < segs; i++)
        r_size += fwrite((const I8s *) ((I8s Hp) ptr + ((unsigned int) i *
            (unsigned int) size * MaxSizeT)), (size_t) size, (size_t) MaxSizeT,
            stream);
    if(rem)
        r_size += fwrite((const I8s *) ((I8s Hp) ptr + ((unsigned int) segs *
        (unsigned int) size * MaxSizeT)), (size_t) size, (size_t) rem, stream);
#endif

#ifdef unix
    r_size = fwrite(ptr, size, n, stream);
#endif

#ifdef __MWERKS__
    r_size = fwrite(ptr, size, n, stream);
#endif /* __MWERKS__ */

#ifdef DJGPP
    r_size = fwrite(ptr, size, n, stream);
#endif

#ifdef IBM3090
    r_size = fwrite(ptr, size, n, stream);
#endif

#ifdef AMIGA
    r_size = fwrite(ptr, size, n, stream);
#endif

#ifdef DECVAX
    r_size = fwrite(ptr, size, n, stream);
#endif

    if(r_size != n)
    {   sprintf(Fbuf,
           "Tierra tfwrite() inconsistency: n = %ld  r_size = %ld", n, r_size);
#ifdef ARG 
        fprintf(stderr, Fbuf);
#else
        FEError(-1202, NOEXIT, NOWRITE, Fbuf);
#endif
    }
    return r_size;
}

#ifdef MEM_PROF

void MemStats(label, af, size)
I32s  label; /* arbitrary label for this kind of allocation */
I32s  af;    /* 1 = allocation, 0 = free */
I32s  size;  /* size in bytes of this transaction (always positive) */
{   I32s  i;

    for (i = 0; i < MemStat.AlocSiz; i++)
        if (label == MemStat.mpa[i].label)
            break;
    if (i == MemStat.AlocSiz)
    {   if (!af)
        {   sprintf(Fbuf,
            "Tierra MemStats(): af zero for new label %ld (free before alloc)",
                label);
            FEError(-1203, EXIT, NOWRITE, Fbuf);
        }
        MemStat.mpa = (MemProf *) realloc(MemStat.mpa,
            (MemStat.AlocSiz + 1) * sizeof(MemProf));
        MemStat.AlocSiz++;
        MemStat.DynMemUse   += sizeof(MemProf);
        MemStat.mpa[0].siz  += sizeof(MemProf);
        MemStat.mpa[i].label = label;
        MemStat.mpa[i].num   = 1;
        MemStat.mpa[i].siz   = size;
    }
    else
    {   MemStat.mpa[i].num  += af ? 1 : -1;
        MemStat.mpa[i].siz  += af ? size : -size;
    }
    MemStat.TotNumAloc += af ? 1 : -1;
    MemStat.DynMemUse  += af ? size : -size;
}

void InitMemProf()
{
    MemStat.mpa = (MemProf *) calloc(1, sizeof(MemProf));
    MemStat.mpa[0].label = 1200;
    MemStat.mpa[0].num   = 1;
    MemStat.mpa[0].siz   = MemStat.DynMemUse = sizeof(MemProf);
    MemStat.AlocSiz = MemStat.TotNumAloc = 1;
}

void MemStatReport()
{   I32s  i, SumNum = 0, SumSiz = 0;

    fprintf(tfp_log, "\n     label      number        size\n\n");
    for (i = 0; i < MemStat.AlocSiz; i++)
    {   fprintf(tfp_log, "%10ld  %10ld  %10ld\n",
            MemStat.mpa[i].label, MemStat.mpa[i].num, MemStat.mpa[i].siz);
        SumNum += MemStat.mpa[i].num;
        SumSiz += MemStat.mpa[i].siz;
    }
    fprintf(tfp_log, "\n      Sums: %10ld  %10ld\n", SumNum, SumSiz);
    fprintf(tfp_log, "TotNumAloc: %10ld  %10ld: DynMemUse\n\n",
        MemStat.TotNumAloc, MemStat.DynMemUse);
}

#endif /* MEM_PROF */

#ifdef AMIGA

#ifndef NET
void sleep(unsigned int delay)
{
   Delay((long)(delay*50));
}
#endif /* NET */

int xseek(FILE *fp, long rpos, int mode)
{
   int error;
   long filsiz,extlen;
   char *extbuf;

   if(!(error=fseek(fp,(long)0,SEEK_END)<0)){
      if(!((filsiz=ftell(fp))<0)){
         if(rpos<=filsiz){
            error=fseek(fp,rpos,SEEK_SET);
         }
         else{
            extlen=rpos-filsiz;
            if(extbuf=calloc((size_t)extlen,
               (size_t)sizeof(char))){
               if(fwrite(extbuf,(size_t)extlen,
                  (size_t)sizeof(char),fp)){
                  error=fseek(fp,rpos,SEEK_SET);
               }
               else{
                  error=-1;
               }
               free(extbuf);
            }
            else{
               error=-1;
            }
         }
      }
      else{
         error=-1;
      }
   }
   return(error);
}

void __regargs _CXBRK(void)
{
#if FRONTEND == STDIO
   TC_Menu=1;
#endif /* FRONTEND == STDIO */
}

#endif /* AMIGA */

#ifdef __MWERKS__
#ifdef NET
int     gettimeofday(struct timeval * time, struct timezone * zone)
{
        unsigned long * seconds;

        GetDateTime(seconds);
        time->tv_sec = (long)seconds;
        time->tv_usec = 0L;

        /* timezone information not returned by this implementation
        zone.tz_minuteswest = 0;
        zone.tz_dsttime = 0;            */
}
/*
int      gethostname(char * name, int maxLength)
{       if (strlen(MacHost) > maxLength)
                return (-1);
        else
                strcpy( name, MacHost );
        return 0;
}
*/
int      getdomainname(char * name, int maxLength)
{       if (strlen(MacSite) > maxLength)
                return (-1);
        else
                strcpy( name, MacSite );
        return 0;
}
char *getenv(const char * variable)
{       char *  envValue;

        if (! strcmp(variable, "HOST"))
        {       strcpy( envValue, MacHost );
                return envValue;
        }
        else
        {
                return NULL;
        }
}
#endif /* NET */
#endif /* __MWERKS__ */

