/* rambank.c   7-3-96  rambank manager for the Tierra Simulator */
/* Tierra Simulator V4.3: Copyright (c) 1991, 1992, 1993, 1994, 1995, 1996 */
/* Tom Ray & Virtual Life */
/* MemProf labels: #14XX, in use: 1400-1406 */
/* FEError labels: #14XX, in use: 1407-1415 1417-1444 */

#ifndef lint
static char sccsid[] = "@(#)rambank.c    1.0        7/21/92";
#endif

#include "license.h"
#include "tierra.h"
#include "globals.h"

#ifdef MON
#include "obs.h"
#include "tmon_global.h"
#endif /* MON */

/*
 * CheckGenotype(ce, flags, lsiz_sl, lsl)
 * Check if cell ce is a new genotype.  If it is a new genotype, it will be
 * assigned a unique label.  If this genotype is not present in the RAM
 * genebank, it will be placed there, but there will be no demographic
 * information updated (this is not assumed to be a birth or death).
 * 
 *  flags: bit 0  (1): check .gen files
 *  flags: bit 1  (2): check .tmp files
 *  flags: bit 2  (4): check files even if rambank does not indicate presence
 *      of genotype on disk (used for startup of old or new soups).
 *  flags: bit 3  (8): use all files in the genebank to assemble the list in
 *      the rambank.  Each time a new size is checked, all genomes of that
 *      size in the genebank will become listed in the rambank as being on
 *      the disk (used for startup of old soups and for cumulative genebanks).
 *      However, the genomes will remain on disk until they are actually
 *      accessed, at which time they will be placed in the genebank in RAM.
 *  flags: bit 4 (16): when reading a file from the disk genebank,
 *      zero bit 1 of the tgl->bits field, otherwise preserve it.
 * 
 *      if soup_in variable CumGeneBnk is non zero, bit 3 is forced ON 
 */

void CheckGenotype(cd, flags, lsiz_sl, lsl)
Dem   *cd;  /* demography structure of cell to be checked */
I32s  flags;
I32s  *lsiz_sl;
SList Fp Fp Fp lsl;
{   I32s si;
    I16s gi;

    si = cd->gen.size;
    cd->gi = Lbl2Int(cd->gen.label);
    if (IsNewSize(si, *lsiz_sl, *lsl))
        NewSize(cd, flags, lsiz_sl, lsl);
    if ((gi = IsInGenQueue(cd, cd->hash, *lsl)) >= 0)
    {   cd->gi = gi;
        return;
    }
    if ((gi = IsInGenBank(cd, cd->hash, flags, *lsl)) >= 0)
    {   cd->gi = gi;
        return;
    }

    /* register new genotype in the lists */
    cd->gi = NewGenotype(cd, cd->hash, *lsl);
}

void DivGenBook(cp, InstExe, reaped, mom, same, disk)
    Pcells  cp ;
    Event   InstExe; /* current time */
    I32s    reaped; /* 1 = the reaper has acted */
    I32s    mom;    /* 1 = this is a mother cell, otherwise it it a daughter */
    I32s    same;   /* 1 = daughter is same genotype as mother */
    I32s    disk;   /* 1 = this creature originated from the disk (Inject) */
{   GList   *tgl, *tcgl;
    float   maxp, maxi;
    int     si, gi;
    Pcells  nc ;

    if (mom) /* this code deals only with the mother of the cell being born */
    {   tcgl = sl[si = cp->d.gen.size]->g[gi = cp->d.gi]; /* mother GList */
        if (tcgl && (I32u) tcgl <= 4)
        {
            tcgl = sl[si]->g[gi] = gq_read(si, gi); /* mother GList */
        }
        if ((I32u) tcgl <= 4)
            FEError(-1408,EXIT,NOWRITE, 
                "Tierra DivGenBook() mother genotype missing");
        if (cp->d.fecundity == 1 && !cp->d.mut && !cp->d.flaw)
            tcgl->d1 = cp->d.d1;
        else if (cp->d.fecundity == 2 && !cp->d.mut && !cp->d.flaw)
        {   tcgl->d2.inst = cp->d.inst + 1 - cp->d.d1.inst;
            tcgl->d2.flags = cp->d.flags - cp->d.d1.flags;
            tcgl->d2.mov_daught = cp->d.mov_daught;
            tcgl->d2.BreedTrue = same;
        }
        if (tcgl->max_cpus < cp->d.max_cpus)
            tcgl->max_cpus = cp->d.max_cpus ;
        return ;
    }
    /* the following code deals only with the daughter cell */
    nc = cp ;
    tgl = sl[si = nc->d.gen.size]->g[nc->d.gi];    /* new cell GList */
    if (!tgl->pop)
    {   NumGenotypes++; 
        sl[si]->num_g++;
        if (disk) /* this creature is an injection */
        {   NumGenDG++;
            SetBit(&tgl->bits, 0, 1);
        }
        if (IsBit(tgl->bits, 0))
            SetBit(&tgl->bits, 1, 1);
    }
    tgl->pop++;
    if (!sl[si]->num_c)
    {   NumSizes++; 
    }
    sl[si]->num_c++;
#ifdef MICRO
   ChkBrkBd();
#endif
#if FRONTEND != STDIO
#ifndef MON
    if ((IMode == SIZ_HIST)|| (IMode == SIZM_HIST) || (IMode == GEN_HIST))
        query_spec_d(si,nc->d.gi);
#endif /* MON */
#endif /* FRONTEND != STDIO */
/* this might be a good place to keep track of multiple parental genotypes. */
    if (reaped)
    {   maxp = (float) tgl->pop / (float) NumCells;
        if (maxp > tgl->MaxPropPop)
        {   tgl->MaxPropPop = maxp;
            tgl->mpp_time   = InstExe;         
        }
        maxi = (float) tgl->pop * nc->d.gen.size / (float) SoupSize;
        if (maxi > tgl->MaxPropInst)
            tgl->MaxPropInst = maxi;
    }

    /* criteria for saving genotype to disk */
    /* note: tgl->bit  0  permanent genotype name, saved in .gen file */
    if (reaped &&
        tgl->pop >= SavMinNum &&
        ((tgl->MaxPropInst > SavThrMem * .5) || (maxi > SavThrMem * .5)))
    {   if (!IsBit(tgl->bits, 0))
        {   SetBit(&tgl->bits, 0, 1);
            SetBit(&tgl->bits, 1, 0);
            extract(nc, 0);
        }
        if (IsBit(tgl->bits, 0) && IsBit(tgl->bits, 1))
        {   SetBit(&tgl->bits, 1, 0);
            extract(nc, 1);
        }
    }
}

void ReapGenBook(cp)
    Pcells  cp;
{   Pgl tgl;
    I32s  si = cp->d.gen.size;
    I16s  gi = cp->d.gi;

    tgl = sl[si]->g[gi];
#ifdef ERROR
    if (gi >= sl[si]->a_num)
    {   sprintf(Fbuf,
            "Tierra ReapBookeep() genotype %hd out of range", gi);
        FEError(-1413, EXIT, NOWRITE, Fbuf);
    }
    if ((I32u) tgl < 5)
    {   sprintf(Fbuf,
            "Tierra ReapBookeep() genotype %hd not in genebank", gi);
        FEError(-1414, EXIT, NOWRITE, Fbuf);
    }
#endif /* ERROR */
    tgl->pop--;    
    if ((!tgl->pop))
    {   if ((I32u) tgl > 4)
        {   if (!IsBit(tgl->bits, 0))
            {   sl[si]->g[gi] = NULL;
                sl[si]->hash[gi] = -1;
            }
            else
            {   extract(cp, 0); /* update .gen file here */
                sl[si]->g[gi] = (Pgl) 1;
            }
            FreeGen(tgl,1);
        }
        NumGenotypes--;
        sl[si]->num_g--;
    }
    sl[si]->num_c--;
    if ((!sl[si]->num_c))
    {   NumSizes--; 
    }
#if FRONTEND != STDIO
    if ((IMode == SIZ_HIST)|| (IMode == SIZM_HIST) || (IMode == GEN_HIST))
        query_spec_d(si,gi);
#endif /* FRONTEND != STDIO */
}

I8s IsNewSize(si, lsiz_sl, lsl)
    I32s  si;
    I32s  lsiz_sl;
    SList Fp Fp lsl;
{
    if (si < lsiz_sl && lsl[si])
        return 0;
    return 1;
}

void NewSize(cd, flags, lsiz_sl, lsl)
    Dem   *cd;
    I32s  flags;
    I32s  *lsiz_sl;
    SList Fp Fp Fp lsl;
{
    static char *ext[] = {"gen", "tmp"};
    I32s i, si, exi;
    I16s gi;
#ifndef AMIGA
    GList Fp Fp  tgl;
#endif
    Pgl sgl = NULL;
    FILE *afp;
    head_t head;
    indx_t *indx, *tindx;
#ifndef AMIGA
    indx_t gindx;
#endif /* AMIGA */

    si = cd->gen.size;
    if (si >= *lsiz_sl)
    {   (*lsl) = (SList Fp Fp) trecalloc((void *) (*lsl),
            sizeof(SList Fp) * (si + 1),
            sizeof(SList Fp) * *lsiz_sl, 1903);
        *lsiz_sl = si + 1;
#ifdef ERROR
        sprintf(Fbuf, "genebank: recalloc, *lsiz_sl = %ld", *lsiz_sl - 1);
        FEPrintf(MSG_X, MSG_Y, 0, 2, Fbuf);
#endif /* ERROR */
    }
    (*lsl)[si] = (SList *) tcalloc(1, sizeof(SList), 1400);
    (*lsl)[si]->a_num = 20;
    (*lsl)[si]->g = (Pgl Fp)  tcalloc(20, sizeof(Pgl),  1401);
    (*lsl)[si]->hash = (I32s Fp) tcalloc(20, sizeof(I32s), 1405);
    if (IsBit(flags, 3))  /* use for old soups, and cumulative genebanks */
        for (exi = 0; exi < 2; exi++) if (IsBit(flags, exi))
    {
        sprintf(Buff, "%s%04ld.%s", GenebankPath, si, ext[exi]);
        afp = fopen(Buff, "rb");
        if (!afp)
            continue;
        head = read_head(afp);
        indx = read_indx(afp, &head);
        for (i = head.n - 1; i >= 0; i--)
        {
            tindx = &indx[i];
            sgl = get_gen(afp, &head, tindx, i);
            gi = Lbl2Int(sgl->gen.label);
            if (gi >= (*lsl)[si]->a_num)
            {   (*lsl)[si]->g = (Pgl Fp)
                    trecalloc((void *) ((*lsl)[si]->g),
                    sizeof(Pgl) * (gi + 1),
                    sizeof(Pgl) * (*lsl)[si]->a_num, 1401);
                (*lsl)[si]->hash = (I32s Fp)
                    trecalloc((void *) ((*lsl)[si]->hash),
                    sizeof(I32s) * (gi + 1),
                    sizeof(I32s) * (*lsl)[si]->a_num, 1405);
                (*lsl)[si]->a_num = gi + 1;
            }
            (*lsl)[si]->hash[gi] = sgl->hash;
            if (!exi)
            {   (*lsl)[si]->g[gi] = (Pgl)1;    /* permanent genotype name */
                FreeGen(sgl,0);    
            }
            else
                (*lsl)[si]->g[gi] = sgl;
        }
        fclose(afp);
        if (exi)
            TmplAdd(si, &tmpl);
        if (indx)
        {   thfree((void *) indx, head.n_alloc * sizeof(indx_t), 600);
            indx = NULL;
        }
    }
}

void TmplAdd(si, tl)
I32s  si;
TmpList  *tl;
{   I32s  i = 0;

    if (!(tl->n))
    {   tl->l = (I32s *) tcalloc(1, sizeof(I32s), 1406);
        tl->n = 1;
        tl->l[0] = si;
    }
    else for (i = 0; i < tl->n; i++)
        if (tl->l[i] == si)
            break;
    if (i == tl->n)
    {   tl->l = (I32s *) trecalloc((void *) tl->l,
            sizeof(I32s) * (tl->n + 1),
            sizeof(I32s) * tl->n, 1406);
        tl->l[tl->n] = si;
        tl->n++;
    }
}

void TmplRem(si, tl)
I32s  si;
TmpList  *tl;
{   I32s  i, j, *p = NULL;

    if (!(tl->n))
    {   sprintf(Fbuf,
            "Tierra RemTmpl() error: si = %ld, no tmp files in tmpl", si);
        FEError(-1415, EXIT, NOWRITE, Fbuf);
    }
    for (i = 0; i < tl->n; i++)
        if (tl->l[i] == si)
            break;
    if (i == tl->n)
        return;
    else
    {   if (tl->n - 1 > 0)
        {   p = (I32s *) tcalloc((tl->n - 1), sizeof(I32s), 1406);
            for (i = 0, j = 0; j < tl->n; i++, j++)
            {   if (tl->l[j] == si)
                {   j++;
                    continue;
                }
                p[i] = tl->l[j];
            }
        }
        tfree(tl->l, tl->n * sizeof(I32s), 1406);
        tl->l = NULL;
        if (p)
            tl->l = p;
        tl->n--;
    }
}

I16s IsInGenQueue(cd, hash, lsl)
    Dem   *cd;            /* returns the index of the genotype in the list */
    I32s  hash;
    SList Fp Fp lsl;
{
    I32s si = cd->gen.size, lhash;
    I16s gi = cd->gi, i;
    GList  *tgl;

    if (gi >= 0)
    {   if (gi < lsl[si]->a_num && (I32u) lsl[si]->g[gi] > 4)
            return gi;
        return -1;
    }
    for (i = 0; i < lsl[si]->a_num; i++)
    {   tgl = lsl[si]->g[i];
        lhash = lsl[si]->hash[i];
#ifdef ERROR
        if ((I32u) tgl > 4)
        {   if (IsSameGen(si, cd->genome, tgl->genome))
            {   if (hash != lhash)
                {   sprintf(Fbuf,
                    "size = %d, index = %hd, label = %s, hash = %d and %d\n",
                    si, i, cd->gen.label, hash, lhash) ;
                    FEPrintf(MSG_X, MSG_Y, 0, 2, Fbuf);
                    FEError(-1417, EXIT, NOWRITE,
                "Tierra IsInGenQueue() error: IsSameGen, but not same hash");
                }
            }
        }
#endif /* ERROR */
        if ((I32u) tgl > 4 && hash == lhash &&
            IsSameGen(si, cd->genome, tgl->genome))
        {
            return i;
        }
    }
    return -1;
}

/*
 * Check to see if ce is in the disk genebank.  This will require the reading
 * of successive genotypes of this size from the .gen and .tmp files on disk.
 * Each genotype that is read will be placed in the genequeue and the complete
 * genome will be placed in the rambank.
 */

I16s IsInGenBank(cd, hash, flags, lsl)
    Dem   *cd;
    I32s  hash, flags;
    SList Fp Fp lsl;
{
    I32s si = cd->gen.size;
#ifndef AMIGA
    I32s i;
#endif /* AMIGA */
    I32u j, n;
    I16s gi = cd->gi, gih = -1;
    Pgl g;
    FILE *afp;
    head_t head;
    indx_t *indx, *tindx;
#ifndef AMIGA
    indx_t gindx;
#endif /* AMIGA */

    /*
     * return -1 if we are looking for a specific geneotype, and it does not
     * appear in the genequeue list, and we are not starting up a soup
     */
    if (gi >= 0 /* we are looking for a specific genotype */
        && (gi >= lsl[si]->a_num /* our gi is off the list */
        || !lsl[si]->g[gi]) /* gi does not exist, either on disk or in RAM */
        && !IsBit(flags, 2)) /* we are not starting up a soup */
        return -1;

    if (!IsBit(flags, 2)) /* we are not starting up a soup */
    {   if (gi >= 0)  /* if we know what genotype we are looking for */
            if (lsl[si]->hash[gi] != hash)
                return -1;
        else
        {   for (j = 0; j < lsl[si]->a_num; j++)
            {   if (!((I32s) lsl[si]->g[j] > 0 && (I32s) lsl[si]->g[j] <= 4))
                    continue;
                if (lsl[si]->hash[j] == hash)
                {   gih = j;
                    break;
                }
            }
            if (gih == -1)
                return -1;
        }
    }

#ifdef IBM3090
    sprintf(Buff, "%04ld.gen.d", si);
#else
    sprintf(Buff, "%s%04ld.gen", GenebankPath, si);
#endif
    if (afp = fopen(Buff, "rb"))
    {   head = read_head(afp);
        indx = read_indx(afp, &head);
    }
    else
        return -1;
    if (gi >= 0)  /* if we know what genotype we are looking for */
    {   if (gi >= lsl[si]->a_num)
        {   lsl[si]->g = (Pgl Fp) trecalloc((void *) (lsl[si]->g),
                sizeof(Pgl) * (gi+1), sizeof(Pgl) * lsl[si]->a_num, 1401);
            lsl[si]->hash = (I32s Fp) trecalloc((void *) (lsl[si]->hash),
                sizeof(I32s) * (gi+1),
                sizeof(I32s) * lsl[si]->a_num, 1405);
            lsl[si]->a_num = gi + 1;
        }
        n = find_gen(indx, Int2Lbl(gi), head.n);
        if (n < head.n)
        {
            tindx = &indx[n];
            lsl[si]->g[gi] = g = get_gen(afp, &head, tindx, n);
            if (IsBit(flags, 4))
                SetBit(&g->bits, 1, 0);
#ifdef ERROR
            if (IsSameGen(si, (FpInst) (cd->genome), g->genome))
            {   if (hash != g->hash)
                    FEError(-1418, EXIT, WRITE,
             "Tierra IsInGenBank() error: IsSameGen, but not same hash");
            }
#endif /* ERROR */
            /* if disk genotype matches soup genotype */
            /* name cell and put genotype in genequeue */
            if (hash == g->hash &&
                IsSameGen(si, (FpInst) (cd->genome), g->genome))
            {   cd->gen = g->gen;
                cd->gi = gi;
                if (indx)
                {   thfree((void *) indx, head.n_alloc * sizeof(indx_t),
                        600);
                    indx = NULL;
                }
                fclose(afp);
                return gi;
            }
        }
    }
    else /* we don't know what genotype we are looking for */
    /*
     * check only genotypes that are listed in the rambank as on
     * disk, but whose genomes are not held in the rambank
     * (0 < lsl[si]->g[j] <= 4); or which are not listed in the
     * rambank at all (!lsl[si]->g[j]), if bit 2 is set, which
     * means we are starting a new or old soup and don't have
     * a list of what is on disk. 
     */
    for (j = 0; j < lsl[si]->a_num; j++)
    {   if (!((I32s) lsl[si]->g[j] > 0 && (I32s) lsl[si]->g[j] <= 4
                || !lsl[si]->g[j] && IsBit(flags, 2)))
            continue;
        if (lsl[si]->hash[j] != hash)
            continue;
        n = find_gen(indx, Int2Lbl(j), head.n);
        if (n < head.n)
        {
            tindx = &indx[n];
            lsl[si]->g[j] = g = get_gen(afp, &head, tindx, n);
#ifdef ERROR
            if (g->hash != hash || g->hash != lsl[si]->hash[j])
            {   sprintf(Fbuf,
"hash values should match: hash = %ld g->hash = %ld lsl[si]->hash[j] = %ld",
                    hash, g->hash, lsl[si]->hash[j]);
                FEError(-1419, EXIT, NOWRITE, Fbuf);
            }
#endif /* ERROR */
            if (IsBit(flags, 4))
                SetBit(&g->bits, 1, 0);
            /* if disk genotype matches soup genotype */
            /* name cell and put genotype in genequeue */
            if (hash == g->hash && g->hash == lsl[si]->hash[j] &&
                IsSameGen(si, cd->genome, g->genome))
            {   cd->gen = g->gen;
                cd->gi = j;
                if (indx)
                {   thfree((void *) indx, head.n_alloc * sizeof(indx_t),
                        600);
                    indx = NULL;
                }
                fclose(afp);
                return (I16s) j;
            }
        }
    }
    if (afp)
    {
        if (indx)
        {   thfree((void *) indx, head.n_alloc * sizeof(indx_t), 600);
            indx = NULL;
        }
        fclose(afp);
    }
    return -1;
}

/*
 * add a new genotype to the RAM list
 */

I16s NewGenotype(cd, hash, lsl)
    Dem   *cd;
    I32s  hash;
    SList Fp Fp lsl;
{
    Pgl   tgl;
    I32s  i, size = cd->gen.size;
#if PLOIDY > 1
    I32s  j;
#endif /* PLOIDY > 1 */
    I16s  gi;
    I8s   found = 0;

    /* find a free name if there is one */
    for (i = 0; i < lsl[size]->a_num; i++) if (!lsl[size]->g[i])
    {   gi = i;
        found = 1;
        break;
    }
    if (!found)
    {   gi = lsl[size]->a_num;
        lsl[size]->g = (Pgl Fp) trecalloc((void *) (lsl[size]->g),
            sizeof(Pgl) * (gi + 4), sizeof(Pgl) * gi, 1401);
        lsl[size]->hash = (I32s Fp) trecalloc((void *) (lsl[size]->hash),
            sizeof(I32s) * (gi + 4), sizeof(I32s) * gi, 1405);
        lsl[size]->a_num += 4;
    }
    lsl[size]->hash[gi] = hash;
    lsl[size]->g[gi] = tgl =
        (GList *) tcalloc(1, sizeof(GList), 1402);
    tgl->gen.size = cd->gen.size = size;
    tgl->genome = (FpInst) tcalloc(size, sizeof(Instruction), 1403);
    tgl->gbits = (FpGenB) tcalloc(size, sizeof(GenBits), 1404);
    for (i = 0; i < size; i++) 
#if PLOIDY == 1
        tgl->genome[i] = cd->genome[i];
#else
        for (j = 0; j < PLOIDY; j++)
            tgl->genome[i][j] = cd->genome[i][j];
#endif
    strcpy(tgl->gen.label, Int2Lbl(gi));
    tgl->originC = time(NULL);
    tgl->originI = InstExe;
    tgl->parent = cd->parent;
    tgl->bits = 0;
    tgl->hash = hash;
    tgl->pop = 0;
    tgl->max_cpus = 0;
    if (reaped)
    {   tgl->MaxPropPop = (float) 1 / (float) NumCells;
        tgl->MaxPropInst = (float) size / (float) SoupSize;
        tgl->mpp_time = InstExe;
    }
    tgl->ploidy = cd->ploidy;
/*
    tgl->track = cd->tr;
 */
    return gi;
}

void FreeGen(g,qr)
    GList  *g;
    I32s   qr;  /* flag to call qr_rem */
{   if (!g)
        return;
    if (g->genome)
    {   tfree((void *) (g->genome), g->gen.size * sizeof(Instruction), 1403);
        g->genome = NULL;
    }
    if (g->gbits)
    {   tfree((void *) (g->gbits), g->gen.size * sizeof(GenBits), 1404);
        g->gbits = NULL;
    }
    tfree((void *) g, sizeof(GList), 1402);
    g = NULL;
}

GList *gq_read(si, gi)
{   GList *p = sl[si]->g[gi];
    I16s n;
    FILE *fp;
    head_t head;
    indx_t *indx, *tindx;
#ifndef AMIGA
   indx_t gindx;
#endif /* AMIGA */

    if ((I32u) p > 4)
        return p;
    sprintf(Buff,
#ifdef IBM3090
        "%04ld.%s.d",
#else
        "%s%04ld.%s", GenebankPath,
#endif
        si, (I32u) p == 1 ? "gen" : "mem");
    if (!(fp = open_ar(Buff, si, GFormat, 0)))
    {   sprintf(Fbuf, "Tierra gq_read() unable to open genome file %s", Buff);
        FEError(-1420, EXIT, WRITE, Fbuf);
    }
    head = read_head(fp);
    indx = read_indx(fp, &head);
    if((n = find_gen(indx, Int2Lbl(gi), head.n))==head.n){
        sprintf(Fbuf, "genome not found");
        FEError(-1444, EXIT, WRITE, Fbuf);
    }
    tindx = &indx[n];
    p = get_gen(fp, &head, tindx, n);
    fclose(fp);
    if (indx)
    {   thfree((void *) indx, head.n_alloc * sizeof(indx_t), 600);
        indx = NULL;
    }
    return p;
}

void VerifyGB() /* verify genebank */
{   I32s  gNumSizes = 0, cNumSizes = 0, cgNumSizes = 0;
    I32s  gNumGenot = 0, cNumGenot = 0, cgNumGenot = 0;
    I32s  gNumCells = 0, cNumCells = 0, cgNumCells = 0;
    I32s  cgsNumGenot = 0, ggNumGenot = 0;
    I32s  cgsNumCells = 0, ggNumCells = 0;
    I32s  tsiz_sl = 1, si, ar, ci;
    I16s  gi;
    Pcells cp;
    GList  Fp pgl;
    SList  Fp Fp tsl, Fp psl;

    /* begin cells array check */
    tsl = (SList Fp Fp) tcalloc(1, sizeof(SList Fp), 1903);
    for (ar = 0; ar < NumCelAr; ar++) for (ci = 0; ci < CelArSiz; ci++)
    {   if (ar == 0 && ci < 2)
            continue;
        cp = &cells[ar][ci];
        if (cp->ld)
        {   cNumCells++;
            si = cp->d.gen.size;
            if (si >= siz_sl)
            {   sprintf(Fbuf,
                 "Tierra VerifyGB() size %ld out of range in genebank", si);
                FEError(-1421, EXIT, NOWRITE, Fbuf);
            }
            psl = sl[si];
            if (!psl)
            {   sprintf(Fbuf,
                 "Tierra VerifyGB() sl[%ld] not allocated in genebank", si);
                FEError(-1422, EXIT, NOWRITE, Fbuf);
            }
            gi = cp->d.gi;
            if (gi >= psl->a_num)
            {   sprintf(Fbuf,
               "Tierra VerifyGB() genome %hd out of range in genebank", gi);
                FEError(-1423, EXIT, NOWRITE, Fbuf);
            }
            pgl = psl->g[gi];
            if ((I32u) pgl < 4)
            {   sprintf(Fbuf,
                 "Tierra VerifyGB() gl[%hd] not allocated in genebank", gi);
                FEError(-1424, EXIT, NOWRITE, Fbuf);
            }
            if (!IsSameGen(si, soup + cp->mm.p, pgl->genome))
                FEError(-1425, EXIT, NOWRITE,
                    "Tierra VerifyGB() cell and genebank do not match");
            if (si >= tsiz_sl)
            {   tsl = (SList Fp Fp) trecalloc((void *) tsl,
                    (si + 1) * sizeof(SList Fp), tsiz_sl * sizeof(SList Fp),
                    1903);
                tsiz_sl = si + 1;
            }
            if (!tsl[si])
            {   tsl[si] = (SList Fp) tcalloc(1, sizeof(SList), 1400);
                tsl[si]->g = (Pgl Fp) tcalloc(gi + 1, sizeof(Pgl), 1401);
                tsl[si]->hash = (I32s Fp) tcalloc(gi + 1, sizeof(I32s), 1405);
                tsl[si]->a_num = gi + 1;
            }
            if (!tsl[si]->num_c)
            {   if (tsl[si]->num_g)
                    FEError(-1426, NOEXIT, NOWRITE,
                    "Tierra VerifyGB() !tsl[si]->num_c but tsl[si]->num_g");
                cNumSizes++;
            }
            tsl[si]->num_c++;
            if (gi >= tsl[si]->a_num)
            {   tsl[si]->g = (Pgl Fp) trecalloc((void *) (tsl[si]->g),
                    (gi + 1) * sizeof(Pgl),
                    tsl[si]->a_num * sizeof(Pgl), 1401);
                tsl[si]->hash = (I32s Fp) trecalloc((void *) (tsl[si]->hash),
                    (gi + 1) * sizeof(I32s),
                    tsl[si]->a_num * sizeof(I32s), 1405);
                tsl[si]->a_num = gi + 1;
            }
            if ((I32u) tsl[si]->g[gi] < 4)
            {   tsl[si]->g[gi] = (GList Fp) tcalloc(1, sizeof(GList), 1402);
                cNumGenot++;
                tsl[si]->num_g++;
            }
            tsl[si]->g[gi]->pop++;
        }
    } /* check and free temporary genebank */
    for (si = 0; si < tsiz_sl; si++)
    {   if (tsl[si])
        {   if (tsl[si]->num_c != sl[si]->num_c)
            {   sprintf(Fbuf,
             "Tierra VerifyGB() tsl[%ld]->num_c != sl[%ld]->num_c", si, si);
                FEError(-1427, NOEXIT, NOWRITE, Fbuf);
            }
            if (tsl[si]->num_g != sl[si]->num_g)
            {   sprintf(Fbuf,
             "Tierra VerifyGB() tsl[%ld]->num_g = %d != sl[%ld]->num_g = %d", 
                 si, tsl[si]->num_g,si,sl[si]->num_g);
                FEError(-1428, NOEXIT, NOWRITE, Fbuf);
            }
            if (tsl[si]->num_c && tsl[si]->g)
            {   cgNumSizes++;
                cgsNumCells += tsl[si]->num_c;
                cgsNumGenot += tsl[si]->num_g;
                for (gi = 0; gi < tsl[si]->a_num; gi++)
                {   if ((I32u) tsl[si]->g[gi] > 4)
                    {   if (tsl[si]->g[gi]->pop != sl[si]->g[gi]->pop)
                        {   sprintf(Fbuf,
"Tierra VerifyGB() tsl[%ld]->g[%hd]->pop = %d != sl[%ld]->g[%hd]->pop = %d",
                       si, gi, tsl[si]->g[gi]->pop,si, gi,sl[si]->g[gi]->pop);
                            FEError(-1429, NOEXIT, NOWRITE, Fbuf);
                        }
                        cgNumGenot++;
                        cgNumCells += tsl[si]->g[gi]->pop;
                        tfree((void *) (tsl[si]->g[gi]), sizeof(GList), 1402);
                    }
                }
                tfree((void *) (tsl[si]->g), tsl[si]->a_num * sizeof(Pgl),
                    1401);
                tfree((void *) (tsl[si]->hash), tsl[si]->a_num * sizeof(I32s),
                    1405);
                tfree((void *) tsl[si], sizeof(SList), 1400);
            }
        }
    }
    tfree((void *) tsl, tsiz_sl * sizeof(SList Fp), 1903);
    if (NumCells != cNumCells || NumCells != cgNumCells ||
        NumCells != cgsNumCells)
        FEError(-1430,NOEXIT,NOWRITE,
            "Tierra VerifyGB() NumCells cells array inconsistency");
    if (NumGenotypes != cNumGenot || NumGenotypes != cgNumGenot ||
        NumGenotypes != cgsNumGenot)
        FEError(-1431,NOEXIT,NOWRITE,
            "Tierra VerifyGB() NumGenot cells array inconsistency");
    if (NumSizes != cNumSizes || NumSizes != cgNumSizes)
    {   sprintf(Fbuf,
         "Tierra VerifyGB() NumSizes cells array inconsistency = %d, %d, %d",
            NumSizes,cNumSizes,cgNumSizes);
        FEError(-1432, NOEXIT, NOWRITE, Fbuf);
    }
    /* end cells array check */

    /* begin genebank check */
    for (si = 0; si < siz_sl; si++)
    {   psl = sl[si];
        if (!psl)
            continue ;
        if (!psl->num_c || !psl->num_g)
            FEError(-1433, NOEXIT, NOWRITE,
                "Tierra VerifyGB() !sl[si]->num_c or !sl[si]->num_g");
        if (sl[si]->num_c)
        {   gNumSizes++;
            ggNumCells += sl[si]->num_c;
        }
        if (sl[si]->num_g)
            ggNumGenot += sl[si]->num_g;
        for (gi = 0; gi < sl[si]->a_num; gi++)
        {   pgl = psl->g[gi];
            if ((I32u) pgl < 4 || !pgl->pop)
                continue ;
            gNumGenot++;
            gNumCells += pgl->pop;
        }
    }
    if (NumCells != gNumCells || NumCells != ggNumCells)
        FEError(-1434,NOEXIT,NOWRITE,
            "Tierra VerifyGB() NumCells genebank inconsistency");
    if (NumGenotypes != gNumGenot || NumGenotypes != ggNumGenot)
        FEError(-1435,NOEXIT,NOWRITE,
            "Tierra VerifyGB() NumGenot genebank inconsistency");
    if (NumSizes != gNumSizes)
    {   sprintf(Fbuf,
            "Tierra VerifyGB() NumSizes genebank inconsistency = %d, %d",
            NumSizes,gNumSizes);
        FEError(-1436, NOEXIT, NOWRITE, Fbuf);
    }
    /* end genebank check */
}

void GBGarbageCollect()
{   I32s    i, j, maxsiz = 0, tail;
    GList   Fp pgl;
    I8s     path[80];
    FILE    *fp;
    head_t  head;
    indx_t  *indx;
#ifndef AMIGA
    indx_t  gindx;
#endif /* AMIGA */

    for (i = siz_sl - 1; i >= 0; i--)      /* for each allocated size class */
    {   if (sl[i])
        {   if (sl[i]->num_c)
            {   if (!maxsiz)                     /* find largest size class */
                    maxsiz = i;
                tail = -1;
                for (j = sl[i]->a_num - 1; j >= 0; j--)
                {   if ((I32u) (pgl = sl[i]->g[j]) > 4 && !pgl->pop
                        && !IsBit(pgl->bits, 0))
                    {
                        if (pgl->genome)
                        {   tfree((void *) (pgl->genome),
                                i * sizeof(Instruction), 1403);
                            pgl->genome = NULL;
                        }
                        if (pgl->gbits)
                        {   tfree((void *) (pgl->gbits),
                                i * sizeof(GenBits), 1404);
                            pgl->gbits = NULL;
                        }
                        tfree((void *) (sl[i]->g[j]), sizeof(GList), 1402);
                        sl[i]->g[j] = NULL;
                    }
                    if (tail < 0 && sl[i]->g[j])
                        tail = j;    /* skip empty geotypes at end of array */
                }
                if (tail < sl[i]->a_num - 1)
                {   if (tail < 0)             /* no genotypes in size class */
                    {   if (sl[i]->g)
                        {   tfree((void *) (sl[i]->g),
                                sl[i]->a_num * sizeof(Pgl), 1401);
                            sl[i]->g = NULL;
                        }
                        if (sl[i]->hash)
                        {   tfree((void *) (sl[i]->hash),
                                sl[i]->a_num * sizeof(I32s), 1405);
                            sl[i]->hash = NULL;
                        }
                        if (sl[i])
                        {   tfree((void *) sl[i], sizeof(SList), 1400);
                            sl[i] = NULL;
                        }
                    }
                    else           /* shorten g arrays to avoid empty tails */
                    {   sl[i]->g = (Pgl Fp) trecalloc((void *) (sl[i]->g),
                            (tail + 1) * sizeof(Pgl),
                            sl[i]->a_num * sizeof(Pgl), 1401);
                        sl[i]->hash = (I32s Fp)
                            trecalloc((void *) (sl[i]->hash),
                            (tail + 1) * sizeof(I32s),
                            sl[i]->a_num * sizeof(I32s), 1405);
                        sl[i]->a_num = tail + 1;
                    }
                }
            }
            else /* no creatures of this size, free sl[i] and sl[i]->g */
            {   sprintf(path, "%s%04ld.gen", GenebankPath, i);
                fp = open_ar(path, i, GFormat, -1);
                head = read_head(fp);
                indx = read_indx(fp, &head);
                for (j = sl[i]->a_num - 1; j >= 0; j--)
                    if ((I32u) (pgl = sl[i]->g[j]) > 4)
                    {   if (pgl->pop)
                        {   sprintf(Fbuf,
   "Tierra GBGarbageCollect(), pgl = %ld, pgl->pop not zero, can't free",
                                (I32s) pgl);
                            FEError(-1437, NOEXIT, NOWRITE, Fbuf);
                        }
                        if (IsBit(pgl->bits, 0)) /* save genome to disk */
                            add_gen(fp, &head, &indx, pgl);
                        FreeGen(pgl,1);
                        sl[i]->g[j] = NULL;
                    }
                fclose(fp);
                if (!head.n)
#ifdef DECVAX
                    delete(path);
#else
                    unlink(path);
#endif
                if (indx)
                {   thfree((void *) indx, head.n_alloc * sizeof(indx_t), 600);
                    indx = NULL;
                }
                if (sl[i]->g)
                {   tfree((void *) (sl[i]->g), sl[i]->a_num * sizeof(Pgl),
                        1401);
                    sl[i]->g = NULL;
                }
                if (sl[i]->hash)
                {   tfree((void *) (sl[i]->hash), sl[i]->a_num * sizeof(I32s),
                        1405);
                    sl[i]->hash = NULL;
                }
                if (sl[i])
                {   tfree((void *) sl[i], sizeof(SList), 1400);
                    sl[i] = NULL;
                }
            }
        }
    }
    if (maxsiz < (I32s) (.5 * (siz_sl - 1)))
    {   sl = (SList Fp Fp) trecalloc((void *) sl,
            (maxsiz + 1) * sizeof(SList Fp),
            siz_sl * sizeof(SList Fp), 1903);
        siz_sl = maxsiz + 1;
    } /* end garbage collect for genebank */
}

void GenExTemp(adrt, tsize)
    I32s     adrt;  /* address of beginning of template */
    I32s    tsize; /* template size */
{
    I32s  i;
    I32u  who;  /* 0 same cell; 1 daughter cell; 2 other cell; */
                /* 3 free memory; 4 daughter of other cell */
    I32s   dist;
    Pgl   tgl, ogl;
    Pcells  ct;

    tgl = sl[ce->d.gen.size]->g[ce->d.gi];
    for (i = 0; i < tsize; i++)
    {   ct = ce;  /* WHAT TO DO WITH THIS? */
        who = WhoIs(&ct, ad(ce->c.c->ip + 1 + i));/* who has templ pattern */
        if (who < 4) tgl->bits |= (I32u) (ONE << (I32u) (12 + who));
        else tgl->bits |= (I32u) (ONE << (I32u) (12 + 2));
        if (!who)
        {   dist = ad(ce->c.c->ip + 1 + i) - ce->mm.p;
            dist = ad(dist);
#ifdef ERROR
            if (tgl->genome == NULL || dist < 0 || dist >= tgl->gen.size)
                FEError(-1438,EXIT,WRITE, "Tierra GenExTemp() error 0");
#endif /* ERROR */
#if PLOIDY == 1
            tgl->gbits[dist] |= 1;
#else /* PLOIDY == 1 */
            tgl->gbits[dist][ce->c.c->so] |= 1;
#endif /* PLOIDY == 1 */
        }
        if (who == 2)
        {   ogl = sl[ct->d.gen.size]->g[ct->d.gi];
            if (IsBit(ogl->bits, 0))
            {   ogl->bits |= (I32u) (ONE << (I32u) (12 + 4));
                dist = ad(ce->c.c->ip + 1 + i) - ct->mm.p;
                dist = ad(dist);
#ifdef ERROR
                if (ogl->genome == NULL || dist < 0 || dist >= ogl->gen.size)
                   FEError(-1439,EXIT,NOWRITE, "Tierra GenExTemp() error 1");
#endif /* ERROR */
#if PLOIDY == 1
                ogl->gbits[dist] |= (1 << 1);
#else /* PLOIDY == 1 */
                ogl->gbits[dist][ce->c.c->so] |= (1 << 1);
#endif /* PLOIDY == 1 */
            }
        }
        ct = ce;
        who = WhoIs(&ct, ad(adrt + i)); /* who has complementary template */
        if (who < 4) tgl->bits |= (I32u) (ONE << (I32u) (7 + who));
        else tgl->bits |= (I32u) (ONE << (I32u) (7 + 2));
        if (!who)
        {   dist = ad(adrt + i) - ce->mm.p;
            dist = ad(dist);
#ifdef ERROR
            if (tgl->genome == NULL || dist < 0 || dist >= tgl->gen.size)
                FEError(-1440,EXIT,WRITE, "Tierra GenExTemp() error 2");
#endif /* ERROR */
#if PLOIDY == 1
            tgl->gbits[dist] |= 1;
#else /* PLOIDY == 1 */
            tgl->gbits[dist][ce->c.c->de] |= 1;
#endif /* PLOIDY == 1 */
        }
        if (who == 2)
        {   ogl = sl[ct->d.gen.size]->g[ct->d.gi];
            if (IsBit(ogl->bits, 0))
            {   ogl->bits |= (I32u) (ONE << (I32u) (7 + 4));
                dist = ad(adrt + i) - ct->mm.p;
                dist = ad(dist);
#ifdef ERROR
                if (ogl->genome == NULL || dist < 0 || dist >= ogl->gen.size)
                    FEError(-1441,EXIT,WRITE, "Tierra GenExTemp() error 3");
#endif /* ERROR */
#if PLOIDY == 1
                ogl->gbits[dist] |= (1 << 1);
#else /* PLOIDY == 1 */
                ogl->gbits[dist][ce->c.c->de] |= (1 << 1);
#endif /* PLOIDY == 1 */
            }
        }
    }
}

void GenExMov(to, from)
    I32s    to, from;
{
    Pcells  ct;
    Pgl     tgl, ogl;
    I32u    who;  /* 0 same cell; 1 daughter cell; 2 other cell; */
                  /* 3 free memory; 4 daughter of other cell */

    tgl = sl[ce->d.gen.size]->g[ce->d.gi];
    if (ce->d.flaw || ce->d.mut || !IsBit(tgl->bits, 0))
        return;
    /* the mov instruction being executed is within your own genome */
    if (ce->mm.p <= ce->c.c->ip && ce->c.c->ip < (ce->mm.p + ce->mm.s))
    {   ct = ce;
        who = WhoIs(&ct, from);    /* who is it moved from */
        if (who < 4) tgl->bits |= (I32u) (ONE << (I32u) (17 + who));
        else tgl->bits |= (I32u) (ONE << (I32u) (17 + 2));
        if (who == 2)
        {   ogl = sl[ct->d.gen.size]->g[ct->d.gi];
            if (IsBit(ogl->bits, 0))
            ogl->bits |= (I32u) (ONE << (I32u) (17 + 4));
        }
        ct = ce;
        who = WhoIs(&ct, to); /* who is it moved to */
        if (who < 4)
            tgl->bits |= (I32u) (ONE << (I32u) (22 + who));
        else tgl->bits |= (I32u) (ONE << (I32u) (22 + 2));
        if (who == 2)
        {   ogl = sl[ct->d.gen.size]->g[ct->d.gi];
            if (IsBit(ogl->bits, 0))
            ogl->bits |= (I32u) (ONE << (I32u) (22 + 4));
        }
    }
    else   /* these are moved from while executing instructions that */
    {   ct = ce;       /* are not your own */
        who = WhoIs(&ct, from);    /* who is it moved from */
        if (who < 4)
            tgl->bits |= (I32u) (ONE << (I32u) (27 + who));
        else tgl->bits |= (I32u) (ONE << (I32u) (27 + 2));
        if (who == 2)   /* ct is cell from which inst is moved */
        {   ogl = sl[ct->d.gen.size]->g[ct->d.gi];
            if (IsBit(ogl->bits, 0))
                ogl->bits |= (I32u) (ONE << (I32u) (27 + 4));
        }
    }
}

void GenExExe(adrt)
    I32s     adrt;
{
    Pcells  ct = ce;
    Pgl tgl;
    I32s    dist;
    I32u    who;  /* 0 same cell; 1 daughter cell; 2 other cell; */
                  /* 3 free memory; 4 daughter of other cell */

    tgl = sl[ce->d.gen.size]->g[ce->d.gi];
    if (ce->d.flaw || ce->d.mut || !IsBit(tgl->bits, 0))
        return;
    who = WhoIs(&ct, adrt);
    if (who < 4)
        tgl->bits |= (I32u) (ONE << (I32u) (2 + who));
    else tgl->bits |= (I32u) (ONE << (I32u) (2 + 2));
    if (!who)  /* who == 0 == same cell */
    {   dist = adrt - ce->mm.p;
#ifdef ERROR
        if (tgl->gbits == NULL || dist < 0 || dist >= tgl->gen.size)
            FEError(-1442,EXIT,WRITE, "Tierra GenExExe() error 0");
#endif /* ERROR */
#if PLOIDY == 1
        tgl->gbits[dist] |= 1;
#else /* PLOIDY == 1 */
        tgl->gbits[dist][ce->c.c->ex] |= 1;
#endif /* PLOIDY == 1 */
    }
    if (who == 2)  /* is other cell */
    {   tgl = sl[ct->d.gen.size]->g[ct->d.gi];
        if (IsBit(tgl->bits, 0))
        {   tgl->bits |= (ONE << (I32u) (2 + 4));
            dist = adrt - ct->mm.p;
#ifdef ERROR
            if (tgl->gbits == NULL || dist < 0 || dist >= tgl->gen.size)
                FEError(-1443,EXIT,WRITE, "Tierra GenExExe() error 1");
#endif /* ERROR */
#if PLOIDY == 1
            tgl->gbits[dist] |= (1 << 1);
#else /* PLOIDY == 1 */
            tgl->gbits[dist][ce->c.c->ex] |= (1 << 1);
#endif /* PLOIDY == 1 */
        }
    }
}

/* rationale for the functioning of the genebank:

The term ``rambank'' refers to a collection of genotypes maintained in RAM
The term ``diskbank'' refers to a collection of genotypes maintained on disk
The term ``genebank'' refers to both the rambank and the diskbank

Genotype names have two parts: size-label, for example 0080aaa, 0045eat,
6666god.

1) When a creature is born its genotype will be compared to that of its parent.
   A) if they are the same, the daughter will be given the same name as the
      mother.
   B) if they are not the same, the genebank will be searched.
      a) if the daughter genotype is found in the genebank, it will be given
         the same name as the genotype that it matches in the bank.
      b) if the daughter genotype does not match any genotype in the bank,
         a new name will be created for it, and it will be entered into the
         rambank.
2) For each birth and death a count of the population of both the genotype
   and the size class involved will be incremented or decremented, so that we
   have a count of the current population of each genotype and each size class.
   This information is maintained in rambank.
3) If a genotype frequency crosses a critical threshold, the genotype name
   will become permanent and the genotype will be saved to the diskbank.
   There may be several types of thresholds: proportion of population
   (e.g., 2%), proportion of soup, or just numbers of creatures.
4) When a genotype frequency drops to zero:
   A) If the genotype never crossed the thresholds, the genotype will be
      removed from the genebank, and its unique name will become available for
      reuse.
   B) If the genotype crossed the threshold, gaining a permanent name, it
      should be retained in the genebank.
5) Periodically, Tierra saves the complete state of the machine (e.g., every
   100 million instructions executed).  At that time, the complete rambank
   is saved to the diskbank.  For this reason, 4 A applies also to genotypes
   which never became permanent, and these must be removed from the diskbank
   as well.  The bitfield in the genotype structure tells us if a genotype is
   saved to the diskbank, and if it is permanent.
6) If the rambank becomes ``too full'', some relatively unused portion of it
   should be swapped to the diskbank.  In DOS, ``too full'' could be signaled
   by a malloc failure.  In unix, ``too full'' could be signaled by some
   specified limit on how big the rambank should get, if this seems wise.
   That portion of the rambank to be swapped to the diskbank might consist of
   the least recently accessed size class.  For this reason it might be
   worthwhile to maintain a queue of size classes, ordered by last use.
*/

