/* @(#)ttools.c    1.1 11/15/94*/
/* ttools.c   7-3-96  support for data analysis of the
 * Tierra Simulator V4.3: Copyright (c) 1991, 1992, 1993, 1994, 1995, 1996
 * Dan Pirone, Tom Ray & Virtual Life, originally written by Dan Pirone
 * MemProf labels: #21XX, in use: 2100-2101
 * FEError labels: #21XX, in use: 2102-2113
 */

#ifndef lint
static char ttools_sccsid[] = "@(#)ttools.c     1.7 7/21/92";
#endif

#include "tierra.h"
#ifdef __GUI__
#define WEXTERN_H
#endif /* __GUI__ */
#include "globals.h"
#include <signal.h>

#ifdef MON
#include "obs.h"
#include "tmon_global.h"
#endif /* MON */

#ifdef ALCOMM
#include "tmonitor.h"
#include "trequest.h"
#include <mlayer.h>
#endif

#ifdef AMIGA
int bg_compare(size_buf *i,size_buf *j);
int bc_compare(size_buf *i,size_buf *j);
int hc_compare(HistType *i,HistType *j);
int hs_compare(HistType *i,HistType *j);
int hg_compare(HistType *i,HistType *j);
#endif /* AMIGA */

/*-----------------------------------------------------------------------*/
/*-----------------------------------------------------------------------*/
/* routines for query_species */
/*-----------------------------------------------------------------------*/
int hg_compare(i,j)
HistType   *i,*j;
   {
   if(i->size == j->size)
     return(i->lbl - j->lbl);
   return(i->size - j->size);
   }
/*-----------------------------------------------------------------------*/
int hs_compare(i,j)
HistType   *i,*j;
   {
   return(i->size - j->size);
   }
/*-----------------------------------------------------------------------*/
int hc_compare(i,j)
HistType   *i,*j;
   {
   return(j->count - i->count);
   }
/*-----------------------------------------------------------------------*/
/* routines for query_size */
/*-----------------------------------------------------------------------*/
int bc_compare(i,j)
 size_buf *i,*j;
   {
   return(j->count - i->count);
   }
/*-----------------------------------------------------------------------*/
int bg_compare(i,j)
 size_buf *i,*j;
   {
     return(i->lbl - j->lbl);
   }
/*-----------------------------------------------------------------------*/

size_buf *InitSizeQuery(size_class, num_spec)
I32s size_class;
int num_spec;
{
   size_buf *buf;
   int qbufidx, genidx;

   buf=(size_buf *)thcalloc(num_spec, sizeof(size_buf), 2100);
   qbufidx=0;
   for(genidx=0; genidx<sl[size_class]->a_num; genidx++){
      if(sl[size_class]->g[genidx]){
         if(sl[size_class]->g[genidx]->pop){
            buf[qbufidx].lbl=genidx;
            buf[qbufidx].count=
               sl[size_class]->g[genidx]->pop;
            buf[qbufidx].moves=
               sl[size_class]->g[genidx]->d1.mov_daught;
            buf[qbufidx].flags=
               sl[size_class]->g[genidx]->d1.flags;
            buf[qbufidx].bits=sl[size_class]->g[genidx]->bits;
            qbufidx++;
         }
      }
   }

   /* sort by genome */
   qsort(buf, qbufidx, sizeof(size_buf), bg_compare);

   return buf;
}

void DispSizeQuery(buf, num_spec, size_class)
size_buf *buf;
int num_spec;
I32s size_class;
{
   int querypgidx, loopflag, displinecnt;
   char cc;

   displinecnt=(fe_lines-MSG_Y)-2;
   querypgidx=0;
   loopflag=1;
   while(loopflag){

      /* display query page */
      DispQueryPage(&querypgidx, buf, displinecnt, num_spec, size_class);

      /* menu bar - page up, page down, quit */
      sprintf(Fbuf,"U/D-page up/down  q-quit->");
      FEPrintf(HLP_X, HLP_Y, 1, 0, Fbuf);
#ifdef AMIGA
#if FRONTEND == BASIC
      nodelay(stdscr, FALSE);
#endif /* FRONTEND == BASIC */
#endif /* AMIGA */
      cc=FEGetch(HLP_X+strlen(&(Fbuf[0])), HLP_Y);
#ifdef AMIGA
#if FRONTEND == BASIC
      nodelay(stdscr, TRUE);
#endif /* FRONTEND == BASIC */
#endif /* AMIGA */

      switch(cc){
         case 'U':
            querypgidx+=displinecnt;
            break;
         case 'D':
            querypgidx-=displinecnt;
            break;
         case 'q':
            loopflag=0;
            break;
         default:
            break;
      }
   }
}

void DispQueryPage(querypgidx, buf, displinecnt, num_spec, size_class)
int *querypgidx;
size_buf *buf;
int displinecnt;
int num_spec;
I32s size_class;
{
   int loopflag, qbufidx, displine, maxquerypgidx, scrnline;

   maxquerypgidx=(displinecnt>num_spec)?0:num_spec-displinecnt;
   (*querypgidx)=((*querypgidx)<0)?
      0:(((*querypgidx)>maxquerypgidx)?
      maxquerypgidx:(*querypgidx));

   /* clear display area */
#if FRONTEND == BASIC
   FEClrmsg(MSG_Y, "");
   FEBanner();
   FEStats();
#endif /* FRONTEND == BASIC */

   /* display query heading */
    sprintf(Fbuf, "Gene:    #  Mem Errs   Move  Bits\n");
    FEPrintf(MSG_X, MSG_Y, 0, 0, Fbuf);
    if (Log && HistPrint)
        fprintf(tfp_log, Fbuf);

   /* display query lines */
   displine=0;
   loopflag=1;
   while(loopflag){
      qbufidx=(*querypgidx)+displine;
      scrnline=(MSG_Y+displine)+1;
      if(qbufidx<num_spec){
         if(displine<displinecnt){
            WritEcoB(buf[qbufidx].bits , Buff);
            sprintf(Fbuf, "%-3s %6ld  %3ld  %3ld    %3ld  %s\n",
               Int2Lbl(buf[qbufidx].lbl), buf[qbufidx].count,
               (I32s) (100.0 * buf[qbufidx].count *
               size_class / SoupSize),
               buf[qbufidx].flags, buf[qbufidx].moves, Buff);
            FEPrintf(MSG_X, scrnline, 0, 0, Fbuf);
            if(Log && HistPrint){
               fprintf(tfp_log, Fbuf);
            }
            displine++;
         }
         else{
            loopflag=0;
         }
      }
      else{
         loopflag=0;
      }
   }
   if(Log && HistPrint){
      fprintf(tfp_log,"\n");
#ifndef AMIGA
#ifndef DECVAX
      fflush(tfp_log);
#endif /* DECVAX */
#endif /* AMIGA */
   }
}
/*-----------------------------------------------------------------------*/

void query_species(num)
I32s num;
{   I32s num_ent = 0, nnn = 0, ci, t, n_star = 0, maxhistostrt;
#ifdef __GUI__
    char *ttbuff[80];
#endif /* __GUI__ */
    int i;

    if ((NumSizes < 1) || (num < 2))
        return;
    if (Hist)
    {   thfree((void*) Hist, HistAlocSiz, 2101);
        Hist = (HistType *) NULL;
        HistAlocSiz = 0;
    }
    Max_hits = -1;
    if (IMode == GEN_HIST)
        num_ent = NumGenotypes;
    else /* IMode == SIZ_HIST  || SIZM_HIST*/
        num_ent = NumSizes;
    Hist  = (HistType *) thcalloc((num_ent + 1) , sizeof(HistType), 2101);
    HistAlocSiz = (num_ent + 1) * sizeof(HistType);
    for(nnn= 0,ci = 1; (ci < siz_sl && nnn < num_ent); ci++)
    {   if (!TNULL(sl[ci]) && sl[ci]->num_c)
        {   if (IMode == GEN_HIST)
            {   for(t= 0; t < sl[ci]->a_num; t++)
                {   if((TNULL(sl[ci]->g[t])) || (sl[ci]->g[t]->pop < 1))
                        continue;
                    Hist[nnn].size = ci;
                    Hist[nnn].lbl =  t; /* int form of label */
                    Hist[nnn].count =  sl[ci]->g[t]->pop;
                    if (Hist[nnn].count >= Max_hits)
                        {Max_hits = Hist[nnn].count;}
                    nnn++;
                }
            }
            else /* IMode == SIZ_HIST  || SIZM_HIST */
            {   Hist[nnn].size = ci;
                Hist[nnn].count =  sl[ci]->num_c;
                if (Hist[nnn].count >= Max_hits)
                {   Max_hits = Hist[nnn].count;
                }
                nnn++;
            }
        } /* end of valid size class */
    }
    if (( Max_hits ) < 1)
        return;
    histoentcnt=nnn;
    nnn = ((nnn > num_ent)? num_ent : nnn);
    nnn = ((nnn > num)? num : nnn);
    nnn = ((nnn > fe_lines - MSG_Y) ? fe_lines - MSG_Y : nnn);
    histodsplnct=nnn;

    maxhistostrt=histoentcnt-histodsplnct;
    histostrtidx=(histostrtidx<0)?
       0:((histostrtidx>maxhistostrt)?
       maxhistostrt:histostrtidx);

    if (IMode == SIZM_HIST)
        for (t = histostrtidx; t < histostrtidx+histodsplnct; t++)
        {   Hist[t].count *= Hist[t].size;
            if (Hist[t].count >= Max_hits)
            {   Max_hits = Hist[t].count;
            }
        }
    if (IMode == GEN_HIST) /* sort top num, by size */
        qsort(Hist, num_ent-1, sizeof(HistType), hg_compare);
    else /* IMode == SIZ_HIST  || SIZM_HIST*/ /* sort top num, by size */
        qsort(Hist, num_ent-1, sizeof(HistType), hs_compare);
    Hist = (HistType *) threcalloc((void *) Hist,
        (histostrtidx+histodsplnct + 1) * sizeof(HistType), 
        (num_ent + 1) * sizeof(HistType), 2101);
    HistAlocSiz = (histostrtidx+histodsplnct + 1) * sizeof(HistType);
    HistSize = nnn;
    HistNStars =  ((float)(fe_width -20) /(float) Max_hits);
#ifndef __GUI__
/******
    FEClrmsg(MSG_Y,
        (FRONTEND == STDIO) ? "" :
        ((INTERFACE == BASICB) || (INTERFACE == BASICM)) ?
"                       Press Any Key for Menu ...                           "
        :
"                    Press Interrupt Key for Menu ...                        "
        );
****/
#else /* __GUI__ */
    FEClrmsg(MSG_Y, "");
#endif /* __GUI__ */

    for (t = histostrtidx, i=0; t < histostrtidx+histodsplnct; t++, i++)
    {   if (Hist[t].count < 1) continue;
#ifdef __GUI__
        strcpy(ttbuff, "*");
#else /* __GUI__ */
        Buff[0]=  '*'; Buff[1]=  '\0';
#endif /* __GUI__ */
        Hist[t].nstar = n_star =  HistNStars * Hist[t].count;
        if (IMode == GEN_HIST)
        {   if ((t > 0) && (Hist[t].size == Hist[t-1].size))
            {   sprintf(Fbuf, "      %3s %6ld | ",
                    Int2Lbl(Hist[t].lbl) ,Hist[t].count);
#ifndef __GUI__
                FEPrintf(MSG_X, MSG_Y + i, 0, 0, Fbuf);
                if (Log && HistPrint)
                    fprintf(tfp_log, Fbuf);
#endif /* __GUI__ */
            }
            else
            {   sprintf(Fbuf, " %5hd%3s %6ld | ",
                    Hist[t].size,Int2Lbl(Hist[t].lbl) ,Hist[t].count);
#ifndef __GUI__
                FEPrintf(MSG_X, MSG_Y + i, 0, 0, Fbuf);
                if (Log && HistPrint)
                    fprintf(tfp_log, Fbuf);
#endif /* __GUI__ */
            }
        }
        else /* IMode == SIZ_HIST || SIZM_HIST */
        {   sprintf(Fbuf, "%5hd %3hd %6ld | ",
                Hist[t].size,sl[Hist[t].size]->num_g,Hist[t].count);
#ifndef __GUI__
            FEPrintf(MSG_X, MSG_Y + i, 0, 0, Fbuf);
            if (Log && HistPrint)
                fprintf(tfp_log, Fbuf);
#endif /* __GUI__ */
        }

#ifdef __GUI__
        for(ci = 0; ci < n_star; ci++, strcat(ttbuff, "*"))
            ;
        strcat(Fbuf, ttbuff);
        strcat(Fbuf, "\n");
        FEPrintf(MSG_X, MSG_Y + i, 0, 0, Fbuf);
#else /* __GUI__ */
        for(ci = 0; ci < n_star; ci++, strcat(Buff, "*"))
            ;
        sprintf(Fbuf, "%s\n",Buff);
        FEPrintf((MSG_X + 19), MSG_Y + i, 0, 0, Fbuf);
#endif /* __GUI__ */
        if (Log && HistPrint)
            fprintf(tfp_log, Fbuf);
    }
    if (Log && HistPrint)
    {   fprintf(tfp_log,"\n");
#ifndef AMIGA
#ifndef DECVAX
        fflush(tfp_log);
#endif /* DECVAX */
#endif /* AMIGA */
    }
#ifdef MON
    MON_nnn = nnn;
#endif /* MON */
}        /* end of query_species */

/*-----------------------------------------------------------------------*/

#if FRONTEND == BASIC || MON

void query_spec_d(size,lbl)
I32s size,lbl;
{   I32s l, n, tcount, i, maxhistostrt;
#ifndef AMIGA
    I32s SigSaveSet;
#endif /* AMIGA */
    I16s empty = 0;
    I8s  old_n_star, ns_dlta;

    if((size > (siz_sl-1)) || (size < 0))
        return;
    if((Hist == NULL))
    {   query_species(fe_lines);
        return;
    }

#ifdef SIGBLOCK
#ifdef BSD
    SigSaveSet = sigblock(SigBlockSet);
#endif /* BSD */
#ifdef TSysV
    sighold(SIGINT);
#endif /* TSysV */
#endif /* SIGBLOCK */

    maxhistostrt=histoentcnt-histodsplnct;
    histostrtidx=(histostrtidx<0)?
       0:((histostrtidx>maxhistostrt)?
       maxhistostrt:histostrtidx);

    Buff[0]=  '*'; Buff[1]=  '\0';
    for (n=histostrtidx, i=0; n < histostrtidx+histodsplnct; n++, i++)
    {   if (TNULL(sl[Hist[n].size]))
        {   empty++;
            continue;
        }
        if (IMode == GEN_HIST)
        {   if (TNULL(sl[Hist[n].size]->g[Hist[n].lbl]))
            {   empty++;
                tcount = 0;
            }
            else
                tcount = sl[Hist[n].size]->g[Hist[n].lbl]->pop;
#ifdef MON
            MDGenHistBuf[n].size = Hist[n].size;
            strncpy (MDGenHistBuf[n].lbl, Int2Lbl (Hist[n].lbl), 3);
            MDGenHistBuf[n].count = tcount;
            MTB_GEN_HIST[n] = MON_UNREAD;
#endif /* MON */
        }
        else if (IMode == SIZM_HIST)
        {   tcount = size * sl[Hist[n].size]->num_c;
        }
        else
            tcount = sl[Hist[n].size]->num_c;
        if ((tcount < 1L) && (++empty > 3))
        {   query_species(fe_lines);
#ifdef SIGBLOCK
#ifdef BSD
            (void) sigsetmask(SigSaveSet);
#endif /* BSD */
#ifdef TSysV
            sigrelse(SIGINT);
#endif /* TSysV */
#endif /* SIGBLOCK */
            return;
        }
        if (Hist[n].size == size)
        {   if ((IMode == GEN_HIST) && (Hist[n].lbl != lbl))
                continue;
            if( tcount > 0L )
            {   old_n_star    = Hist[n].nstar;
                l             = HistNStars * tcount;
                Hist[n].nstar = (l > 255) ? 255 : l;
                ns_dlta = Hist[n].nstar - old_n_star;
            }
            else
            {   old_n_star = 0L;
                Hist[n].nstar = 0;
                ns_dlta = 0L;
                Buff[0] = ' ';
            }
            if (IMode == GEN_HIST)
            {   sprintf(Fbuf, "%6ld", tcount);
#ifdef __GUI__
                l = strlen(Fbuf);
                while (l--)
                    *((FEw + MSG_Y)->szDesc + MSG_X + 10 + l)
                        = *(Fbuf + l);
                FEPrintf(MSG_X, MSG_Y, 0, 0, (FEw + MSG_Y)->szDesc);
#else /* __GUI__ */
                FEPrintf(MSG_X + 10, MSG_Y + i, 0, 0, Fbuf);
#endif /* __GUI__ */
            }
            else
            {   sprintf(Fbuf, "%3hd %6ld", sl[Hist[n].size]->num_g, tcount);
#ifdef __GUI__
                l = strlen(Fbuf);
                while (l--)
                    *((FEw + MSG_Y)->szDesc + MSG_X + 6 + l)
                        = *(Fbuf + l);
                FEPrintf(MSG_X, MSG_Y, 0, 0, (FEw + MSG_Y)->szDesc);
#else /* __GUI__ */
                FEPrintf(MSG_X + 6, MSG_Y + i, 0, 0, Fbuf);
#endif /* __GUI__ */
            }
            if (Hist[n].nstar && !ns_dlta)
                break;
            if (Hist[n].nstar > (fe_width))
            {   Buff[0] = '+';
                old_n_star = 1+fe_width;
                ns_dlta = 0;
                Hist[n].nstar=fe_width;
            }
            else
            {   if ( ns_dlta > 1L)
                {   for(l= 0;l < ns_dlta;l++,strcat(Buff, "*"))
                        ;
                    ns_dlta=0;
                }
            }
#ifdef __GUI__
            l = strlen(Buff);
            while (l--)
                *((FEw + MSG_Y)->szDesc + MSG_X + 19 + old_n_star+ns_dlta)
                    = *(Buff + l);
            sprintf(Fbuf, "%s\n", (FEw + MSG_Y)->szDesc);
            FEPrintf(MSG_X, MSG_Y, 0, 0, Fbuf);
#else /* __GUI__ */
            sprintf(Fbuf, "%s\n", Buff);
            FEPrintf((MSG_X + 19 + old_n_star + ns_dlta), MSG_Y + i, 0, 0,
                Fbuf);
#endif /* __GUI__ */
#ifdef MON
            if(IMode == SIZ_HIST) {
              MDSizeHistBuf[n].size = size;
              MDSizeHistBuf[n].num = sl[Hist[n].size]->num_g;
              MDSizeHistBuf[n].count = sl[Hist[n].size]->num_c;
              MTB_SIZ_HIST[n] = MON_UNREAD;
            }
            if(IMode == SIZM_HIST) {
              MDMemHistBuf[n].size = size;
              MDMemHistBuf[n].num = sl[Hist[n].size]->num_g;
              MDMemHistBuf[n].count = size * sl[Hist[n].size]->num_c;
              MTB_MEM_HIST[n] = MON_UNREAD;
            }
#endif /* MON */
            break;
        }
    }
#ifdef SIGBLOCK
#ifdef BSD
            (void) sigsetmask(SigSaveSet);
#endif /* BSD */
#ifdef TSysV
            sigrelse(SIGINT);
#endif /* TSysV */
#endif /* SIGBLOCK */
}

#endif /* FRONTEND == BASIC || MON */

#ifndef THINK_C

/*-----------------------------------------------------------------------*/
#ifdef unix
/*-----------------------------------------------------------------------*/
/* THIS CODE IS NOT FINISHED  */
#ifndef __SOLARIS2__
void T_sig_read(sig,code,scp,addr)
    I32s  sig,code;
    /* struct sigcontext *scp; */
    I32s *scp; /* DO NOT USE !!!!! */
    I8s  *addr;
#else /* __SOLARIS2__ */
void T_sig_read(sig)
I32s sig;
#endif /* __SOLARIS2__ */
{
    FILE *inf;
    I8s data[85];
    I32s i;
    struct s_list **tsl;
    signal(SIGUSR1, T_sig_read);
    FEError(-2102, NOEXIT, NOWRITE,
        "==========================================================");
    sprintf(Fbuf, "TIERRA: re-read soup_in (%s) ...", soup_fn);
    FEError(-2103, NOEXIT, NOWRITE, Fbuf);
    inf = fopen(soup_fn, "r");
    if (inf == NULL)
    {   sprintf(Fbuf,
            "Tierra T_sig_read() GetSoupIn: file %s not opened, exiting",
                soup_fn);
        FEError(-2104, EXIT, NOWRITE, Fbuf);
    }
    fgets(data, 84, inf);
    sscanf(data, "tierra core:");       /* file header line */
    fgets(data, 84, inf);  /* blank line */
    fgets(data, 84, inf);
    while (strlen(data) > 3)
    {   if (!GetAVar(data))
        {   sprintf(Fbuf, "Tierra T_sig_read() bad soup_in line: %s", data);
            FEError(-2105, NOEXIT, NOWRITE, Fbuf);
        }
        fgets(data, 84, inf);
    }

    FEError(-2106,NOEXIT,NOWRITE,
        "==========================================================");
 ToggleLog();
}
#endif /* unix */

/*-----------------------------------------------------------------------*/
#ifdef unix
/*-----------------------------------------------------------------------*/
/* THIS CODE IS NOT FINISHED  */
#ifndef __SOLARIS2__
void T_sig_write(sig,code,scp,addr)
I32s  sig,code;
I32s *scp; /* DO NOT USE !!!!! */ /* struct sigcontext *scp; */
I8s  *addr;
#else /* __SOLARIS2__ */
void T_sig_write(sig)
I32s sig;
#endif /* __SOLARIS2__ */
{   signal(SIGHUP, T_sig_write);
    FEError(-2107, NOEXIT, NOWRITE,
         "==========================================================");
    FEError(-2108, NOEXIT, WRITE, "TIERRA: writing soup ...");
    FEError(-2109, NOEXIT, NOWRITE,
         "==========================================================");
}
#endif /* unix */

/*-----------------------------------------------------------------------*/
#ifdef unix
/*-----------------------------------------------------------------------*/
/* THIS CODE IS NOT FINISHED  */
#ifndef __SOLARIS2__
void T_sig_info(sig,code,scp,addr)
I32s  sig,code;
/* struct sigcontext *scp; */
I32s *scp; /* DO NOT USE !!!!! */
I8s  *addr;
#else /* __SOLARIS2__ */
void T_sig_info(sig)
I32s  sig;
#endif /* __SOLARIS2__ */
{   I32s t_hist_print = HistPrint;

    signal(SIGUSR2, T_sig_info);
    if (!HistPrint)
        HistPrint = 1L;
    sprintf(Fbuf,
        "InstExe.m    = %6ld  InstExec.i = %6ld  NumCells = %4ld ",
        InstExe.m,InstExe.i,NumCells);
    FEPrintf(MSG_X, MSG_Y, 0, 2, Fbuf);
    if (GeneBnker)
    {   sprintf(Fbuf,
            "NumGenotypes =   %4ld  NumSizes   =   %4ld",
            NumGenotypes, NumSizes);
        FEPrintf(MSG_X, MSG_Y + 1, 0, 2, Fbuf);
    }
    query_species(20);
    HistPrint = t_hist_print;
}
#endif /* unix */

#endif /* THINK_C */

void ToggleLog()
{
#ifndef NET
    if (Log  && (tfp_log == NULL))
    {   if(!new_soup)
            tfp_log = fopen("tierra.log","a");
        else
            tfp_log = fopen("tierra.log","w");
        if (tfp_log  == NULL)
            FEError(-2110, NOEXIT, NOWRITE,
                "Tierra ToggleLog() Problem opening - tierra.log ");
    }
    if (!Log && ( tfp_log ))
    {   fclose(tfp_log);
        tfp_log = NULL;
    }
    sprintf(Fbuf, "TIERRA: LOG = %s, Histogram Logging = %s \n",
        (Log) ? "on" : "off", (HistPrint) ? "on" : "off");
    FEPrintf(MSG_X, MSG_Y + 1, 0, 2, Fbuf);
#else /* NET */
    I8s TierraLogFile[30];
#ifdef DYNIPADR
   sprintf(TierraLogFile,"tierra.log");
#else
#ifdef AMIGA
    if(getenv("HOSTNAME") == NULL)
        FEError(-2111, EXIT, NOWRITE, "setenv HOSTNAME missed.\n");
    sprintf(TierraLogFile1,"tierra-%s.log",getenv("HOSTNAME"));
    strncpy(&(TierraLogFile[0]),
      &(TierraLogFile1[0]),
      sizeof(TierraLogFile)-1);
    TierraLogFile[sizeof(TierraLogFile)-1]='\0';
#else
    if(getenv("HOST") == NULL)
        FEError(-2111, EXIT, NOWRITE, "setenv HOST missed.\n");
    sprintf(TierraLogFile,"tierra-%s.log",getenv("HOST"));
#endif /* AMIGA */
#endif /* DYNIPADR */
    if (Log  && (tfp_log == NULL))
    {   if (!new_soup)
            tfp_log = fopen(TierraLogFile,"a");
        else
            tfp_log = fopen(TierraLogFile,"w");
        if(tfp_log  == NULL)
        {   sprintf(Fbuf,
                "Tierra ToggleLog() Problem opening - %s", TierraLogFile);
            FEError(-2112, NOEXIT, NOWRITE, Fbuf);
        }
    }
    if (!Log && ( tfp_log ))
    {   fclose(tfp_log);
        tfp_log = NULL;
    }
    sprintf(Fbuf, "TIERRA: LOG = %s, Histogram Logging = %s \n",
        (Log) ? "on" : "off", (HistPrint) ? "on" : "off");
    FEPrintf(MSG_X, MSG_Y, 0, 2, Fbuf);
#endif /* NET */
}

void FESoupImage(fp, x, y, n)
FILE * fp;
I32s x, y, n;
{   I32s addr, fsz, sz;
    register int nfp = 0;
    Pcells cp;
    I8s tmd;
    float cpb;

    if (!fp && (nfp = 1) && ( (fp = fopen("soup.img","w")) == NULL))
        FEError(-2113, NOEXIT, NOWRITE,
            "FESoupImage - can't open file soup.img\n");
    addr = 0;
    cpb = (float) ( x * y ) / (float) SoupSize;     /* char per byte */
    fsz = cpb * ( MinCellSize -1);  /* chars per Free step */
    while(1)
    {   if (IsFree(addr))
        {   FEPrintSI(fp,addr,x,n,fsz, 0);    /* 0 is char for Free */
            addr += MinCellSize -1;
        }
        else
        {   WhichCell(addr,&cp,&tmd);            /*chars for this size */
            sz = cpb * (tmd == 'm') ? cp->mm.s : cp->md.s;
            FEPrintSI(fp,addr,x,n,sz,tmd);
            addr += cp->mm.s + 1;
        }
        if ( addr >= SoupSize)
            break;
        addr = ad(addr);
    }
    if (nfp)
        fclose(fp);
}

void FEPrintSI(fp, addr, x, n, sz, md)
FILE * fp;
I32s x, addr, n, sz, md;
{   I32s l;
    I8s ch;

    ch =  (md) ? ((md == 'm') ? ('A' + (sz % n)) : ('a' + (sz % n))) : '.';
    for (l = 0; l < sz; l++)
        fprintf(fp, "%c",ch);
    if (((addr % x) + sz) > x)
        fprintf(fp, "\n");
}

/*-----------------------------------------------------------------------*/
