
/* crithist.c  18-12-92  code for computing avelanche distributions of the
      diversity indices */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <ctype.h>
#ifdef __TURBOC__
#include <alloc.h>
#endif /* __TURBOC__ */

#ifdef __STDC__           /* ANSI prototyping */
#define P_(A) A
#define const const
#else                 /* non-ANSI prototyping */
#define P_(A) ()
#define const
#endif

typedef long int       I32s;

struct range {
    double  n;  /* minimum value */
    double  x;  /* maximum value */
    I32s    m;  /* count of values */
    int     f;  /* first time through = 1 */
    } range, lrange, start, prop;

typedef struct {
    double  AvSize;
    double  StartVal;
    } Avel;

void main P_((int  argc, char  **argv));
void minmax P_((double  v, struct range  *r));

I32s    Num, *Bins, NumBins = 512;
double  Min, Max, BinMin, BinMax, BinWidth;
char    infile[13], ofile[13], vrbl[9], data[81];
Avel    av;
FILE    *inf, *ouf;

void main (argc, argv)
int  argc;
char  **argv;
{
#ifdef __TURBOC__
    float  f = sin(0);
#endif /* __TURBOC__ */
    int  insiz, i;

    if (argc <= 1)
    {
        printf ("crithist tool: to accept default values, hit Enter\n\n");

        printf("takes input from: numcell.cri numsize.cri sizediv.cri\n");
        printf("    or: agesize.cri numgeno.cri genodiv.cri agegeno.cri\n\n");

        printf("input file = ", infile);
        gets(data);
        sprintf(infile, "%s", data);
        sscanf(data,"%[^.]", vrbl);

        inf = fopen("critsums","r");
        if(inf == NULL)
        {   printf("file critsums not found\n");
            return ;
        }
/*
Avelanche Size Ranges:

             Minimum      Maximum     Number

numcell            1           87       7635
numsize            1            8        409
sizediv   2.6226e-06     0.347835       9026
agesize  0.000840439      2.30466        842
numgeno            1           19       1756
genodiv  2.86102e-06     0.432941       9498
agegeno  5.59311e-05     0.576081       4097
*/

        while(1)
        {   if (fgets(data,80,inf) == NULL)
            {   printf("input variable %s not recognized\n", vrbl);
                return;
            }
            if (!strncmp(data,vrbl,7))
                break;
        }
        sscanf(data,"%*s%lf%lf%ld", &Min, &Max, &Num);
        fclose(inf);

        inf = fopen(infile,"rb");
        if(inf == NULL)
        {   printf("file %s not found\n", infile);
            return ;
        }

        printf("Number of data: %ld\n", Num);
        printf("Number of bins for histogram (default: %ld) = ", NumBins);
        gets(data);
        if (strlen(data))
        {   sscanf(data,"%ld", &NumBins);
        }

        printf("Log data range: %g to %g\n", log10(Min), log10(Max));
        printf("Enter BinMin BinMax: ");
        gets(data);
        sscanf(data, "%lf%lf", &BinMin, &BinMax);
        BinWidth = (BinMax - BinMin) / (double) NumBins;

    }
    range.f = lrange.f = start.f = prop.f = 1;
    
    Bins = (I32s *) calloc(NumBins, sizeof(I32s));
    if (!Bins)
    {   printf("calloc(NumBins, sizeof(I32s)) failure\n");
        printf("NumBins = %ld  sizeof(I32s) = %d\n", NumBins, sizeof(I32s));
        exit(0);
    }

    for(;;)
    {   if (!fread(&av, sizeof(Avel), 1, inf))
            break;
        minmax(av.AvSize, &range);
        minmax(log10(av.AvSize), &lrange);
        minmax(av.StartVal, &start);
        minmax(av.AvSize / av.StartVal, &prop);
        ++(Bins[(int) ((log10(av.AvSize) - BinMin) / BinWidth)]);
    }
    fclose(inf);
    sprintf(ofile,"%s.his",vrbl);
    ouf = fopen(ofile,"w");
    fprintf(ouf,"%s Histogram:\n\n", vrbl);
    fprintf(ouf,"%s range: %le to %lf  N = %ld\n",
        vrbl, range.n, range.x, range.m);
    fprintf(ouf,"%s Log range: %lf to %lf  N = %ld\n",
        vrbl, lrange.n, lrange.x, lrange.m);
    fprintf(ouf,"StartVal range: %lf to %lf  N = %ld\n",
        start.n, start.x, start.m);
    fprintf(ouf,"AvSize / StartVal range: %le to %lf  N = %ld\n",
        prop.n, prop.x, prop.m);
    fprintf(ouf,"BinMin = %g  BinMax = %g  BinWidth = %g\n\n",
        BinMin, BinMax, BinWidth);
    for (i = 0; i < NumBins; i++)
        fprintf(ouf,"%ld\n", Bins[i]);
    fclose(ouf);
}

void minmax(v, r)
double  v;
struct range  *r;
{   if(r->f)
    {   r->f = 0;
        r->n = r->x = v;
    }
    if(v < r->n) r->n = v; if(v > r->x) r->x = v;
    r->m++;
}
