
/* genomebd.c  20-11-92  tool to extract genomes using birth death records,
   by scanning the break.X files for a frequency threshold,
   the selected genomes will be extracted from the genebank
   Compile this file together with arg_inc.c */
/*** genomebd: Version 1.0  Copyright (c) 1992  Tom Ray ***/

#include <ctype.h>
#include "license.h"
#include "tierra.h"

typedef unsigned int   Uint;
typedef unsigned long  Ulong;

struct last_out {
    Ulong  time;   /* += itime % 1000000 */
    Ulong  ctime;  /* count of millions of cpu cycles */
    Ulong  itime;  /* time interval since last birth or death */
    char   bd;
    Uint   size;
    char   label[4];
    } lo;

struct gene_dat {
    struct tnode  *t;
    long          index;
    } ;

struct pop_dat {
    long  pop;  /* current population of this size or genotype */
    char  saved;  /* 0 = not saved to disk, 1 = saved to disk */
    } ;

struct tnode {
    long            size;  /* genome size */
    struct pop_dat  sd;    /* pop_dat for this size class */
    int             gsize; /* allocated size of *g array */
    struct pop_dat  *g;    /* array of pop_dat structures */
    struct tnode    *l;    /* left sub-tree */
    struct tnode    *r;    /* right sub-tree */
    } ;

struct tnode * AddTree P_((struct tnode  *p));
void ExtractGen P_((int  pop));
int t_read P_((char  data[], struct last_out  *lo, int  *first,
     int  *genotypes));

int     genotypes, reaped = 0, NumCells = 0, NumExtr = 0;
int     StartTime, EndTime, ThrStFrq;
I32s    mtime = 0, otime = 0;
float   thr, ThrSt;
double  ttime;
char    infile[13], ofile[13], bifile[13], lop;
char    indir[80], gbdir[80], opdir[80], data[81];
FILE    *inf, *ouf;

/* begin needed to link with genio.c */
indx_t  GIndx;
I32s    GFormat, NopS, Nop1, Nop0;
I8s     Swap, mes[10][80];
InstDef *id;
/* end needed to link with genio.c */

/* begin for debugging: */
/* double         dtime = 0., a = 0, b = 1; */
/* end for debugging */

int main (argc, argv)
int  argc;
char  *argv[];
{   unsigned long  size;
    int            binum = 1;
    int            first = 1, first2 = 1;
    char           c, buf[80];
    struct tnode   *troot = NULL;
    double         tmp;

    ttime = 0.;

    sprintf(infile,"break.1");
    sprintf(indir,"td/");
    sprintf(gbdir,"gb1/");
    sprintf(opdir,"genomes/");
    thr = .05; ThrSt = -.005; ThrStFrq = 250; StartTime = EndTime = 0;

    if (argc <= 1)
    {   printf ("breakgen tool: to accept default values, hit Enter\n\n");

        printf("input file (default: %s) = ", infile);
        gets(buf);
        if (strlen(buf))
            sprintf(infile, "%s", buf);

        printf("input directory (default: %s) = ", indir);
        gets(buf);
        if (strlen(buf))
            sprintf(indir, "%s", buf);

        sprintf(data, "%s%s", indir, infile);
        inf = fopen(data,"r");
        if(inf == NULL)
        {   printf("file %s not found\n", infile);
            return ;
        }
        fclose(inf);

        printf("genebank directory (default: %s) = ", gbdir);
        gets(buf);
        if (strlen(buf))
            sprintf(gbdir, "%s", buf);

        printf("output directory (default: %s) = ", opdir);
        gets(buf);
        if (strlen(buf))
            sprintf(opdir, "%s", buf);

        printf("save threshold (default: %f) = ", thr);
        gets(buf);
        if (strlen(buf))
            sscanf(buf, "%f", &thr);

        printf("threshold step (default: %f) = ", ThrSt);
        gets(buf);
        if (strlen(buf))
            sscanf(buf, "%f", &ThrSt);

        printf("threshold step frequency (default: %d) = ", ThrStFrq);
        gets(buf);
        if (strlen(buf))
            sscanf(buf, "%d", &ThrStFrq);

        printf("Start Time in millions (default: %d) = ", StartTime);
        gets(buf);
        if (strlen(buf))
            sscanf(buf, "%d", &StartTime);

        printf("End Time in millions (default: %d) = ", EndTime);
        gets(buf);
        if (strlen(buf))
            sscanf(buf, "%d", &EndTime);
    }
        
    sprintf(data, "%sgenomebd.log", opdir);
    ouf = fopen(data,"w");
    if (!ouf)
    {   sprintf(data,"%s", opdir);
        data[strlen(opdir) - 1] = 0;
        mkdir(data);
        sprintf(data, "%sgenomebd.log", opdir);
        ouf = fopen(data,"w");
        if(!ouf)
        {   printf("\n Output File %s NOT opened. \n", data);
            exit(1);
        }
    }
    sprintf(data, "%sopcode.map", gbdir);
    GetAMap(data);
    sscanf(infile,"%[^.].%d", bifile, &binum);
    sprintf(data, "%s%s", indir, infile);
    inf = fopen(data,"r");
    if(inf == NULL)
    {   printf("\n Input File %s NOT found. \n", data);
        exit(1);
    }
    lo.time = lo.ctime = lo.size = 0;
    for(;;)
    {   if(fgets(data,80,inf) == NULL)
        {   binum++;
            sprintf(infile,"%s%s.%d", indir, bifile, binum);
            fclose(inf);
            inf = fopen(infile,"r");
            if(inf == NULL)
                break ;
            if(fgets(data,80,inf) == NULL)
                break ;
        }
        t_read(data, &lo, &first, &genotypes);

/* begin: for debugging purposes */
/*      if (ttime > dtime)
            a = b;
 */
/* end: for debugging purposes */

        if(first2)
        {   first2 = 0;
            if (!genotypes)
            {   printf("input file %s does not contain genotypes, exiting\n",
                    infile);
                exit(1);
            }
#ifdef __TURBOC__
            printf("\nTime = %4ld million  Coreleft = %6lu\r",
                0L, coreleft());
#else  /* __TURBOC__ */
            if (argc <= 1)
                printf("\nTime = %4d million\r", first2);
#endif /* __TURBOC__ */
            fflush(stdout);
            lop = 't';
        }
        troot = AddTree(troot);

        ttime += (double) lo.itime;
        mtime = (I32s) (ttime / 1000000L);
        if (mtime > otime)
        {
#ifdef __TURBOC__
            printf("Time = %4ld million  Coreleft = %6lu\r",
                mtime, coreleft());
#else  /* __TURBOC__ */
            if (argc <= 1)
                printf("Time = %4ld million\r", mtime);
#endif /* __TURBOC__ */
            fflush(stdout);
            lop = 't';
            otime = mtime;
            if (mtime && !(mtime % ThrStFrq))
                thr += ThrSt;
            if (EndTime && mtime > EndTime)
                break;
        }
    }
}

struct tnode * AddTree(p)
struct tnode  *p;
{   int  i, j, osize;
    struct pop_dat  *pd;

    if(p == NULL) /* this is a new size class */
    {   if(lo.bd == 'd')
        {   printf("new node is a death, exiting\n");
            fprintf(ouf,"new node is a death, exiting\n");
            exit(0);
        }
        p = (struct tnode  *) calloc(1,sizeof(struct tnode));
        if(p == NULL)
        {   printf("calloc failure, exiting\n");
            fprintf(ouf,"calloc failure, exiting\n");
            exit(0);
        }
        p->size = lo.size;         /* initialize this size */
        p->sd.pop = 1;
        i = Lbl2Int(lo.label);  /* allocate genotype array */
        p->gsize = i + 1;       /* & initialize first genotype */
        p->g = (struct pop_dat  *)
            calloc(p->gsize,sizeof(struct pop_dat));
        if(p->g == NULL)
        {   printf("calloc failure, exiting\n");
            fprintf(ouf,"calloc failure, exiting\n");
            exit(0);
        }
        pd = p->g + i;
        pd->pop = 1;
        p->l = p->r = NULL;
        NumCells++;
        if (reaped && !pd->saved && ((float) pd->pop / (float) NumCells > thr)
            && (mtime >= StartTime))
        {   ExtractGen(pd->pop);
            pd->saved = 1;
        }
    }
    else if(lo.size < p->size) /* if not this size */
        p->l = AddTree(p->l);
    else if(lo.size > p->size)
        p->r = AddTree(p->r);
    else                        /* this size, but not new */
    {   i = Lbl2Int(lo.label); /* reallocating gsize array */
        if(i >= p->gsize)
        {   osize = p->gsize;
            p->gsize = i + 5;
            if(p->g == NULL)
                p->g = (struct pop_dat  *)
                    calloc(p->gsize, sizeof(struct pop_dat));
            else
                p->g = (struct pop_dat  *)
                    realloc(p->g, p->gsize * sizeof(struct pop_dat));
            if(p->g == NULL)
            {   printf("realloc failure, exiting\n");
                fprintf(ouf,"realloc failure, exiting\n");
                exit(0);
            }
            for(j = osize; j < p->gsize; j++)
            {   pd = p->g + j;
                pd->pop = 0;
            }
        }
        pd = p->g + i;
        if(lo.bd == 'b') /* this is a birth */
        {   p->sd.pop++;
            pd->pop++;
            NumCells++;
            if (reaped && !pd->saved &&
                (float) pd->pop / (float) NumCells > thr)
            if (reaped && !pd->saved &&
                ((float) pd->pop / (float) NumCells > thr)
                && (mtime >= StartTime))
            {   ExtractGen(pd->pop);
                pd->saved = 1;
            }
        }
        else /* this is a death */
        {   p->sd.pop--;
            pd->pop--;
            NumCells--;
            reaped = 1;
/* BEGIN NEW CODE */
            if (!p->sd.pop)/* eliminate gen array if size class extinct */
            {   if (p->g)
                {  free(p->g);
                   p->g = NULL;
                }
                p->gsize = 0;
            }
/* END NEW CODE */
        }
    }
    return p;
}

void ExtractGen(int  pop)
{
    char    gename[9], ifile[90], ofile[90];
    I32s    i, j;
    FILE    *afp = 0;
    GList   *g = 0;
    head_t  head;
    indx_t  *indx, *tindx, gindx;

    sprintf(gename, "%04hd%3s", lo.size, lo.label);
    sprintf(ifile, "%s%04hd.gen", gbdir, lo.size);
    sprintf(ofile, "%s%s", opdir, gename);
    if (!(afp = fopen(ifile, "rb")))
    {   printf("file %s not opened\n", ifile);
        fprintf(ouf,"file %s not opened\n", ifile);
        exit(9);
    }
    head = read_head(afp);
#ifdef __TURBOC__
    indx = &gindx;
#else  /* __TURBOC__ */
    indx = read_indx(afp, &head);
#endif /* __TURBOC__ */
    if ((j = find_gen(afp, indx, lo.label, head.n)) == head.n)
    {   fprintf(stderr, "%s not in archive\n", lo.label);
        fprintf(ouf, "%s not in archive\n", lo.label);
    }
#ifdef __TURBOC__
    tindx = indx;
#else  /* __TURBOC__ */
    tindx = &indx[j];
#endif /* __TURBOC__ */
    g = get_gen(afp, &head, tindx, j);
    fclose(afp);
    WritAscFile(g, ofile);
    NumExtr++;
    if (lop == 't')
    {   lop = 'g';
        printf("\n");
    }
    printf("%04hd%3s  %6d    %d/%d   %5ld,%06ld\n",
        lo.size, lo.label, NumExtr, pop, NumCells, mtime, lo.time);
    fprintf(ouf,"%04hd%3s  %6d    %d/%d   %5ld,%06ld\n",
        lo.size, lo.label, NumExtr, pop, NumCells, mtime, lo.time);
    if (g)
    {   if (g->genome)
        {   free(g->genome);
            g->genome = NULL;
        }
        if (g->gbits)
        {   free(g->gbits);
            g->gbits = NULL;
        }
        free(g);
        g = NULL;
    }
}

I16s Lbl2Int(s)
char  s[];
{   if(s[0] == '-') return 0;
    return 1 + (s[2]- 'a') + (26 * (s[1] - 'a')) + (676 * (s[0] - 'a'));
}

int t_read(data, lo, first, genotypes)
char  data[];
struct last_out  *lo;
int  *first;
int  *genotypes;
{   struct last_out  ti;
    int    nargs;
    char   v2[9], v3[9], v4[9];

    sscanf(data,"%s", v2);
    if(!strcmp(v2,"num_sp")) return 0;
    nargs = sscanf(data,"%lx%s%s%s", &ti.time, v2, v3, v4);
    lo->itime = ti.time;
    if(*first)
    {   *first = 0;
        if(nargs == 4) *genotypes = 1;
        else *genotypes = 0;
        lo->time += ti.time;     /* assumes lo structure initialized to zero */
        if(lo->time >= 1000000L)
        {   lo->time %= 1000000L;
            lo->ctime++;
        }
        lo->bd = v2[0];
        sscanf(v3,"%u", &lo->size);
        if(*genotypes) strcpy(lo->label,v4);
        else strcpy(lo->label,"");
    }
    else
    {   lo->time += ti.time;
        if(lo->time >= 1000000L)
        {   lo->time %= 1000000L;
            lo->ctime++;
        }
        if(*genotypes) switch(nargs)
        {   case 1: break;
            case 2:
            {   if(isdigit(v2[0]))
                {   sscanf(v2,"%u", &lo->size); break; }
                else
                {   if(strlen(v2) == 1)
                    {   lo->bd = v2[0]; break; }
                    else
                    {   strcpy(lo->label,v2); break; }
                }
            }
            case 3:
            {   if(isdigit(v2[0]))
                {   sscanf(v2,"%u", &lo->size);
                    strcpy(lo->label,v3);
                }
                else
                {   lo->bd = v2[0];
                    if(isdigit(v3[0]))
                        sscanf(v3,"%u", &lo->size);
                    else
                        strcpy(lo->label,v3);
                }
                break;
            }
            case 4:
            {   lo->bd = v2[0];
                sscanf(v3,"%u", &lo->size);
                strcpy(lo->label,v4);
                break;
            }
        }
        else switch(nargs)
        {   case 1: break;
            case 2:
            {   if(isdigit(v2[0]))
                    sscanf(v2,"%u", &lo->size);
                else
                    lo->bd = v2[0];
                break;
            }
            case 3:
            {   lo->bd = v2[0];
                sscanf(v3,"%u", &lo->size);
                break;
            }
        }
    }
    return 1;
}
